#include <string>
#include <set>
#include <map>
#include <fstream>
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
#include "R3MyGraphics/R3MyGraphics.h"
using namespace std;
static char common_setting_filename[] = "configs/config.cfg";
static char specific_setting_filename[] = "configs/gen_seg_descriptors.cfg";
static int
ParseArgs(int argc, char **argv)
{
    // Parse arguments
    argc--; argv++;
    while (argc > 0) {
        if ((*argv)[0] == '-') {
			if (!strcmp(*argv, "-config"))
			{
				argv++; argc--;
				MyConfigParser.Import(*argv);
			}
            else if (!strcmp(*argv, "-Help")) {
                MyConfigParser.PrintAllSettings();
                exit(-1);
            }
			else {
				if (MyConfigParser.HasSection((*argv)+1)) {
					char* sec = (*argv) + 1;
					argv++; argc--;
					char* opt = *argv;
					argv++; argc--;
					char* val = *argv;
					MyConfigParser.Set(sec, opt, val);
				}
				else {
					fprintf(stderr, "Invalid program argument: %s", (*argv)+1);
					exit(1);
				}
			}
            argv++; argc--;
        }
        else {
            fprintf(stderr, "Invalid program argument: %s", *argv); exit(1);
            argv++; argc--;
        }
    }
    
    // Return OK status
    return 1;
}

static const map<string, vector<double> >
GetAreaFraction(PointCloudPropertySet* pSegmentation)
{
    PointCloud* pCloud = pSegmentation->GetPointCloud();
    map<string, vector<double> > descs;
    int totalNumPoints = pCloud->NPoints();
    for (int i=0; i<pSegmentation->NProperties(); i++) {
        PointCloudProperty* pSeg = pSegmentation->Property(i);
        descs[pSeg->Name()] = vector<double>();
        int count = 0;
        for (int j=0; j<totalNumPoints; j++) {
            if (pSeg->VertexValue(j) > 0.5) count ++;
        }
        descs[pSeg->Name()].push_back(double(count)/totalNumPoints);
    }
    return descs;
}

static const map<string, vector<double> >
GetBBoxDimensions(PointCloudPropertySet* pSegmentation)
{
    PointCloud* pCloud = pSegmentation->GetPointCloud();
    map<string, vector<double> > descs;
    int totalNumPoints = pCloud->NPoints();
    for (int i=0; i<pSegmentation->NProperties(); i++) {
        PointCloudProperty* pSeg = pSegmentation->Property(i);
        descs[pSeg->Name()] = vector<double>();
        R3Box box = R3null_box;
        for (int j=0; j<totalNumPoints; j++) {
            if (pSeg->VertexValue(j) > 0.5) {
                box.Union(pCloud->Vertex(j)->Position());
            }
        }
        descs[pSeg->Name()].push_back(box.XLength());
        descs[pSeg->Name()].push_back(box.YLength());
        descs[pSeg->Name()].push_back(box.ZLength());
    }
    return descs;
}

int main(int argc, char **argv)
{
	// Configs
    MyConfigParser.Import(common_setting_filename, TRUE);
    MyConfigParser.Import(specific_setting_filename, TRUE);
	if (!ParseArgs(argc, argv)) exit(-1);

    // input point cloud
    string pointCloudFilename = MyConfigParser.Get("Input", "point_cloud_filename");
    PointCloud pointCloud;
    pointCloud.ReadFile(pointCloudFilename.c_str());
    pointCloud.SetName(FileSystemUtil::GetBasename(pointCloudFilename, false).c_str());
    // input segmentation
    string segmentationFilename = MyConfigParser.Get("Input", "segmentation_filename");
    PointCloudPropertySet segmentation(&pointCloud);
    segmentation.ReadFile(segmentationFilename.c_str());
    
    map< string,vector<double> > descriptors;
    string descriptorType = MyConfigParser.Get("DescriptorType", "type");
    if (descriptorType == "area_fraction") {
        descriptors = GetAreaFraction(&segmentation);
    }
    else if (descriptorType == "bbox_dimensions") {
        descriptors = GetBBoxDimensions(&segmentation);
    }
    else {
        fprintf(stderr, "Unable to recognize descriptor type %s!\n", descriptorType.c_str());
        exit(-1);
    }
    // output
    // outputDirname/partName.sd
    string outputDirname = MyConfigParser.Get("Output", "shape_descriptor_dirname");
    FileSystemUtil::CreateDirectory(outputDirname);
    for (map<string, vector<double> >::const_iterator it=descriptors.begin();
         it!=descriptors.end(); it++) {
        const string& partName = it->first;
        const vector<double>& descs = it->second;
        string filename = outputDirname+"/"+partName+".sd";
        ofstream fout(filename.c_str());
        for (int i=0; i<(int)descs.size(); i++) {
            fout << descs[i] << " ";
        }
        fout.close();
    }
	return 0;
}
