#include "PointCloudViewer.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
using namespace std;
int PointCloudViewer::window_width = 0;
int PointCloudViewer::window_height = 0;

R3Vector PointCloudViewer::initial_camera_towards(0, 1.0, -0.3);
R3Vector PointCloudViewer::initial_camera_up(R3posz_vector);
R3Point PointCloudViewer::initial_camera_origin(0,0,0);
RNBoolean PointCloudViewer::initial_camera_flag = FALSE;

R3Viewer* PointCloudViewer::viewer = NULL;
bool PointCloudViewer::exit_flag = false;

string PointCloudViewer::saved_image_name = "";
int PointCloudViewer::saved_image_counter = 0;

PointCloud* PointCloudViewer::m_pPointCloud = NULL;
PointCloudPropertySet* PointCloudViewer::m_pPropertySet = NULL;
Voxelization* PointCloudViewer::m_pVoxelization = NULL;
vector<R3Box>* PointCloudViewer::m_pCuboidArray = NULL;
int PointCloudViewer::m_iCurrentPropertyIndex = -1;
vector< vector<int> > PointCloudViewer::region_array;
int PointCloudViewer::region_center_node_index = -1;

RNInterval PointCloudViewer::value_range = RNInterval(0, 0);
RNInterval PointCloudViewer::percentile_range = RNInterval(0, 0);

bool PointCloudViewer::show_values_in_text_flag = false;
bool PointCloudViewer::show_statistics_flag = true;
bool PointCloudViewer::show_normals_flag = false;
bool PointCloudViewer::show_node_indices_flag = false;
bool PointCloudViewer::show_selected_region_flag = false;
bool PointCloudViewer::show_voxel_flag = false;
bool PointCloudViewer::show_cuboids_flag = false;
bool PointCloudViewer::show_points_flag = true;

static void
ParseRangeDescriptionFile(const string& filename,
                          map<string, double>& minMapping, map<string, double>& maxMapping)
{
    ifstream fin(filename.c_str());
    string keyword;
    double minVal, maxVal;
    while (fin>>keyword>>minVal>>maxVal) {
        minMapping[keyword] = minVal;
        maxMapping[keyword] = maxVal;
    }
    fin.close();
}

void PointCloudViewer::
Init(int* argc, char **argv, const std::string& windowName, int width, int height,
     PointCloud* pCloud, PointCloudPropertySet* pPropertySet, Voxelization* pVoxel,
     vector<R3Box>* pCuboidArray)
{
	ConfigInit();
#ifdef GLUT_RENDERING
    GLUTInit(argc, argv, windowName, width, height);
#elif defined MESA_RENDERING
    MesaInit(width, height);
#endif
	// Window
	window_width = width;
	window_height = height;
    // Initialize background color
    glClearColor(1.0, 1.0, 1.0, 1.0);
	// Initialize lighting
	LightingInit();
    // Initialize graphics modes
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
	// Input data
    m_pPointCloud = pCloud;
    m_pPropertySet = pPropertySet;
    m_pVoxelization = pVoxel;
    m_pCuboidArray = pCuboidArray;
	// Init viewer
	viewer = CreateBirdsEyeViewer();
	// Viewport
	// glViewport(0, 0, window_width, window_height);
    // Check
    if (!m_pPointCloud) {
        fprintf(stderr, "PointCloudViewer::Init m_pPointCloud == NULL!\n");
        exit(-1);
    }
    percentile_range.Reset(10, 90);
    if (m_pPropertySet->NProperties() > 0) {
        m_iCurrentPropertyIndex = 0;
        RNScalar min_value = GetCurrentProperty()->Percentile(percentile_range.Min());
        RNScalar max_value = GetCurrentProperty()->Percentile(percentile_range.Max());
        value_range.Reset(min_value, max_value);
    }
    else m_iCurrentPropertyIndex = -1;
    // value range from external source
    if (MyConfigParser.IsEnabled("PropertyVis", "external_range_description_flag")) {
        string filename = MyConfigParser.Get("Input", "range_description_filename");
        if (!FileSystemUtil::PathExist(filename)) {
            fprintf(stderr, "PointCloudViewer::Init unable to find %s!\n", filename.c_str());
            exit(-1);
        }
        if (m_pPropertySet->NProperties() != 1) {
            fprintf(stderr, "PointCloudViewer::Init m_pPropertySet have >= 2 properties!\n");
            exit(-1);
        }
        map<string, double> minMapping, maxMapping;
        ParseRangeDescriptionFile(filename, minMapping, maxMapping);
        string query = "all:";
        query = query + m_pPropertySet->Property(0)->Name();
        if (minMapping.find(query) == minMapping.end() ||
            maxMapping.find(query) == maxMapping.end()) {
            fprintf(stderr, "PointCloudViewer::Init unable to find keyword %s in %s!\n",
                    query.c_str(), filename.c_str());
            exit(-1);
        }
        RNScalar min_value = minMapping[query];
        RNScalar max_value = maxMapping[query];
        value_range.Reset(min_value, max_value);
        printf("value_range is set [%g , %g] according to %s!\n", min_value, max_value, filename.c_str());
    }
    // Region array
    string region_array_filename = MyConfigParser.Get("Input", "selected_region_filename");
    if (region_array_filename != "NULL") {
        ReadRegionArray(region_array_filename);
        if ((int)region_array.size() >= 1) {
            region_center_node_index = 0;
        }
    }
}

void PointCloudViewer::
ReadRegionArray(const string& filename)
{
    vector<string> lines = FileUtil::ReadLines(filename);
    region_array.clear();
    for (int i=0; i<(int)lines.size(); i++) {
        vector<string> segs = StringUtil::Split(lines[i], ' ');
        vector<int> indices(segs.size());
        for (int j=0; j<(int)segs.size(); j++)
            indices[j] = TypeUtil::String2Integer(segs[j]);
        region_array.push_back(indices);
    }
}

void PointCloudViewer::
ConfigInit(void)
{
	// Starting camera
	std::vector<RNScalar> camera_towards_vec = MyConfigParser.GetScalarArray("StartingCamera", "initial_camera_towards");
	std::vector<RNScalar> camera_up_vec = MyConfigParser.GetScalarArray("StartingCamera", "initial_camera_up");
	initial_camera_towards[0] = camera_towards_vec[0];
	initial_camera_towards[1] = camera_towards_vec[1];
	initial_camera_towards[2] = camera_towards_vec[2];
	initial_camera_up[0] = camera_up_vec[0];
	initial_camera_up[1] = camera_up_vec[1];
	initial_camera_up[2] = camera_up_vec[2];
    // Voxel flag
    show_voxel_flag = MyConfigParser.IsEnabled("StartingCamera", "voxel_flag");
    // Points flag
    show_points_flag = MyConfigParser.IsEnabled("StartingCamera", "points_flag");
    // Cuboids flag
    show_cuboids_flag = MyConfigParser.IsEnabled("StartingCamera", "cuboids_flag");
}

void PointCloudViewer::
LightingInit(void)
{
	// Initialize lighting
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    std::vector<RNScalar> lmodel_ambient_v = MyConfigParser.GetScalarArray("StartingLighting", "ambient_light");
    static GLfloat lmodel_ambient[] = { lmodel_ambient_v[0], lmodel_ambient_v[1], lmodel_ambient_v[2], lmodel_ambient_v[3] };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_LIGHTING);
    
    // Initialize headlight 0
    std::vector<RNScalar> light0_diffuse_v = MyConfigParser.GetScalarArray("StartingLighting", "light0_diffuse");
    static GLfloat light0_diffuse[] = { light0_diffuse_v[0], light0_diffuse_v[1], light0_diffuse_v[2], light0_diffuse_v[3] };
    std::vector<RNScalar> light0_position_v = MyConfigParser.GetScalarArray("StartingLighting", "light0_position");
    static GLfloat light0_position[] = { light0_position_v[0], light0_position_v[1], light0_position_v[2], light0_position_v[3] };
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glEnable(GL_LIGHT0);
    
    // Initialize headlight 1
    std::vector<RNScalar> light1_diffuse_v = MyConfigParser.GetScalarArray("StartingLighting", "light1_diffuse");
    static GLfloat light1_diffuse[] = { light1_diffuse_v[0], light1_diffuse_v[1], light1_diffuse_v[2], light1_diffuse_v[3] };
    std::vector<RNScalar> light1_position_v = MyConfigParser.GetScalarArray("StartingLighting", "light1_position");
    static GLfloat light1_position[] = { light1_position_v[0], light1_position_v[1], light1_position_v[2], light1_position_v[3] };
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    glEnable(GL_LIGHT1);
	
}


double PointCloudViewer::
GetNormalizedValue(PointCloudVertex* pVertex)
{
    PointCloudProperty* pProperty = GetCurrentProperty();
    // return value between 0 and 1
    assert(pProperty);
    double value = pProperty->VertexValue(pVertex);
    double diameter = value_range.Diameter();
    if (diameter > 0) value = (value - value_range.Min()) / diameter;
    else if (value > value_range.Min()) value = 1;
    else if (value < value_range.Min()) value = 0;
    else value = 0.5;
    if (value < 0) value = 0;
    else if (value > 1) value = 1;
    return value;
}

RNRgb PointCloudViewer::
GetNormalizedColor(PointCloudVertex* pVertex)
{
    double value = GetNormalizedValue(pVertex);
    RNRgb c(0, 0, 0);
    // small values -> blue
    if (value < 0.5) {
        c[2] = 1 - 2 * value;
        c[1] = 2 * value;
    }
    // big values -> red
    else {
        c[1] = 1 - 2 * (value - 0.5);
        c[0] = 2 * (value - 0.5);
    }
    return c;
}

PointCloudProperty* PointCloudViewer::
GetCurrentProperty(void)
{
    if (m_iCurrentPropertyIndex<0 || m_iCurrentPropertyIndex>= m_pPropertySet->NProperties()) {
        return NULL;
    }
    else {
        return m_pPropertySet->Property(m_iCurrentPropertyIndex);
    }
}

void PointCloudViewer::
Draw(void)
{
    viewer->Camera().Load();
    glClearColor(1.0, 1.0, 1.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_LIGHTING);
    
    RNLength radius = 0.003 * m_pPointCloud->BBox().LongestAxisLength();
    // draw highlighted region
    if (show_selected_region_flag && region_center_node_index>=0) {
        // all nodes
        static GLfloat regularMaterial[] = {1,1,0,1};
        glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, regularMaterial);
        for (int i=0; i<m_pPointCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
            R3Sphere(pVertex->Position(), radius).Draw();
        }
        // highlighted nodes
        RNLength highlighted_radius = 1.2*radius;
        static GLfloat highlightedMaterial[] = {1,0,0,1};
        glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, highlightedMaterial);
        for (int i=0; i<(int)region_array[region_center_node_index].size(); i++) {
            PointCloudVertex* pVertex = m_pPointCloud->Vertex(region_array[region_center_node_index][i]);
            if (i == 0) R3Sphere(pVertex->Position(), 2.0*highlighted_radius).Draw();
            else R3Sphere(pVertex->Position(), highlighted_radius).Draw();
        }
    }
    // just draw point cloud
    else if (!GetCurrentProperty()) {
        if (show_points_flag) {
            static GLfloat regularMaterial[] = {0,1,0,1};
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, regularMaterial);
            for (int i=0; i<m_pPointCloud->NPoints(); i++) {
                PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
                R3Sphere(pVertex->Position(), radius).Draw();
            }
        }
    }
    // draw values
    else {
        static GLfloat material[4] = {0.8,0.8,0.8,1};
        for (int i=0; i<m_pPointCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
            RNRgb color = GetNormalizedColor(pVertex);
            material[0] = color[0];
            material[1] = color[1];
            material[2] = color[2];
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
            R3Sphere(pVertex->Position(), radius).Draw();
        }
    }
    // draw normals
    if (show_normals_flag) {
        glDisable(GL_LIGHTING);
        glLineWidth(3.0f);
        static GLfloat color[4] = {0,0,0,1};
        for (int i=0; i<m_pPointCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
            glColor4fv(color);
            R3Span(pVertex->Position(), pVertex->Position() + (4 * radius) * pVertex->Normal()).Draw();
        }
    }
    // draw voxels
    if (show_voxel_flag && !m_pVoxelization->IsEmpty()) {
        m_pVoxelization->Draw();
    }
    // draw cuboid
    if (show_cuboids_flag && m_pCuboidArray) {
        glDisable(GL_LIGHTING);
        glLineWidth(0.1f);
        glColor3f(0, 0, 0);
        for (int i=0; i<(int)m_pCuboidArray->size(); i++) {
            (*m_pCuboidArray)[i].Outline();
        }
    }
}

R3Viewer* PointCloudViewer::
CreateBirdsEyeViewer(void)
{
	// Setup camera view looking down the Z axis
	R3Box bbox = m_pPointCloud->BBox();
	assert(!bbox.IsEmpty());
	RNLength r = bbox.DiagonalRadius();
	if ((r < 0.0) || !RNIsFinite(r)) {
		fprintf(stderr, "Unable to initialize the viewpoint (r < 0.0 or r is infinity)\n");
		exit(-1);
	}
	if (!initial_camera_flag) {
		initial_camera_origin = bbox.Centroid() - initial_camera_towards * (3.0* r);
	}
	R3Camera camera(initial_camera_origin, initial_camera_towards, initial_camera_up, 0.4, 0.4, 0.01 * r, 100.0 * r);
	R2Viewport viewport(0, 0, window_width, window_height);
	return new R3Viewer(camera, viewport);
}

void PointCloudViewer::
TakeSnapshot(const std::string& filename)
{
    saved_image_name = filename;
}


