#include <fstream>
#include <sstream>
#include "ConfigParser.h"
#include "Utils/Utils.h"

using namespace std;
ConfigParser MyConfigParser;

ConfigParser::
ConfigParser(void)
:isLoaded(FALSE)
{}

void ConfigParser::
Import(const char* filename, RNBoolean create_new)
{
    ifstream fin(filename);
	if (!fin)
	{
		fprintf(stderr, "Unable to open setting file %s\n", filename);
		return;
	}
    string buffer;
	string current_section = "";
	int line_counter = 0;
	RNBoolean fail_flag = FALSE;
    while (!fin.eof())
    {
        getline(fin, buffer);
		line_counter ++;
		// comments
        if (buffer[0] == '#') continue;
		// section name
		else if (buffer[0] == '[') {
			size_t last_pos = buffer.find_last_of("]");
			if (last_pos == string::npos) {
				fprintf(stderr, "ConfigParser::Import syntax error found at"
						" line %d\n", line_counter);
				fail_flag = TRUE;
				break;
			}
			else {
				current_section = buffer.substr(1, last_pos-1);
			}
		}
		// option name
		else {
			// empty line
			if (buffer.length() == 0) continue;
			if (current_section.length() == 0) {
				fprintf(stderr, "ConfigParser::Import syntax error found at"
						" line %d\n", line_counter);
				fail_flag = TRUE;
				break;
			}
			size_t equal_pos = buffer.find_last_of("=");
			if (equal_pos == string::npos) {
				fprintf(stderr, "ConfigParser::Import syntax error found at"
						" line %d\n", line_counter);
				fail_flag = TRUE;
				break;
			}
			string opt_part = buffer.substr(0, equal_pos);
			string val_part = buffer.substr(equal_pos+1);
			Set(current_section, StringUtil::Trim(opt_part, " "),
				StringUtil::Trim(val_part, " "),
				create_new);
		}
    }
    fin.close();
	if (!fail_flag) {
		isLoaded = TRUE;
	}
	else
	{
		fprintf(stderr, "ConfigParser::Import fail to import %s\n",
				filename);
	}
    ImportFrequentlyUsedParameters();
}

void ConfigParser::
ImportFrequentlyUsedParameters(void)
{
    vector<double> direction = GetScalarArray("CurvatureHistogram", "prescribed_direction");
    prescribed_direction = R3Vector(direction[0], direction[1], direction[2]);
}

RNBoolean ConfigParser::
HasSection(const string& sec) const
{
	return (values.find(sec) != values.end());
}

RNBoolean ConfigParser::
HasOption(const string& sec, const string& opt) const
{
	if (!HasSection(sec)) return FALSE;
	map<string, map<string, string> >::const_iterator sec_map =
	values.find(sec);
	return ((*sec_map).second.find(opt) != (*sec_map).second.end());
}

RNBoolean ConfigParser::
IsLoaded(void) const
{
	return isLoaded;
}

void ConfigParser::
Set(const std::string& sec, const std::string& opt,
	const std::string& val, RNBoolean create_new)
{
	if (!create_new && !HasOption(sec, opt)) {
		fprintf(stderr, "ConfigParser::Set Option %s::%s doesn't exist!\n", sec.c_str(), opt.c_str());
		return;
	}
	values[sec][opt] = val;
}

const string ConfigParser::
Get(const std::string& sec, const std::string& opt) const
{
	if (!HasOption(sec, opt)) {
		fprintf(stderr, "ConfigParser::Get Unable to find %s::%s\n",
				sec.c_str(), opt.c_str());
	}
	map<string, map<string, string> >::const_iterator sec_map = values.find(sec);
	map<string, string>::const_iterator opt_map = (*sec_map).second.find(opt);
	return (*opt_map).second;
}

RNScalar ConfigParser::
GetScalar(const std::string& sec, const std::string& opt) const
{
	return TypeUtil::String2Number(Get(sec, opt));
}

int ConfigParser::
GetInt(const std::string& sec, const std::string& opt) const
{
	return int(GetScalar(sec, opt));
}

const char* ConfigParser::
GetCharArray(const std::string& sec, const std::string& opt) const
{
	return Get(sec, opt).c_str();
}

const std::vector<std::string> ConfigParser::
GetStringArray(const std::string& sec, const std::string& opt) const
{
    std::string str = Get(sec, opt);
	return StringUtil::Split(str, ',');
}

const std::vector<RNScalar> ConfigParser::
GetScalarArray(const std::string& sec,
			   const std::string& opt) const
{
	std::string str = Get(sec, opt);
	std::vector<RNScalar> scalar_array;
    // an empty array
    if (str == "NULL") {
        return scalar_array;
    }
	std::vector<std::string> segments = StringUtil::Split(str, ',');
	for (int i=0; i<int(segments.size()); i++) {
		scalar_array.push_back(TypeUtil::String2Number(segments[i]));
	}
	return scalar_array;
}

const std::vector<int> ConfigParser::
GetIntArray(const std::string& sec, const std::string& opt) const
{
	std::string str = Get(sec, opt);
	std::vector<std::string> segments = StringUtil::Split(str, ',');
	std::vector<int> intArray;
	for (int i=0; i<int(segments.size()); i++) {
		intArray.push_back(TypeUtil::String2Integer(segments[i]));
	}
	return intArray;
}

RNBoolean ConfigParser::
IsEnabled(const std::string& sec, const std::string& opt) const
{
	return Get(sec, opt) == "on";
}

void ConfigParser::
PrintAllSettings(void) const
{
    printf("ConfigParser::PringAllSettings ...\n");
    for (map< string, map<string,string> >::const_iterator it1=values.begin();
         it1!=values.end(); it1++) {
        printf("[%s]\n", it1->first.c_str());
        const map<string,string>& ref = it1->second;
        for (map<string,string>::const_iterator it2=ref.begin();
             it2!=ref.end(); it2++) {
            printf("%20s = %s\n", it2->first.c_str(), it2->second.c_str());
        }
        printf("\n");
    }
}
