#include <string>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
#include "RNBasics/RNBasics.h"
// How to use clapack in Linux and mac: http://www.ualberta.ca/~kbeach/lapack.html
#ifdef __APPLE__
#include <Accelerate/Accelerate.h>
typedef __CLPK_integer integer;
#else
typedef int integer;
extern "C" double dlange_(char *norm, integer *m, integer *n, double *a, integer *lda, double *work);
extern "C" int dgelsy_(integer *m, integer *n, integer *nrhs, double *a, integer *lda, double *b, integer *ldb, 
        integer * jpvt, double *rcond, integer *rank, double *work, integer * lwork, integer *info);
#endif
// end
using namespace std;
using namespace ShapeDescriptors;

// have to use typedef in this case
typedef double (*funcPtr)(double, double, const R3Vector&, const R3Vector&);

// curvature features
static double
K1(double k1, double k2,
   const R3Vector& v1, const R3Vector& v2)
{
    return k1;
}

static double
K2(double k1, double k2,
   const R3Vector& v1, const R3Vector& v2)
{
    return k2;
}

static double
K1MinusK2(double k1, double k2,
          const R3Vector& v1, const R3Vector& v2)
{
    return k1-k2;
}

static double
K1Ordered(double k1, double k2,
          const R3Vector& v1, const R3Vector& v2)
{
    if ( fabs(k1) < fabs(k2) ) std::swap( k1, k2 );
    return k1;
}

static double
K2Ordered(double k1, double k2,
          const R3Vector& v1, const R3Vector& v2)
{
    if ( fabs(k1) < fabs(k2) ) std::swap( k1, k2 );
    return k2;
}

static double
K1MinusK2Ordered(double k1, double k2,
                 const R3Vector& v1, const R3Vector& v2)
{
    if ( fabs(k1) < fabs(k2) ) std::swap( k1, k2 );
    return k1-k2;
}

static double
K1K2Average(double k1, double k2,
            const R3Vector& v1, const R3Vector& v2)
{
    return (k1+k2)*0.5;
}

static double
K1K2Product(double k1, double k2,
            const R3Vector& v1, const R3Vector& v2)
{
    return k1*k2;
}

static double
K1Abs(double k1, double k2,
      const R3Vector& v1, const R3Vector& v2)
{
    return fabs(k1);
}

static double
K2Abs(double k1, double k2,
      const R3Vector& v1, const R3Vector& v2)
{
    return fabs(k2);
}

static double
K1MinusK2Abs(double k1, double k2,
             const R3Vector& v1, const R3Vector& v2)
{
    return fabs(k1-k2);
}

static double
K1OrderedAbs(double k1, double k2,
             const R3Vector& v1, const R3Vector& v2)
{
    if ( fabs(k1) < fabs(k2) ) std::swap( k1, k2 );
    return fabs(k1);
}

static double
K2OrderedAbs(double k1, double k2,
             const R3Vector& v1, const R3Vector& v2)
{
    if ( fabs(k1) < fabs(k2) ) std::swap( k1, k2 );
    return fabs(k2);
}

static double
K1MinusK2OrderedAbs(double k1, double k2,
                    const R3Vector& v1, const R3Vector& v2)
{
    if ( fabs(k1) < fabs(k2) ) std::swap( k1, k2 );
    return fabs(k1-k2);
}

static double
K1K2AverageAbs(double k1, double k2,
               const R3Vector& v1, const R3Vector& v2)
{
    return fabs((k1+k2)*0.5);
}

static double
K1K2ProductAbs(double k1, double k2,
               const R3Vector& v1, const R3Vector& v2)
{
    return fabs(k1*k2);
}

static double
PrescribedDirection(double k1, double k2,
                    const R3Vector& v1, const R3Vector& v2)
{
    R3Vector direction = MyConfigParser.prescribed_direction;
    double c1 = v1.Dot(direction);
    double c2 = v2.Dot(direction);
    return k1*c1*c1 + k2*c2*c2;
}

static funcPtr
GetFeatureFunc(const string& funcName)
{
    if (funcName == "K1") return &K1;
    else if (funcName == "K2") return &K2;
    else if (funcName == "K1MinusK2") return &K1MinusK2;
    else if (funcName == "K1Ordered") return &K1Ordered;
    else if (funcName == "K2Ordered") return &K2Ordered;
    else if (funcName == "K1MinusK2Ordered") return &K1MinusK2Ordered;
    else if (funcName == "K1K2Average") return &K1K2Average;
    else if (funcName == "K1K2Product") return &K1K2Product;
    else if (funcName == "K1Abs") return &K1Abs;
    else if (funcName == "K2Abs") return &K2Abs;
    else if (funcName == "K1MinusK2Abs") return &K1MinusK2Abs;
    else if (funcName == "K1OrderedAbs") return &K1OrderedAbs;
    else if (funcName == "K2OrderedAbs") return &K2OrderedAbs;
    else if (funcName == "K1MinusK2OrderedAbs") return &K1MinusK2OrderedAbs;
    else if (funcName == "K1K2AverageAbs") return &K1K2AverageAbs;
    else if (funcName == "K1K2ProductAbs") return &K1K2ProductAbs;
    else if (funcName == "PrescribedDirection") return &PrescribedDirection;
    else {
        fprintf(stderr, "CurvatureHistogram::GetFeatureFunc unable to find function %s!\n", funcName.c_str());
        exit(-1);
    }
}

static void
WriteAdjacentSampleIndexArray(const vector< vector<int> >& adjacentArray, const string& filename)
{
    ofstream fout(filename.c_str());
    for (int i=0; i<(int)adjacentArray.size(); i++) {
        fout<<i<<" ";
        for (int j=0; j<(int)adjacentArray[i].size(); j++) {
            fout<<adjacentArray[i][j]<<" ";
        }
        fout<<std::endl;
    }
    fout.close();
}

CurvatureExporter::
CurvatureExporter(PointCloud* pCloud)
:GeodesicComputer(pCloud)
{}

void CurvatureExporter::
Run(void)
{
    // construct connected graph and compute geodesic distance
    if (!m_bIsDistanceLoaded) {
        ConstructConnectedGraphAndComputeGeodesicDistances();
    }
    
    string fittingScheme = MyConfigParser.Get("CurvatureHistogram", "fitting_scheme");
    if (fittingScheme == "cubic") {
        RunCubic();
    }
    
    else if (fittingScheme == "quadratic") {
        RunQuadratic();
    }
}

// quadratic
void CurvatureExporter::
RunQuadratic(void)
{
    printf("CurvatureExporter::RunQuadratic processing %d samples ...\n", NSamples());
    // curvature neighborhood, normal different cut-off
    double normalThreshold = MyConfigParser.GetScalar("CurvatureHistogram", "normal_threshold");
    // see PointFeatureExporter::exportCurvatureFeaturesPerPoint
    // options of scales
    double epsilon = MyConfigParser.GetScalar("CurvatureHistogram", "epsilon");
    vector<double> scaleArray = MyConfigParser.GetScalarArray("CurvatureHistogram", "scale_array");
    vector<string> featureArray = MyConfigParser.GetStringArray("CurvatureHistogram", "feature_array_per_scale");
    int numScales = (int)scaleArray.size();
    int numFeaturesPerScale = (int)featureArray.size();
    vector<funcPtr> featureFuncArray(numFeaturesPerScale, NULL);
    // mapping from feature index to feature func
    for (int i=0; i<numFeaturesPerScale; i++)
        featureFuncArray[i] = GetFeatureFunc(featureArray[i]);
    // delete m_pPropertySet;
    m_pPropertySet = new PointCloudPropertySet(m_pPointCloud);
    for (int i=0; i<numScales; i++) {
        for (int j=0; j<numFeaturesPerScale; j++) {
            string name = string("Scale(")+StringUtil::Number2String(scaleArray[i])+")-"+featureArray[j];
            m_pPropertySet->Insert(new PointCloudProperty(m_pPointCloud, name));
        }
    }
    // auxillary
    vector<int> adjacentSampleIndexArray;
    adjacentSampleIndexArray.reserve(NSamples());
    vector<double> adjacentGeodesicArray;
    adjacentGeodesicArray.reserve(NSamples());
    
    // outer loop: scales
    for (int scaleIdx=0; scaleIdx<numScales; scaleIdx++) {
        // scale is the absolute length (average max geo distance ~ 5)
        double scale = scaleArray[scaleIdx]/**globalMaxGeodDistance*/;
        printf("ShapeDescriptors::CurvatureHistogram computing curvatures at scale(%g) for mesh ...\n",
               scaleArray[scaleIdx]);
        // vector< vector<int> > adjacentSampleIndexArrayArray;
        // inner loop: samples
//        PointCloud cloudWithPrinciple;
        for (int sampleIdx=0; sampleIdx<NSamples(); sampleIdx++) {
            adjacentSampleIndexArray.clear();
            adjacentGeodesicArray.clear();
            ComputeClosestSamples(adjacentSampleIndexArray, adjacentGeodesicArray, sampleIdx, scale*0.5, 10, normalThreshold);
            //// push to adjacentSampleIndexArrayArray
            // adjacentSampleIndexArrayArray.push_back(adjacentSampleIndexArray);
            // local coordinate frame
            R3Vector n = GetKthSample(sampleIdx).Normal();
            R3Vector t = ( GetKthSample(sampleIdx).Position() - GetKthSample(adjacentSampleIndexArray[0]).Position() ) % n;
            t.Normalize();
            R3Vector s = t % n;
            double C[9] = {t[0], t[1], t[2], s[0], s[1], s[2], n[0], n[1], n[2]};
            // call clapack to solve linear problem
            char NORM = '1';
            // number of rows (data points+normals)
            integer M = adjacentSampleIndexArray.size(); // quadratic surface fitting
            R3Vector principleDir[2];
            principleDir[0] = t; principleDir[1] = s;
            double k1 = 0, k2 = 0;
            if (int(adjacentSampleIndexArray.size()) < 3) {
                printf("Sample(%d) has only %d adjacent samples, where we are unable to compute curvatures!\n", sampleIdx, adjacentSampleIndexArray.size());
            }
            else {
                integer N = 3; // quadratic surface fitting
                integer NRHS = 1;
                integer *JPVT = new integer[N];
                for (int j=0; j<N; j++) JPVT[j] = 0;
                integer RANK;
                integer LWORK = 3 * M * N;
                double *WORK = new double[ LWORK ];
                double *B = new double[ M ];
                double *A = new double[ M*N ];
                integer INFO;
                for( int j = 0; j < (int)adjacentSampleIndexArray.size(); j++ )  // adjacent face 0 is the source face i
                {
                    R3Vector diff = GetKthSample(sampleIdx).Position() - GetKthSample(adjacentSampleIndexArray[j]).Position();
                    // position in the local frame
                    double x = C[0] * diff[0] + C[1] * diff[1] + C[2] * diff[2];
                    double y = C[3] * diff[0] + C[4] * diff[1] + C[5] * diff[2];
                    double z = C[6] * diff[0] + C[7] * diff[1] + C[8] * diff[2];
                    double norm = 1.0 / NSamples();
                    
                    A[      j   ] = 0.5f * x*x * norm;
                    A[  M + j   ] =        x*y * norm;
                    A[2*M + j   ] = 0.5f * y*y * norm;
                    B[      j	] = z * norm;
                    
                }
                
                double RCOND = dlange_( &NORM, &M, &N, A, &M, WORK );
                dgelsy_( &M, &N, &NRHS, A, &M, B, &M, JPVT, &RCOND, &RANK, WORK, &LWORK, &INFO );
                if (INFO != 0)
                {
                    std::cerr << std::endl << "Point " << sampleIdx << "possibly did not acquire right curvature values during dgelsy_ call! INFO = " << INFO << std::endl;
                }
                
                double trace = B[0] + B[2];
                double det = B[0] * B[2] - B[1] * B[1];
                k1 = .5f * ( trace + sqrt( trace*trace - 4.0f * det ) );
                k2 = .5f * ( trace - sqrt( trace*trace - 4.0f * det ) );
    //            // the principle direction
    //            double coeff_t = 0;
    //            if (fabs(k1) > fabs(k2)) {
    //                coeff_t = B[1]/(k1 - B[0]);
    //            }
    //            else {
    //                coeff_t = B[1]/(k2 - B[0]);
    //            }
    //            R3Vector principleDir = t*coeff_t + s;
    //            principleDir.Normalize();
    //            cloudWithPrinciple.AddVertex(GetKthSample(sampleIdx).Position(), principleDir);
                
                // the principle direction
                
                principleDir[0] = B[1]/(k1 - B[0])*t + s;
                principleDir[0].Normalize();
                
                principleDir[1] = B[1]/(k2 - B[0])*t + s;
                principleDir[1].Normalize();

                delete[] A;
                delete[] B;
                delete[] JPVT;
                delete[] WORK;
            }
            
            // go through all features
            for (int featureIdx=0; featureIdx<(int)featureArray.size(); featureIdx++) {
                double value = (*(featureFuncArray[featureIdx]))(k1, k2, principleDir[0], principleDir[1]);
                m_pPropertySet->Property(featureIdx+scaleIdx*numFeaturesPerScale)->SetVertexValue(sampleIdx, value);
            }
        }
//        cloudWithPrinciple.WriteFile("cloud_with_principle.pts");
    }
}

// We compute the curvatures on each point of the surface
// Two principle curvatures
// The curvature depends on the size of neighborhood under consideration,
// The size of neighbhorhood is specified in the config, relative to the max pairwise geodesic distance in the mesh
// The outerloop is about which part of the mesh is computed
// The middleloop is about the scales
// The innerloop is about features (k1 or k2)
// cubic surface fitting
void CurvatureExporter::
RunCubic(void)
{
    printf("CurvatureExporter::RunCubic processing %d samples ...\n", NSamples());
    // curvature neighborhood, normal different cut-off
    double normalThreshold = MyConfigParser.GetScalar("CurvatureHistogram", "normal_threshold");
    // see PointFeatureExporter::exportCurvatureFeaturesPerPoint
    // options of scales
    double epsilon = MyConfigParser.GetScalar("CurvatureHistogram", "epsilon");
    vector<double> scaleArray = MyConfigParser.GetScalarArray("CurvatureHistogram", "scale_array");
    vector<string> featureArray = MyConfigParser.GetStringArray("CurvatureHistogram", "feature_array_per_scale");
    int numScales = (int)scaleArray.size();
    int numFeaturesPerScale = (int)featureArray.size();
    vector<funcPtr> featureFuncArray(numFeaturesPerScale, NULL);
    // mapping from feature index to feature func
    for (int i=0; i<numFeaturesPerScale; i++)
        featureFuncArray[i] = GetFeatureFunc(featureArray[i]);
    // delete m_pPropertySet;
    m_pPropertySet = new PointCloudPropertySet(m_pPointCloud);
    for (int i=0; i<numScales; i++) {
        for (int j=0; j<numFeaturesPerScale; j++) {
            string name = string("Scale(")+StringUtil::Number2String(scaleArray[i])+")-"+featureArray[j];
            m_pPropertySet->Insert(new PointCloudProperty(m_pPointCloud, name));
        }
    }
    // auxillary
    vector<int> adjacentSampleIndexArray;
    adjacentSampleIndexArray.reserve(NSamples());
    vector<double> adjacentGeodesicArray;
    adjacentGeodesicArray.reserve(NSamples());
    
    // outer loop: scales
    for (int scaleIdx=0; scaleIdx<numScales; scaleIdx++) {
        // scale is the absolute length (average max geo distance ~ 5)
        double scale = scaleArray[scaleIdx]/**globalMaxGeodDistance*/;
        printf("ShapeDescriptors::CurvatureHistogram computing curvatures at scale(%g) for mesh ...\n",
               scaleArray[scaleIdx]);
        // PointCloud cloudWithPrinciple;
        // vector< vector<int> > adjacentSampleIndexArrayArray;
        // inner loop: samples
        for (int sampleIdx=0; sampleIdx<NSamples(); sampleIdx++) {
            adjacentSampleIndexArray.clear();
            adjacentGeodesicArray.clear();
            ComputeClosestSamples(adjacentSampleIndexArray, adjacentGeodesicArray, sampleIdx, scale*0.5, 10, normalThreshold);
            //// push to adjacentSampleIndexArrayArray
            // adjacentSampleIndexArrayArray.push_back(adjacentSampleIndexArray);
            // local coordinate frame
            R3Vector n = GetKthSample(sampleIdx).Normal();
            R3Vector t = ( GetKthSample(sampleIdx).Position() - GetKthSample(adjacentSampleIndexArray[0]).Position() ) % n;
            t.Normalize();
            R3Vector s = t % n;
            double C[9] = {t[0], t[1], t[2], s[0], s[1], s[2], n[0], n[1], n[2]};
            // call clapack to solve linear problem
            char NORM = '1';
            // number of rows (data points+normals)
            integer M = adjacentSampleIndexArray.size()*3; // cubic surface fitting
            R3Vector principleDir[2];
            principleDir[0] = t;
            principleDir[1] = s;
            double k1=0, k2=0;
            if (int(adjacentSampleIndexArray.size()) < 3) {
                printf("Sample(%d) has only %d adjacent samples, where we are unable to compute curvatures!\n", sampleIdx, adjacentSampleIndexArray.size());
            }
            else {
                integer N = 7; // cubic surface fitting
                integer NRHS = 1;
                integer *JPVT = new integer[N];
                for (int j=0; j<N; j++) JPVT[j] = 0;
                integer RANK;
                integer LWORK = 3 * M * N;
                double *WORK = new double[ LWORK ];
                double *B = new double[ M ];
                double *A = new double[ M*N ];
                integer INFO;
                for( int j = 0; j < (int)adjacentSampleIndexArray.size(); j++ )  // adjacent face 0 is the source face i
                {
                    R3Vector diff = GetKthSample(sampleIdx).Position() - GetKthSample(adjacentSampleIndexArray[j]).Position();
                    // position in the local frame
                    double x = C[0] * diff[0] + C[1] * diff[1] + C[2] * diff[2];
                    double y = C[3] * diff[0] + C[4] * diff[1] + C[5] * diff[2];
                    double z = C[6] * diff[0] + C[7] * diff[1] + C[8] * diff[2];
                    double norm = 1.0 / NSamples();
                    // normal in the local frame
                    R3Vector otherNormal = GetKthSample(adjacentSampleIndexArray[j]).Normal();
                    double xn = C[0] * otherNormal[0] + C[1] * otherNormal[1] + C[2] * otherNormal[2];
                    double yn = C[3] * otherNormal[0] + C[4] * otherNormal[1] + C[5] * otherNormal[2];
                    double zn = C[6] * otherNormal[0] + C[7] * otherNormal[1] + C[8] * otherNormal[2];
                    double u = -xn / (zn+epsilon);
                    double v = -yn / (zn+epsilon);
                    
                    A[      3*j   ] = 0.5f * x*x * norm; // only for cubic
                    A[  M + 3*j   ] =        x*y * norm; // only for cubic
                    A[2*M + 3*j   ] = 0.5f * y*y * norm; // only for cubic
                    A[3*M + 3*j   ] =      x*x*x * norm; // only for cubic
                    A[4*M + 3*j   ] =	   x*x*y * norm; // only for cubic
                    A[5*M + 3*j   ] =      x*y*y * norm; // only for cubic
                    A[6*M + 3*j   ] =      y*y*y * norm; // only for cubic
                    B[      3*j	  ] = z * norm;          // only for cubic
                    
                    A[      3*j+1 ] =          x * norm; // only for cubic
                    A[  M + 3*j+1 ] =          y * norm; // only for cubic
                    A[2*M + 3*j+1 ] = 0;                 // only for cubic
                    A[3*M + 3*j+1 ] = 3.0f * x*x * norm; // only for cubic
                    A[4*M + 3*j+1 ] = 2.0f * x*y * norm; // only for cubic
                    A[5*M + 3*j+1 ] =        y*y * norm; // only for cubic
                    A[6*M + 3*j+1 ] = 0;                 // only for cubic
                    B[      3*j+1 ] = u * norm;          // only for cubic
                    
                    A[      3*j+2 ] = 0;                 // only for cubic
                    A[  M + 3*j+2 ] =          x * norm; // only for cubic
                    A[2*M + 3*j+2 ] =          y * norm; // only for cubic
                    A[3*M + 3*j+2 ] = 0;                 // only for cubic
                    A[4*M + 3*j+2 ] =	     x*x * norm; // only for cubic
                    A[5*M + 3*j+2 ] = 2.0f * x*y * norm; // only for cubic
                    A[6*M + 3*j+2 ] = 3.0f * y*y * norm; // only for cubic
                    B[      3*j+2 ] = v * norm;          // only for cubic
                }
                
                double RCOND = dlange_( &NORM, &M, &N, A, &M, WORK );
                dgelsy_( &M, &N, &NRHS, A, &M, B, &M, JPVT, &RCOND, &RANK, WORK, &LWORK, &INFO );
                if (INFO != 0)
                {
                    std::cerr << std::endl << "Point " << sampleIdx << "possibly did not acquire right curvature values during dgelsy_ call! INFO = " << INFO << std::endl;
                }
                
                double trace = B[0] + B[2];
                double det = B[0] * B[2] - B[1] * B[1];
                k1 = .5f * ( trace + sqrt( trace*trace - 4.0f * det ) );
                k2 = .5f * ( trace - sqrt( trace*trace - 4.0f * det ) );
    //            // the principle direction
    //            double coeff_t = 0;
    //            if (fabs(k1) > fabs(k2)) {
    //                coeff_t = B[1]/(k1 - B[0]);
    //            }
    //            else {
    //                coeff_t = B[1]/(k2 - B[0]);
    //            }
    //            R3Vector principleDir = t*coeff_t + s;
    //            principleDir.Normalize();
    //            cloudWithPrinciple.AddVertex(GetKthSample(sampleIdx).Position(), principleDir);
                
                // the principle direction
                principleDir[0] = B[1]/(k1 - B[0])*t + s;
                principleDir[0].Normalize();
                
                principleDir[1] = B[1]/(k2 - B[0])*t + s;
                principleDir[1].Normalize();
                
                delete[] A;
                delete[] B;
                delete[] JPVT;
                delete[] WORK;
            }
            
            // go through all features
            for (int featureIdx=0; featureIdx<(int)featureArray.size(); featureIdx++) {
                double value = (*(featureFuncArray[featureIdx]))(k1, k2, principleDir[0], principleDir[1]);
                m_pPropertySet->Property(featureIdx+scaleIdx*numFeaturesPerScale)->SetVertexValue(sampleIdx, value);
            }
        }
        // cloudWithPrinciple.WriteFile("cloud_with_principle.pts");
    }
}

//const vector<double>& CurvatureExporter::
//Get(void) const
//{
//    // find the feature array with label = all
//    for (int i=0; i<(int)m_vFeaturePerLabelArray.size(); i++) {
//        if (m_vFeaturePerLabelArray[i].labelName == "all") {
//            return m_vFeaturePerLabelArray[i].descriptors;
//        }
//    }
//    fprintf(stderr, "CurvatureExporter::Get unable to find 'all'!\n");
//    exit(-1);
//}
//
//const vector<DescriptorsPerLabel>& CurvatureExporter::
//GetPerLabel(void) const
//{
//    return m_vFeaturePerLabelArray;
//}
//
//const vector<double> ShapeDescriptors::
//CurvatureHistogram(PointCloud* pCloud)
//{
//    CurvatureExporter exporter(pCloud);
//    exporter.Run();
//    return exporter.Get();
//}
//
//const vector<DescriptorsPerLabel> ShapeDescriptors::
//CurvatureHistogramPerLabel(PointCloud* pCloud)
//{
//    CurvatureExporter exporter(pCloud);
//    exporter.Run();
//    return exporter.GetPerLabel();
//}
