// Source file for the shape descriptor computation


////////////////////////////////////////////////////////////////////////
// Include files
////////////////////////////////////////////////////////////////////////

#include "Descriptor/ShapeDescriptors.h"

////////////////////////////////////////////////////////////////////////
// Utility functions
////////////////////////////////////////////////////////////////////////

static int
PCA(PointCloud* pCloud, R3CoordSystem *result_coordinate_system = NULL, RNScalar *result_variances = NULL)
{
    // Just checking
    if (pCloud->NPoints() == 0) return 0;
    
    // Get center
    R3Point centroid = pCloud->Centroid();
    
    // Compute 2D covariance matrix
    RNScalar m[4] = { 0 };
    RNCoord zmin = FLT_MAX;
    for (int i = 0; i < pCloud->NPoints(); i++) {
        PointCloudVertex* pVertex = pCloud->Vertex(i);
        const R3Point& p = pVertex->Position();
        RNScalar x = p[0] - centroid[0];
        RNScalar y = p[1] - centroid[1];
        if (p[2] < zmin) zmin = p[2];
        m[0] += x*x;
        m[1] += x*y;
        m[2] += x*y;
        m[3] += y*y;
    }
    
    // Normalize covariance matrix
    for (int i = 0; i < 4; i++) {
        m[i] /= pCloud->NPoints();
    }
    
    // Compute eigenvalues and eigenvectors
    RNScalar U[4];
    RNScalar W[2];
    RNScalar Vt[4];
    RNSvdDecompose(2, 2, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
    
    // Compute coordinate system
    if (result_coordinate_system) {
        // Extract principle axis direction from first eigenvector
        R3Vector xaxis(Vt[0], Vt[1], 0);
        
        // Flip principle axis so that "heavier" on positive side
        int positive_count = 0;
        int negative_count = 0;
        for (int i = 0; i < pCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = pCloud->Vertex(i);
            const R3Point& p = pVertex->Position();
            RNScalar x = p[0] - centroid[0];
            RNScalar y = p[1] - centroid[1];
            R3Vector vertex_vector(x, y, 0);
            RNScalar dot = xaxis.Dot(vertex_vector);
            if (dot > 0.0) positive_count++;
            else negative_count++;
        }
        if (positive_count < negative_count) {
            xaxis.Flip();
        }
        
        // Build coordinate system
        R3Vector yaxis(-xaxis.Y(), xaxis.X(), 0);
        R3Triad axes(xaxis, yaxis, R3posz_vector);
        result_coordinate_system->Reset(centroid, axes);
    }
    
    // Return variances
    if (result_variances) {
        // Set xy variances based on eigenvalues
        result_variances[0] = W[0];
        result_variances[1] = W[1];
        
        // Compute z variance
        result_variances[2] = 0;
        for (int i = 0; i < pCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = pCloud->Vertex(i);
            const R3Point& p = pVertex->Position();
            RNScalar dz = p.Z() - centroid.Z();
            result_variances[2] += dz * dz;
        }
        result_variances[2] /= pCloud->NPoints();
    }
    
    // Return success
    return 1;
}


////////////////////////////////////////////////////////////////////////
// Descriptor functions
////////////////////////////////////////////////////////////////////////

static void
InsertBBoxDescriptorValues(std::vector<double>& descriptor,
                           PointCloud* pCloud, double length_to_feature_scale)
{
    R3Box bbox = pCloud->BBox();
    // Initialize descriptor values
    RNScalar minz = -1;
    RNScalar maxz = -1;
    RNScalar lengthz = -1;
    RNScalar length1 = -1;
    RNScalar length2 = -1;
    RNScalar aspect21 = -1;
    RNScalar aspectz1 = -1;
    RNScalar aspect2z = -1;
    
    // Compute bounding box properties
    if (!bbox.IsEmpty()) {
        minz = length_to_feature_scale * bbox.ZMin();
        maxz = length_to_feature_scale * bbox.ZMax();
        RNScalar lengthx = length_to_feature_scale * bbox.XLength();
        RNScalar lengthy = length_to_feature_scale * bbox.YLength();
        lengthz = length_to_feature_scale * bbox.ZLength();
        length1 = (lengthx > lengthy) ? lengthx : lengthy;
        length2 = (lengthx < lengthy) ? lengthx : lengthy;
        aspect21 = (length1 > 0) ? length2 / length1 : 10;
        aspectz1 = (length1 > 0) ? lengthz / length1 : 10;
        aspect2z = (lengthz > 0) ? length2 / lengthz : 10;
        if (aspectz1 > 10) aspectz1 = 10;
        if (aspect2z > 10) aspect2z = 10;
    }
    
    // Insert descriptor values
    descriptor.push_back(minz);
    descriptor.push_back(maxz);
    descriptor.push_back(lengthz);
    descriptor.push_back(length1);
    descriptor.push_back(length2);
    descriptor.push_back(aspect21);
    descriptor.push_back(aspectz1);
    descriptor.push_back(aspect2z);
}



static void
InsertPCADescriptorValues(std::vector<double>& descriptor,
                          PointCloud* pCloud, double length_to_feature_scale)
{
    R3Box bbox = pCloud->BBox();
    // Initialize descriptor values
    RNScalar zcentroid = -1;
    RNScalar length1 = -1;
    RNScalar length2 = -1;
    RNScalar lengthz = -1;
    RNScalar aspect21 = -1;
    RNScalar aspectz1 = -1;
    RNScalar aspect2z = -1;
    
    // Compute PCA properties
    if (!bbox.IsEmpty()) {
        if (pCloud->NPoints() > 0) {
            R3CoordSystem cs;
            RNScalar variances[3];
            if (PCA(pCloud, &cs, variances)) {
                R3Point centroid = cs.Origin();
                R3Triad axes = cs.Axes();
                zcentroid = length_to_feature_scale * centroid.Z();
                length1 = length_to_feature_scale * sqrt(variances[0]);
                length2 = length_to_feature_scale * sqrt(variances[1]);
                lengthz = length_to_feature_scale * sqrt(variances[2]);
                aspect21 = (length1 > 0) ? length2 / length1 : 10;
                aspectz1 = (length1 > 0) ? lengthz / length1 : 10;
                aspect2z = (lengthz > 0) ? length2 / lengthz : 10;
                if (aspectz1 > 10) aspectz1 = 10;
                if (aspect2z > 10) aspect2z = 10;
            }
        }
    }
    
    // Insert descriptor values
    descriptor.push_back(zcentroid);
    descriptor.push_back(length1);
    descriptor.push_back(length2);
    descriptor.push_back(lengthz);
    descriptor.push_back(aspect21);
    descriptor.push_back(aspectz1);
    descriptor.push_back(aspect2z);
}



static void
InsertEGIDescriptorValues(std::vector<double>& descriptor,
                          PointCloud* pCloud, double length_to_feature_scale)
{
    // Initialize values
    RNScalar values[2] = { -1, -1 };
    
    // Check points
    if (pCloud->NPoints() > 0) {
        // Zero values
        values[0] = 0;
        values[1] = 0;
        
        // Compute values
        for (int j = 0; j < pCloud->NPoints(); j++) {
            PointCloudVertex* pVertex = pCloud->Vertex(j);
            R3Vector normal = pVertex->Normal();
            int dim = normal.MaxDimension();
            if (dim == RN_Z) values[1]++;
            else values[0]++;
        }
        
        // Normalize values
        RNScalar total = values[0] + values[1];
        if (total > 0) {
            values[0] /= total;
            values[1] /= total;
        }
    }
    
    // Insert descriptor values
    descriptor.push_back(values[0]);
    descriptor.push_back(values[1]);
}



static void
InsertZSliceDescriptorValues(std::vector<double>& descriptor,
                             PointCloud* pCloud,
                             double length_to_feature_scale, int zslice_count)
{
    R3Box bbox = pCloud->BBox();
    // Initialize zslice values
    if (zslice_count == 0) return;
    RNScalar *zslice_values = new RNScalar [ zslice_count ]; // alright
    for (int j = 0; j < zslice_count; j++) zslice_values[j] = -1;
    
    // Check bounding box
    if (!bbox.IsEmpty()) {
        if (pCloud->NPoints() > 0) {
            // Get zslice parameters
            double zslice_height = bbox.ZLength() / zslice_count;
            if (RNIsZero(zslice_height)) zslice_height = RN_EPSILON * zslice_count;
            double zslice_min = bbox.ZMin();
            
            // Zero values
            for (int j = 0; j < zslice_count; j++) {
                zslice_values[j] = 0;
            }
            
            // Compute values
            for (int j = 0; j < pCloud->NPoints(); j++) {
                PointCloudVertex* pVertex = pCloud->Vertex(j);
                R3Point position = pVertex->Position();
                double zslice_bin = (position.Z() - zslice_min) / zslice_height;
                int bin0 = (int) zslice_bin;
                int bin1 = bin0 + 1;
                double t = zslice_bin - bin0;
                if (bin0 < 0) bin0 = 0;
                if (bin1 < 0) bin1 = 0;
                if (bin0 > zslice_count-1) bin0 = zslice_count-1;
                if (bin1 > zslice_count-1) bin1 = zslice_count-1;
                zslice_values[bin0] += 1.0 - t;
                zslice_values[bin1] += t;
            }
            
            // Normalize values
            for (int j = 0; j < zslice_count; j++) {
                zslice_values[j] /= pCloud->NPoints();
            }
        }
    }
    
    // Insert descriptor values
    for (int j = 0; j < zslice_count; j++) {
        descriptor.push_back(zslice_values[j]);
    }
    
    // Delete zslice values
    delete [] zslice_values;
}



static void
InsertD2DescriptorValues(std::vector<double>& descriptor,
                         PointCloud* pCloud,
                         double length_to_feature_scale, int nbins = 8)
{
    R3Box bbox = pCloud->BBox();
    // Get undersampling factor
    int undersampling_factor = pCloud->NPoints() / 1024;
    if (undersampling_factor == 0) undersampling_factor = 1;
    
    // Initialize bins
    if (nbins <= 0) return;
    double *bins = new double [ nbins ]; // alright
    for (int i = 0; i < nbins; i++) bins[i] = -1;
    
    // Check bounding box
    if (!bbox.IsEmpty()) {
        if (pCloud->NPoints() > 0) {
            // Zero bins
            for (int i = 0; i < nbins; i++) bins[i] = 0;
            
            // Get distance samples
            int nsamples = 0;
            int max_samples = pCloud->NPoints() * (pCloud->NPoints() - 1) / 2;;
            RNLength *samples = new RNLength [ max_samples ]; // alright
            for (int i = 0; i < pCloud->NPoints(); i += undersampling_factor) {
                PointCloudVertex* pVertex0 = pCloud->Vertex(i);
                for (int j = i + undersampling_factor; j < pCloud->NPoints(); j += undersampling_factor) {
                    PointCloudVertex* pVertex1 = pCloud->Vertex(j);
                    if (nsamples >= max_samples) break;
                    RNLength d = R3Distance(pVertex0->Position(), pVertex1->Position());
                    samples[nsamples++] = d;
                }
            }
            
            // Compute scale from distance to bins
            RNScalar sum = 0;
            for (int i = 0; i < nsamples; i++) sum += samples[i];
            double max_distance = 3.0 * sum / nsamples;
            double xscale = (nbins-1) / max_distance;
            
            // Fill bins
            for (int i = 0; i < nsamples; i++) {
                RNScalar x = xscale * samples[i];
                if (x >= nbins - 1) x = nbins - 1 - RN_EPSILON;
                else if (x < 0) x = 0;
                int ix = (int) x;
                assert(ix < nbins);
                RNScalar dx = x - ix;
                bins[ix] += 1 - dx;
                bins[ix+1] += dx;
            }
            
            // Normalize bins (area under curve is 10);
            RNScalar yscale = (double) nbins / (double) nsamples;
            for (int i = 0; i < nbins; i++) {
                bins[i] = yscale * bins[i];
            }
            
            // Delete distance samples
            delete [] samples;
        }
    }
    
    // Insert descriptor values
    for (int j = 0; j < nbins; j++) {
        descriptor.push_back(bins[j]);
    }
    
    // Delete bins
    delete [] bins;
}



////////////////////////////////////////////////////////////////////////
// Top level function
////////////////////////////////////////////////////////////////////////

const std::vector<double> ShapeDescriptors::
D2(PointCloud* pCloud, int z_slices, int d2_nbins,
   double length_to_feature_scale)
{
    // Create descriptor
    std::vector<double> descriptor;
    InsertBBoxDescriptorValues(descriptor, pCloud, length_to_feature_scale);
    InsertPCADescriptorValues(descriptor, pCloud, length_to_feature_scale);
    InsertEGIDescriptorValues(descriptor, pCloud, length_to_feature_scale);
    InsertZSliceDescriptorValues(descriptor, pCloud, length_to_feature_scale, z_slices);
    InsertD2DescriptorValues(descriptor, pCloud, length_to_feature_scale, d2_nbins);
    // Return descriptor
    return descriptor;
}



