#include <string>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
#include "RNBasics/RNBasics.h"

using namespace std;
using namespace ShapeDescriptors;

GeodesicComputer::
GeodesicComputer(PointCloud* pCloud)
:DescriptorExporter(NULL, pCloud),
m_bIsDistanceLoaded(false)
{}

// construct a connected graph based on nearest neighbors
// output: m_vAdjacentIndexList
void GeodesicComputer::
ConstructConnectedGraphAndComputeGeodesicDistances(void)
{
    int numNearestNeighbors = MyConfigParser.GetInt("CurvatureHistogram", "starting_number_nearest_neighbors");
    m_vAdjacentIndexList.clear();
    // PointFeatureExporter::generateSamples L547
    printf("GeodesicComputer::ConstructConnectedGraphAndComputeGeodesicDistances numNearestNeighbors = %d ...\n", numNearestNeighbors);
    fflush(stdout);
    while (true) {
        // find the samples that are cloest according to euclidean distances
        m_vAdjacentIndexList = vector< vector<int> >(NSamples());
        printf("    compute adjacent points ... \n");
        FileSystemUtil::DrawProgressBar(0.0);
        RNTime time;
        time.Read();
        for (int i=0; i<NSamples(); i++) {
            // list all euclidean distances
            vector<double> euclideanDistances(NSamples(), 0);
            for (int j=0; j<NSamples(); j++) {
                euclideanDistances[j] = R3Distance(GetKthSample(i).Position(), GetKthSample(j).Position());
            }
            // pick k nearest neighbors
            euclideanDistances[i] = FLT_MAX;
            m_vAdjacentIndexList[i] = MathUtil::KthMinimum(euclideanDistances, numNearestNeighbors);
            FileSystemUtil::DrawProgressBar(double(i+1)/NSamples());
        }
        printf("    %.3f s\n", time.Elapsed()); fflush(stdout);
        printf("    compute geodesic distances ... \n");
        time.Read();
        ComputeGeodesicDistances();
        printf("    %.3f s\n", time.Elapsed()); fflush(stdout);
        break;
        //if (GetMaxGeodesicDistance() < FLT_MAX) break;
        //numNearestNeighbors *= 2;
        //printf("    the graph is disconnected, resetting numNearestNeighbors = %d\n", numNearestNeighbors); fflush(stdout);
        // clean up
        //m_vGeodClosestIndexList.clear();
    }
    printf("GeodesicComputer::ConstructConnectedGraphAndComputeGeodesicDistances Max Geodesic Distance = %g\n", GetMaxGeodesicDistance());
    m_bIsDistanceLoaded = true;
}

// compute geodesic distance given m_vAdjacentIndexList
// output: m_vGeodDistanceTable, m_vGeodClosestIndexList
void GeodesicComputer::
ComputeGeodesicDistances(void)
{
    // PointFeatureExporter::computeGeodesicDistances
    m_vGeodDistanceTable = vector< vector<double> >(NSamples());
    m_vGeodClosestIndexList = vector< vector<int> >(NSamples());
    FileSystemUtil::DrawProgressBar(0.0);
    HeapData tempHeapData;
    HeapData* heapDataArray = new HeapData[NSamples()];
    for (int i=0; i<NSamples(); i++) {
        vector<bool> visitedFlags(NSamples(), false);
        // initialize
        RNHeap<HeapData*> minHeap2(&tempHeapData, &(tempHeapData.value), &(tempHeapData.pointer));
        for (int j=0; j<NSamples(); j++) {
            heapDataArray[j].value = FLT_MAX;
            heapDataArray[j].pointer = NULL;
            heapDataArray[j].sampleIndex = j;
            minHeap2.Push(&heapDataArray[j]);
        }
        m_vGeodDistanceTable[i] = vector<double>(NSamples(), FLT_MAX);
        // initial setting
        heapDataArray[i].value = 0;
        minHeap2.Update(&heapDataArray[i]);
        // Djistra algorithm
        for (int j=0; j<NSamples(); j++) {
            HeapData* heapPtr2 = minHeap2.Peek();
            int index2 = heapPtr2->sampleIndex;
            double minUnvisitedDistHeap2 = heapPtr2->value;
            if (minUnvisitedDistHeap2 >= FLT_MAX || visitedFlags[index2]) break;
            // put index into known group
            visitedFlags[index2] = true;
            m_vGeodDistanceTable[i][index2] = heapPtr2->value;
            // set as infinite
            heapPtr2->value = FLT_MAX;
            minHeap2.Update(heapPtr2);
            // update neighbors of index
            for (int k=0; k<(int)m_vAdjacentIndexList[index2].size(); k++) {
                int indexOfNeighbor = m_vAdjacentIndexList[index2][k];
                if (visitedFlags[indexOfNeighbor]) continue;
                double newDistance = m_vGeodDistanceTable[i][index2] + R3Distance(GetKthSample(index2).Position(), GetKthSample(indexOfNeighbor).Position());
                if (newDistance < heapDataArray[indexOfNeighbor].value) {
                    heapDataArray[indexOfNeighbor].value = newDistance;
                    assert (heapDataArray[indexOfNeighbor].sampleIndex == indexOfNeighbor);
                    minHeap2.Update(&heapDataArray[indexOfNeighbor]);
                }
            }
        }
        // update m_vGeodClosestIndexList[i]
        MathUtil::Sort(m_vGeodDistanceTable[i], m_vGeodClosestIndexList[i]);
        FileSystemUtil::DrawProgressBar(double(i+1)/NSamples());
    }
    delete[] heapDataArray;
}

// return a bunch of samples that are
// (1) close spatially
// (2) having similar orientation
void GeodesicComputer::
ComputeClosestSamples(vector<int>& adjacentSampleIndexArray, vector<double>& adjacentGeodDistanceArray,
                      int startSampleIndex, double maxGeodDistance, int minNumAdjacentSamples, double normalThreshold)
{
    const PointCloudVertex& pointSample = GetKthSample(startSampleIndex);
    for (int i=1; i<NSamples(); i++) {
        int otherSampleIndex = GetKthClosestSampleIndex(startSampleIndex, i);
        const PointCloudVertex& otherSample = GetKthSample(otherSampleIndex);
        // has to add more
        if ((int)adjacentGeodDistanceArray.size() <= minNumAdjacentSamples) {
            // if normal meets requirement
            if (pointSample.Normal().Dot(otherSample.Normal()) > normalThreshold) {
                adjacentSampleIndexArray.push_back(otherSampleIndex);
                adjacentGeodDistanceArray.push_back(GetGeodesicDistanceByIndexPair(startSampleIndex, otherSampleIndex));
            }
        }
        // don't have to add
        else {
            if (GetGeodesicDistanceByIndexPair(startSampleIndex, otherSampleIndex) > maxGeodDistance) break;
            else if (pointSample.Normal().Dot(otherSample.Normal()) > normalThreshold) {
                adjacentSampleIndexArray.push_back(otherSampleIndex);
                adjacentGeodDistanceArray.push_back(GetGeodesicDistanceByIndexPair(startSampleIndex, otherSampleIndex));
            }
        }
    }
}

// quick reference
int GeodesicComputer::
NSamples(void) const
{
    return m_pPointCloud->NPoints();
}

const PointCloudVertex& GeodesicComputer::
GetKthSample(int k) const
{
    assert(k>=0 && k<NSamples());
    return *(m_pPointCloud->Vertex(k));
}

double GeodesicComputer::
GetGeodesicDistanceByIndexPair(int index1, int index2) const
{
    assert(index1>=0 && index1<(int)m_vGeodDistanceTable.size());
    assert(index2>=0 && index2<=(int)m_vGeodDistanceTable[index1].size());
    return m_vGeodDistanceTable[index1][index2];
}

int GeodesicComputer::
GetKthClosestSampleIndex(int startSampleIndex, int k) const
{
    assert(startSampleIndex>=0 && startSampleIndex<(int)m_vGeodClosestIndexList.size());
    assert(k>=0 && k<(int)m_vGeodClosestIndexList[startSampleIndex].size());
    return m_vGeodClosestIndexList[startSampleIndex][k];
}

double GeodesicComputer::
GetMaxGeodesicDistance(void) const
{
    // only need to check the largest distance associated to each starting sample
    double maxDistance = -FLT_MAX;
    for (int i=0; i<NSamples(); i++) {
        int index = GetKthClosestSampleIndex(i, NSamples()-1);
        double distance = GetGeodesicDistanceByIndexPair(i, index);
        if (distance > maxDistance) maxDistance = distance;
    }
    return maxDistance;
}