#ifndef __SHAPE_DESCRIPTORS_H
#define __SHAPE_DESCRIPTORS_H
#include <vector>
#include <string>
#include "R3Shapes/R3Shapes.h"
#include "R3MyGraphics/R3MyGraphics.h"
#include "Utils/Utils.h"
namespace ShapeDescriptors
{
    // compute property set from mesh (mesh is transformed with segmentation)
    // PointCloudPropertySet is associated with a PointCloud, which contains segmentation
    std::vector<PointCloudPropertySet*> ComputeShapeDescriptor(R3Mesh* pMesh, PointCloud* pCloud,
                                                               const std::vector<std::string>& descriptorNameArray);
    std::vector<PointCloudPropertySet*> ComputeShapeDescriptorPerSegment(R3Mesh* pMesh,
                                                                         PointCloud* pCloud,
                                                                         PointCloudPropertySet* pSegmentation,
                                                                         const std::vector<std::string>& descriptorNameArray);
    void SplitPointCloud(PointCloudPropertySet* pSegmentation, std::vector<int>& segIdxArray, std::vector<int>& vtxIdxArray,
                         std::vector<PointCloud*>& pointCloudArray);
    struct HeapData
    {
        double value;
        HeapData** pointer;
        int sampleIndex;
    };
    
    // Function for public
    /*Base class for descriptor*/
    class DescriptorExporter
    {
    public:
        DescriptorExporter(R3Mesh* pMesh, PointCloud* pCloud);
        PointCloudPropertySet* Get(void);
    protected:
        virtual void Run(void) = 0;
    protected:
        R3Mesh* m_pMesh;
        PointCloud* m_pPointCloud;
        PointCloudPropertySet* m_pPropertySet;
    };
    
    /*Base class for curvature and spectrum*/
    class GeodesicComputer : public DescriptorExporter
    {
    public:
        GeodesicComputer(PointCloud* pCloud);
    protected:
        // construct a connected graph from sampleArray using proximity
        // (so that geodesic distance can be defined upon)
        void ConstructConnectedGraphAndComputeGeodesicDistances(void);
        // order neighbors by geodesic distance
        // (PointFeatureExporter::computeGeodesicDistances)
        void ComputeGeodesicDistances(void);
        // (PointFeatureExporter::getClosestSamples)
        void ComputeClosestSamples(std::vector<int>& adjacentSampleIndexArray,
                                   std::vector<double>& adjacentGeodDistanceArray,
                                   int startSampleIndex, double maxGeodDistance,
                                   int minNumAdjacentSamples, double normalThreshold);
        // quick reference
        int NSamples(void) const;
        const PointCloudVertex& GetKthSample(int k) const;
        int GetKthClosestSampleIndex(int startSampleIndex, int k) const;
        double GetGeodesicDistanceByIndexPair(int index1, int index2) const;
        double GetMaxGeodesicDistance(void) const;
    protected:
        std::vector< std::vector<int> > m_vAdjacentIndexList;
        std::vector< std::vector<double> > m_vGeodDistanceTable;
        std::vector< std::vector<int> > m_vGeodClosestIndexList;
        // std::vector<DescriptorsPerLabel> m_vFeaturePerLabelArray;
        bool m_bIsDistanceLoaded;
    };
    
    /*Curvature histogram*/
    class CurvatureExporter : public GeodesicComputer
    {
    public:
        CurvatureExporter(PointCloud* pCloud);
    protected:
        virtual void Run(void);
        void RunQuadratic(void);
        void RunCubic(void);
    };
//    
//    /*Laplacian spectrum*/
//    class LaplacianSpectrumExporter : public GeodesicComputer
//    {
//    public:
//        LaplacianSpectrumExporter(PointCloud* pCloud);
//    protected:
//        virtual void Run(void);
//    protected:
//        void GenerateLaplacianMatrix(PointCloudProperty* mask, MathUtil::SparseMatrix& sparseMat);
//    };
//    const std::vector<double> LaplacianSpectrum(PointCloud* pCloud);
//    const std::vector<DescriptorsPerLabel> LaplacianSpectrumPerLabel(PointCloud* pCloud);
    
    /*Shape diameter function*/
    class ShapeDiameterExporter : public DescriptorExporter
    {
    public:
        ShapeDiameterExporter(R3Mesh* pMesh, PointCloud* pCloud);
    protected:
        virtual void Run(void);
        //std::vector<DescriptorsPerLabel> m_vFeaturePerLabelArray;
    };
}
#endif
