#include <string>
#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
using namespace std;
using namespace ShapeDescriptors;

#define sind(x) (sin(fmod((x),360) * M_PI / 180))
#define cosd(x) (cos(fmod((x),360) * M_PI / 180))

ShapeDiameterExporter::
ShapeDiameterExporter(R3Mesh* pMesh, PointCloud* pCloud)
:DescriptorExporter(pMesh, pCloud)
{}

// We compute the SDF for each point on the surface
// First, the SDF value is normalized by the size of the model's bbox
// This doesn't influence the shape descriptors that just depend on the histogram
// But do make the values comparable across models that are not in the same scale
// Additional to the SDF values (SDF-Original), we also add additional dimensions where values are logarithmized,
// Details can be found in the paper "Contextual Part Analogies in 3D objects"
void ShapeDiameterExporter::
Run(void)
{
    printf("ShapeDiameterExporter::Run processing %d samples with %d faces ...\n",
           m_pPointCloud->NPoints(), m_pMesh->NFaces());
    double phiMax = MyConfigParser.GetScalar("ShapeDiameterHistogram", "sampling_phi_max");
    double phiStep = MyConfigParser.GetScalar("ShapeDiameterHistogram", "sampling_phi_step");
    double thetaStep = MyConfigParser.GetScalar("ShapeDiameterHistogram", "sampling_theta_step");
    double epsilon = MyConfigParser.GetScalar("ShapeDiameterHistogram", "epsilon");
    // Reference: PointFeatureExporter::exportSDFFeaturesPerPoint
    // all properties
    // delete m_pPropertySet;
    m_pPropertySet = new PointCloudPropertySet(m_pPointCloud);
    string name = string("SDF-Original");
    m_pPropertySet->Insert(new PointCloudProperty(m_pPointCloud, name));
    // fill SDF-Original, the first property
    printf("Compute basic SDF ...\n");
    FileSystemUtil::DrawProgressBar(0.0);
    for (int i=0; i<m_pPointCloud->NPoints(); i++) {
        vector<double> sdfSamples;
        PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
        R3Vector rayCenterDir = -pVertex->Normal();
        R3Point rayOrigin = pVertex->Position() + rayCenterDir*epsilon;
        // create a triad (zaxis = rayCenterDir)
        R3Triad triad = MathUtil::GenerateTriad(rayCenterDir);
        R3Vector u = triad.Axis(RN_X);
        R3Vector v = triad.Axis(RN_Y);
//        printf("center[%g %g %g](norm = %g) u[%g %g %g](norm = %g) v[%g %g %g](norm = %g)\n",
//               rayCenterDir[0], rayCenterDir[1], rayCenterDir[2],
//               rayCenterDir.Length(),
//               u[0], u[1], u[2], u.Length(),
//               v[0], v[1], v[2], v.Length());
//        getchar();
        for (double phi=0; phi<=phiMax; phi+=phiStep) {
            for (double theta=0; theta<360.0; theta+=thetaStep) {
                R3Vector rayDir = rayCenterDir*cosd(phi) + u*sind(phi)*cosd(theta) + v*sind(phi)*sind(theta);
                rayDir.Normalize();
                R3Ray ray(rayOrigin, rayDir);
                // TODO: can be optimized by building kd-tree
                R3MeshIntersection intersection;
                R3MeshType type = m_pMesh->Intersection(ray, &intersection);
                if (type != R3_MESH_NULL_TYPE) {
                    // we normalize by the size of model, in order to have a value comparable across models
                    double rayLength = intersection.t;
                    if (rayLength > 0.0) {
                        sdfSamples.push_back(rayLength);
                    }
                    else {
                        sdfSamples.push_back(-1.0);
                    }
                }
            }
        }
        m_pPropertySet->Property(0)->SetVertexValue(i, MathUtil::Median(sdfSamples));
        FileSystemUtil::DrawProgressBar(double(i+1)/m_pPointCloud->NPoints());
    }
    printf("\n");
//    // remove outlier
//    m_pPropertySet->Property(0)->RemoveOutliers(0.95);
//    // normalize
//    m_pPropertySet->Property(0)->Normalize(); // scale to [0, 1]
//    // logarithmize, and fill the remaining properties
//    printf("Compute logarithmized SDFs ...\n");
//    for (int i=0; i<(int)logBaseArray.size(); i++) {
//        string name = string("SDF-")+StringUtil::Number2String(logBaseArray[i]);
//        m_pPropertySet->Insert(new PointCloudProperty(m_pPointCloud, name));
//    }
//
//    FileSystemUtil::DrawProgressBar(0.0);
//    for (int baseIdx=0; baseIdx<(int)logBaseArray.size(); baseIdx++) {
//        double base = logBaseArray[baseIdx];
//        for (int sampleIdx=0; sampleIdx<m_pPointCloud->NPoints(); sampleIdx++) {
//            double rawVal = m_pPropertySet->Property(0)->VertexValue(sampleIdx);
//            double val = log(rawVal*base+1.0)/log(base+1.0);
//            m_pPropertySet->Property(baseIdx+1)->SetVertexValue(sampleIdx, val);
//        }
//        FileSystemUtil::DrawProgressBar(double(baseIdx+1)/logBaseArray.size());
//    }
//    printf("\n");
//    // compute features, which are histograms of each dimensions
//    // see FeatureMesh::compute_features
//    m_vFeaturePerLabelArray.clear();
//    vector<int> numBinArray = MyConfigParser.GetIntArray("ShapeDiameterHistogram", "number_histogram_bins");
//    // create masks
//    PointCloudPropertySet maskSet(m_pPointCloud);
//    maskSet.CreateSegmentMasks();
//    for (int k=0; k<maskSet.NProperties(); k++) {
//        DescriptorsPerLabel descPerLabel;
//        descPerLabel.labelName = maskSet.Property(k)->Name();
//        // different features
//        for (int i=0; i<m_pPropertySet->NProperties(); i++) {
//            // different numbers of bins
//            for (int j=0; j<(int)numBinArray.size(); j++) {
//                vector<double> features = m_pPropertySet->Property(i)->HistogramGivenMask(numBinArray[j], 0, 1.0, maskSet.Property(k));
//                
//                descPerLabel.descriptors.insert(descPerLabel.descriptors.end(),
//                                                features.begin(), features.end());
//            }
//        }
//        m_vFeaturePerLabelArray.push_back(descPerLabel);
//    }
}

//const vector<double>& ShapeDiameterExporter::
//Get(void) const
//{
//    // find the feature array with label = all
//    for (int i=0; i<(int)m_vFeaturePerLabelArray.size(); i++) {
//        if (m_vFeaturePerLabelArray[i].labelName == "all") {
//            return m_vFeaturePerLabelArray[i].descriptors;
//        }
//    }
//    fprintf(stderr, "ShapeDiameterExporter::Get unable to find 'all'!\n");
//    exit(-1);
//}
//
//const vector<DescriptorsPerLabel>& ShapeDiameterExporter::
//GetPerLabel(void) const
//{
//    return m_vFeaturePerLabelArray;
//}
//
//const vector<double> ShapeDescriptors::
//ShapeDiameterHistogram(R3Mesh* pMesh, PointCloud* pCloud)
//{
//    ShapeDiameterExporter exporter(pMesh, pCloud);
//    exporter.Run();
//    return exporter.Get();
//}
//
//const vector<DescriptorsPerLabel> ShapeDescriptors::
//ShapeDiameterHistogramPerLabel(R3Mesh* pMesh, PointCloud* pCloud)
//{
//    ShapeDiameterExporter exporter(pMesh, pCloud);
//    exporter.Run();
//    return exporter.GetPerLabel();
//}