#include "R3MyGraphics/R3MyGraphics.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
using namespace std;

DeformableCircle::
DeformableCircle(int numSamples)
:m_aNormals(numSamples),radius(1.0), center(0, 0, 0)
{
    for (int i=0; i<numSamples; i++) {
        double angle = 2.0*RN_PI*i/numSamples;
        m_aNormals[i] = R3Vector(0, cos(angle), sin(angle));
    }
}

void DeformableCircle::
SetRadius(double _radius)
{
    radius = _radius;
}

void DeformableCircle::
SetCenter(const R3Point& pos)
{
    center = pos;
}

int DeformableCircle::
NSamples(void) const
{
    return (int)m_aNormals.size();
}

const R3Vector DeformableCircle::
Normal(int idx) const
{
    assert(idx>=0 && idx<NSamples());
    return m_aNormals[idx];
}

const R3Point DeformableCircle::
Position(int idx) const
{
    assert(idx>=0 && idx<NSamples());
    return center + m_aNormals[idx]*radius;
}

const R3Point& DeformableCircle::
Center(void) const
{
    return center;
}

double DeformableCircle::
Radius(void) const
{
    return radius;
}

DeformableCylinder::
DeformableCylinder(int numCircles, int numSamplesPerCircle)
:PointCloud()
{
    for (int i=0; i<numCircles; i++)
        m_aCircleArray.push_back(new DeformableCircle(numSamplesPerCircle));
}

DeformableCylinder::
~DeformableCylinder(void)
{
    for (int i=0; i<NCircles(); i++) delete m_aCircleArray[i];
    m_aCircleArray.clear();
}

void DeformableCylinder::
CoarseFit(PointCloud* pCloud)
{
    printf("DeformableCylinder::CoarseFit fit to %s ...\n", pCloud->Name().c_str());
    // default upright direction is X
    // position
    R3Box bbox = pCloud->BBox();
    double xmax = bbox.XMax();
    double xmin = bbox.XMin();
    for (int i=0; i<NCircles(); i++) {
        double xpos = (xmax-xmin)*(i+0.5)/NCircles() + xmin;
        m_aCircleArray[i]->SetCenter(R3Point(xpos, 0, 0));
    }
    // radius
    double avgRadius = 0;
    for (int i=0; i<pCloud->NPoints(); i++) {
        R3Point pos = pCloud->Vertex(i)->Position();
        double radius = sqrt(pos[1]*pos[1]+pos[2]*pos[2]);
        avgRadius += radius;
    }
    avgRadius /= pCloud->NPoints();
    for (int i=0; i<NCircles(); i++) {
        m_aCircleArray[i]->SetRadius(avgRadius);
    }
    UpdatePoints();
}

void DeformableCylinder::
FineFit(PointCloud* pCloud)
{
    printf("DeformableCylinder::FineFit fit to %s ...\n", pCloud->Name().c_str());
    // for each point in pCloud, find the closest circle
    vector< vector<R3Point> > closestPointArray(NCircles());
    vector<double> xArray(NCircles());
    double threshold = 0.2;
    for (int i=0; i<(int)xArray.size(); i++) xArray[i] = m_aCircleArray[i]->Position(0)[0];
    for (int i=0; i<pCloud->NPoints(); i++) {
        R3Point pos = pCloud->Vertex(i)->Position();
        for (int j=0; j<(int)xArray.size(); j++) {
            double diff = fabs(xArray[j] - pos[0]);
            if (diff < threshold)
                closestPointArray[j].push_back(pos);
        }
    }
    // estimate the best position and radius for each circle
    for (int i=0; i<(int)closestPointArray.size(); i++) {
        // center in y-z
        double avgY = 0;
        double avgZ = 0;
        double avgRadius = 0;
        if (closestPointArray[i].size() == 0) {
            printf("closetPointArray[i].size() == 0\n");
        }
        else {
            for (int j=0; j<(int)closestPointArray[i].size(); j++) {
                avgY += closestPointArray[i][j][1];
                avgZ += closestPointArray[i][j][2];
            }
            avgY /= closestPointArray[i].size();
            avgZ /= closestPointArray[i].size();
            // average radius
            for (int j=0; j<(int)closestPointArray[i].size(); j++) {
                double diffY = closestPointArray[i][j][1] - avgY;
                double diffZ = closestPointArray[i][j][2] - avgZ;
                avgRadius += sqrt(diffY*diffY + diffZ*diffZ);
            }
            avgRadius /= closestPointArray[i].size();
        }
        printf("Center = [%g %g %g] Radius = %g\n", xArray[i], avgY, avgZ, avgRadius);
        m_aCircleArray[i]->SetCenter(R3Point(xArray[i], avgY, avgZ));
        m_aCircleArray[i]->SetRadius(avgRadius);
    }
    UpdatePoints();
}

int DeformableCylinder::
NCircles(void) const
{
    return (int)m_aCircleArray.size();
}

void DeformableCylinder::
UpdatePoints(void)
{
    // clear points
    Clear();
    // add vertices
    for (int i=0; i<NCircles(); i++) {
        for (int j=0; j<m_aCircleArray[i]->NSamples(); j++) {
            AddVertex(m_aCircleArray[i]->Position(j), m_aCircleArray[i]->Normal(j));
        }
    }
}

PointCloudProperty* DeformableCylinder::
GenerateThickness(PointCloud* pSkeleton) const
{
    PointCloudProperty* pPrp = new PointCloudProperty(pSkeleton, "Thickness");
    for (int i=0; i<NCircles(); i++) {
        if (m_aCircleArray[i]->Radius() != m_aCircleArray[i]->Radius()) {
            pPrp->SetVertexValue(i, 0);
        }
        else {
            pPrp->SetVertexValue(i, m_aCircleArray[i]->Radius()*2.0);
        }
    }
    return pPrp;
}

PointCloudProperty* DeformableCylinder::
GenerateTapering(PointCloud* pSkeleton) const
{
    PointCloudProperty* pPrp = new PointCloudProperty(pSkeleton, "Tapering");
    // arc length
    vector<double> arcLengthArray;
    arcLengthArray.push_back(0);
    for (int i=1; i<pSkeleton->NPoints(); i++) {
        double base = arcLengthArray[arcLengthArray.size()-1];
        arcLengthArray.push_back(base + R3Distance(pSkeleton->Vertex(i)->Position(),
                                                   pSkeleton->Vertex(i-1)->Position()));
    }
    double radius = MyConfigParser.GetScalar("TaperingHistogram", "scale")*0.5;
    printf("DeformableCylinder::GenerateTapering radius = %g\n", radius);
    for (int i=0; i<pSkeleton->NPoints(); i++) {
        double arcLength = arcLengthArray[i];
        vector<double> radiusArray;
        vector<double> arcArray;
        // backward (including)
        for (int idx=i; idx>=0 && arcLength-arcLengthArray[idx]<radius; idx--) {
            radiusArray.push_back(m_aCircleArray[idx]->Radius());
            arcArray.push_back(arcLengthArray[idx]);
        }
        // forward (not including)
        for (int idx=i+1; idx<pSkeleton->NPoints() && arcLengthArray[idx]-arcLength<radius; idx++) {
            radiusArray.push_back(m_aCircleArray[idx]->Radius());
            arcArray.push_back(arcLengthArray[idx]);
        }
        double tapering = MathUtil::Tapering(radiusArray, arcArray);
        printf("    Vertex(%d) associated with %d neighbors, tapering = %g!\n", i, (int)radiusArray.size(), tapering);
        pPrp->SetVertexValue(i, tapering);
        // debug: create a local point cloud and write to file
        // PointCloud localPatch;
        //for (int j=0; j<(int)positionArray.size(); j++) {
        //    localPatch.AddVertex(positionArray[j], tangentArray[j]);
        // }
        // string filename = dirname + "/" + StringUtil::Number2String(i) + ".pts";
        // localPatch.WriteFile(filename.c_str());
        // printf("        write to %s!\n", filename.c_str());
    }
    return pPrp;
}


//PointCloudProperty* DeformableCylinder::
//GenerateCurvature(PointCloud* pSkeleton) const
//{
//    PointCloudProperty* pPrp = new PointCloudProperty(pSkeleton, "Curvature");
//    // two ends don't have curvature
//    pPrp->SetVertexValue(0, 0);
//    pPrp->SetVertexValue(NCircles()-1, 0);
//    for (int i=1; i<NCircles()-1; i++) {
//        R3Point pos[3];
//        pos[0] = pSkeleton->Vertex(i-1)->Position() - pSkeleton->Vertex(i)->Position().Vector();
//        pos[1] = R3Point(0, 0, 0);
//        pos[2] = pSkeleton->Vertex(i+1)->Position() - pSkeleton->Vertex(i)->Position().Vector();
//        R3Triad triad = MathUtil::GenerateTriad(pos[0].Vector() % pos[2].Vector());
//        R3Vector xDir = triad.Axis(RN_X);
//        R3Vector yDir = triad.Axis(RN_Y);
//        double x[2], y[2];
//        x[0] = pos[0].Vector().Dot(xDir);
//        y[0] = pos[0].Vector().Dot(yDir);
//        x[1] = pos[2].Vector().Dot(xDir);
//        y[1] = pos[2].Vector().Dot(yDir);
//        double t[2];
//        t[0] = (x[0]*x[0] + y[0]*y[0])*0.5;
//        t[1] = (x[1]*x[1] + y[1]*y[1])*0.5;
//        double curvature = 0;
//        if (y[0]*x[1] - y[1]*x[0] == 0.0 || x[0]*y[1] - x[1]*y[0] == 0.0) {
//            curvature = 0;
//        }
//        else {
//            double b = (t[0]*x[1] - t[1]*x[0]) / (y[0]*x[1] - y[1]*x[0]);
//            double a = (t[0]*y[1] - t[1]*y[0]) / (x[0]*y[1] - x[1]*y[0]);
//            curvature = 1.0/sqrt(a*a+b*b);
//        }
//        if (curvature != curvature) {
//            printf("pos[0] = [%g %g %g] pos[2] = [%g %g %g], curvature = NaN\n\n\n", pos[0][0], pos[0][1], pos[0][2],
//                   pos[2][0], pos[2][1], pos[2][2]);
//            exit(-1);
//        }
//        pPrp->SetVertexValue(i, curvature);
//    }
//    return pPrp;
//}

PointCloudProperty* DeformableCylinder::
GenerateCurvature(PointCloud* pSkeleton) const
{
    PointCloudProperty* pPrp = new PointCloudProperty(pSkeleton, "Curvature");
    vector<double> arcLengthArray;
    arcLengthArray.push_back(0);
    for (int i=1; i<pSkeleton->NPoints(); i++) {
        double base = arcLengthArray[arcLengthArray.size()-1];
        arcLengthArray.push_back(base + R3Distance(pSkeleton->Vertex(i)->Position(),
                                                   pSkeleton->Vertex(i-1)->Position()));
    }
    vector<double> radiusArray = MyConfigParser.GetScalarArray("CurvatureHistogram", "scale_array");
    double radius = radiusArray[0]*0.5;
    printf("DeformableCylinder::GenerateCurvature radius = %g\n", radius);
    // string dirname = string("localpatch/") + Name() + "/";
    // FileSystemUtil::CreateDirectory(dirname);
    for (int i=0; i<pSkeleton->NPoints(); i++) {
        double arcLength = arcLengthArray[i];
        vector<R3Point> positionArray;
        vector<R3Vector> tangentArray;
        // backward (including)
        for (int idx=i; idx>=0 && arcLength-arcLengthArray[idx]<radius; idx--) {
            positionArray.push_back(pSkeleton->Vertex(idx)->Position());
            tangentArray.push_back(pSkeleton->Vertex(idx)->Normal());
        }
        // forward (not including)
        for (int idx=i+1; idx<pSkeleton->NPoints() && arcLengthArray[idx]-arcLength<radius; idx++) {
            positionArray.push_back(pSkeleton->Vertex(idx)->Position());
            tangentArray.push_back(pSkeleton->Vertex(idx)->Normal());
        }
        double curvature = 0;
//        if (Name() == "leg-3" && i == 6) {
//            curvature = MathUtil::CurvatureDebugging(positionArray, tangentArray);
//        }
//        else {
            curvature = MathUtil::Curvature(positionArray, tangentArray);
//        }
        printf("    Vertex(%d) associated with %d neighbors, curvature = %g!\n", i, (int)positionArray.size(), curvature);
        pPrp->SetVertexValue(i, curvature);
        // debug: create a local point cloud and write to file
        // PointCloud localPatch;
        //for (int j=0; j<(int)positionArray.size(); j++) {
        //    localPatch.AddVertex(positionArray[j], tangentArray[j]);
        // }
        // string filename = dirname + "/" + StringUtil::Number2String(i) + ".pts";
        // localPatch.WriteFile(filename.c_str());
        // printf("        write to %s!\n", filename.c_str());
    }
    return pPrp;
}

PointCloudProperty* DeformableCylinder::
GenerateSkeletonProperty(const string& prpName, PointCloud* pSkeleton) const
{
    if (prpName == "Thickness") {
        return GenerateThickness(pSkeleton);
    }
    else if (prpName == "Curvature") {
        return GenerateCurvature(pSkeleton);
    }
    else if (prpName == "Tapering") {
        return GenerateTapering(pSkeleton);
    }
    else {
        fprintf(stderr, "DeformableCylinder::GenerateSkeletonProperty %s is unrecognizable!\n",
                prpName.c_str());
        exit(-1);
    }
}

PointCloudPropertySet* DeformableCylinder::
GetSkeletonProperties(const vector<string>& propertyNames) const
{
    // skeleton
    PointCloud* pSkeleton = new PointCloud;
    vector<R3Point> positionArray;
    // please note that Center() is not transformed
    for (int i=0; i<NCircles(); i++) {
        R3Vector dir(0, 0, 0);
        if (i == 0) {
            dir = m_aCircleArray[i+1]->Center() - m_aCircleArray[i]->Center();
            dir.Normalize();
        }
        else if (i == NCircles()-1) {
            dir = m_aCircleArray[i]->Center() - m_aCircleArray[i-1]->Center();
            dir.Normalize();
        }
        else {
            dir = m_aCircleArray[i+1]->Center() - m_aCircleArray[i-1]->Center();
            dir.Normalize();
        }
        pSkeleton->AddVertex(m_aCircleArray[i]->Center(), dir);
    }
    // property set
    PointCloudPropertySet* pPrpSet = new PointCloudPropertySet(pSkeleton);
    for (int i=0; i<(int)propertyNames.size(); i++) {
        pPrpSet->Insert(GenerateSkeletonProperty(propertyNames[i], pSkeleton));
    }
    return pPrpSet;
}

void DeformableCylinder::
Transform(const R3Transformation& transformation)
{
    // points
    for (int i=0; i<(int)m_vPointArray.size(); i++) {
        m_vPointArray[i]->Transform(transformation);
    }
    // circles
    for (int i=0; i<NCircles(); i++) {
        R3Point center = m_aCircleArray[i]->Center();
        center.Transform(transformation);
        m_aCircleArray[i]->SetCenter(center);
    }
}