#include "R3MyGraphics/R3MyGraphics.h"
#include "Utils/Utils.h"
using namespace std;

PointCloudVertex::
PointCloudVertex(const R3Point& _p, const R3Vector& _n, PointCloud* pCloud)
:position(_p), normal(_n), indexInCloud(-1),segIndex(-1), m_pCloud(pCloud)
{}

const R3Point& PointCloudVertex::
Position(void) const
{
    return position;
}

const R3Vector& PointCloudVertex::
Normal(void) const
{
    return normal;
}

int PointCloudVertex::
IndexInCloud(void) const
{
    return indexInCloud;
}

int PointCloudVertex::
SegmentIndex(void) const
{
    return segIndex;
}

int PointCloudVertex::
FaceIndex(void) const
{
    return faceIndex;
}

const string& PointCloudVertex::
SegmentName(void) const
{
    return m_pCloud->KthSegmentName(segIndex);
}

void PointCloudVertex::
Transform(const R3Transformation& transformation)
{
    position.Transform(transformation);
    normal.Transform(transformation);
    normal.Normalize();
}

void PointCloudVertex::
SetPosition(const R3Point& _p)
{
    position = _p;
}

void PointCloudVertex::
SetNormal(const R3Vector& _n)
{
    normal = _n;
}

void PointCloudVertex::
SetIndexInCloud(int index)
{
    indexInCloud = index;
}

void PointCloudVertex::
SetSegmentIndex(int index)
{
    segIndex = index;
}

void PointCloudVertex::
SetFaceIndex(int index)
{
    faceIndex = index;
}

PointCloud::
PointCloud(void)
:m_sName("")
{}

PointCloud::
~PointCloud(void)
{
    Clear();
}

R3Box PointCloud::
BBox(void) const
{
    R3Box bbox = R3null_box;
    for (int i=0; i<(int)m_vPointArray.size(); i++) {
        bbox.Union(m_vPointArray[i]->Position());
    }
    return bbox;
}

R3Point PointCloud::
Centroid(void) const
{
    return BBox().Centroid();
}

void PointCloud::
Transform(const R3Transformation& transformation)
{
    for (int i=0; i<(int)m_vPointArray.size(); i++) {
        m_vPointArray[i]->Transform(transformation);
    }
}

int PointCloud::
NPoints(void) const
{
    return (int)m_vPointArray.size();
}

PointCloudVertex* PointCloud::
Vertex(int k) const
{
    assert(k>=0 && k<NPoints());
    return m_vPointArray[k];
}

int PointCloud::
VertexSegmentID(int k) const
{
    return Vertex(k)->SegmentIndex();
}

PointCloudVertex* PointCloud::
ClosestVertex(PointCloudVertex* pVtx, PointCloudProperty* pMask,
              bool isCorresWithinSegment) const
{
    PointCloudVertex* pVertex = NULL;
    double shortestDist = FLT_MAX;
    R3Point position = pVtx->Position();
    for (int i=0; i<(int)m_vPointArray.size(); i++) {
        // mask
        if (pMask->VertexValue(i) < 0.5) continue;
        PointCloudVertex* _pVertex = m_vPointArray[i];
        // must be in the same segment
        if (isCorresWithinSegment && pVtx->SegmentName() != _pVertex->SegmentName()) continue;
        double dist = R3Distance(position, _pVertex->Position());
        if (dist < shortestDist) {
            shortestDist = dist;
            pVertex = _pVertex;
        }
    }
    // it is possible that pVtx is an extra segment, doesn't have correspondence
    // in the current point cloud if isCorresWithinSegment = true
    return pVertex;
}

int PointCloud::
ReadFile(const char* filename)
{
    const char *extension;
    if (!(extension = strrchr(filename, '.'))) {
        printf("Filename %s has no extension (e.g., .pts)\n", filename);
        return 0;
    }
    
    // Read file of appropriate type
    if (!strncmp(extension, ".pts", 4)) {
        if (!ReadPtsFile(filename)) return 0;
    }
    else {
        RNFail("Unable to read file %s (unrecognized extension: %s)\n", filename, extension);
        return 0;
    }
    return 1;
}

int PointCloud::
ReadPtsFile(const char* filename)
{
    Clear();
    // set name
    SetName(FileSystemUtil::GetBasename(filename, false));
    FILE* fp = fopen(filename, "rb");
    if (!fp) {
        fprintf(stderr, "PointCloud::ReadPtsFile unable to open points file %s!\n", filename);
        return 0;
    }
    float coordinates[6];
    while (fread(coordinates, sizeof(float), 6, fp) == (unsigned int) 6) {
        R3Point position(coordinates[0], coordinates[1], coordinates[2]);
        R3Vector normal(coordinates[3], coordinates[4], coordinates[5]);
        AddVertex(position, normal);
    }
    fclose(fp);
    return 1;
}

int PointCloud::
WriteFile(const char* filename) const
{
    const char* extension;
    if (!(extension = strrchr(filename, '.'))) {
        printf("Filename %s has no extension (e.g., .pts)\n", filename);
        return 0;
    }
    
    if (!strncmp(extension, ".pts", 4)) {
        if (!WritePtsFile(filename)) return 0;
    }
    else {
        RNFail("Unable to read file %s (unrecognized extension: %s)\n",
               filename, extension);
        return 0;
    }
    return 1;
}

int PointCloud::
WritePtsFile(const char* filename) const
{
    FILE* fp = fopen(filename, "wb");
    if (!fp) {
        fprintf(stderr, "PointCloud::WritePtsFile unable to open points file %s\n", filename);
        return 0;
    }
    float coordinates[6];
    for (int i=0; i<(int)m_vPointArray.size(); i++) {
        coordinates[0] = m_vPointArray[i]->Position()[0];
        coordinates[1] = m_vPointArray[i]->Position()[1];
        coordinates[2] = m_vPointArray[i]->Position()[2];
        coordinates[3] = m_vPointArray[i]->Normal()[0];
        coordinates[4] = m_vPointArray[i]->Normal()[1];
        coordinates[5] = m_vPointArray[i]->Normal()[2];
        fwrite(coordinates, sizeof(float), 6, fp);
    }
    fclose(fp);
    return 1;
}

void PointCloud::
Clear(void)
{
    for (int i=0; i<(int)m_vPointArray.size(); i++)
        delete m_vPointArray[i];
    m_vPointArray.clear();
}

void PointCloud::
AddVertex(const R3Point& position, const R3Vector& normal, int faceIdx)
{
    PointCloudVertex* pVertex = new PointCloudVertex(position, normal, this);
    pVertex->SetIndexInCloud(m_vPointArray.size());
    pVertex->SetFaceIndex(faceIdx);
    m_vPointArray.push_back(pVertex);
}

int PointCloud::
Import(R3Mesh* pMesh, int numSamples, bool isPhongShading)
{
    Clear();
    // total area
    double totalArea = 0;
    for (int i=0; i<pMesh->NFaces(); i++) {
        totalArea += pMesh->FaceArea(pMesh->Face(i));
    }
    int count = 0;
    RNSeedRandomScalar();
    int maxNumSamples = numSamples*2;
    for (int i=0; i<pMesh->NFaces(); i++) {
        R3MeshFace* pFace = pMesh->Face(i);
        // Get vertex positions
        R3MeshVertex *v0 = pMesh->VertexOnFace(pFace, 0);
        R3MeshVertex *v1 = pMesh->VertexOnFace(pFace, 1);
        R3MeshVertex *v2 = pMesh->VertexOnFace(pFace, 2);
        const R3Point& p0 = pMesh->VertexPosition(v0);
        const R3Point& p1 = pMesh->VertexPosition(v1);
        const R3Point& p2 = pMesh->VertexPosition(v2);
        const R3Vector& n0 = pMesh->VertexNormal(v0);
        const R3Vector& n1 = pMesh->VertexNormal(v1);
        const R3Vector& n2 = pMesh->VertexNormal(v2);
        
        // Determine number of points for face
        double idealNumSamplesPerFace = numSamples * pMesh->FaceArea(pFace) / totalArea;
        int numSamplesPerFace = (int) idealNumSamplesPerFace;
        double remainder = idealNumSamplesPerFace - numSamplesPerFace;
        if (remainder > RNRandomScalar()) numSamplesPerFace++;
        
        // Generate random points in face
        for (int j = 0; j < numSamplesPerFace; j++) {
            if (count == maxNumSamples) break;
            RNScalar r1 = sqrt(RNRandomScalar());
            RNScalar r2 = RNRandomScalar();
            RNScalar t0 = (1.0 - r1);
            RNScalar t1 = r1 * (1.0 - r2);
            RNScalar t2 = r1 * r2;
            R3Point position = t0*p0 + t1*p1 + t2*p2;
            R3Vector normal;
            if (isPhongShading) {
                normal = t0*n0 + t1*n1 + t2*n2;
                normal.Normalize();
            }
            else {
                normal = pMesh->FaceNormal(pFace);
            }
            AddVertex(position, normal, pMesh->FaceID(pFace));
            count ++;
        }
    }
    assert(count <= maxNumSamples);
//    // copy segment name array
//    m_vSegmentNameArray.clear();
//    for (int i=0; i<pMesh->NSegments(); i++) {
//        m_vSegmentNameArray.push_back(pMesh->KthSegmentName(i));
//    }
    return 1;
}

void PointCloud::
UpdateSegmentationBasedOnTriangleIndices(R3Mesh* pMesh)
{
    // copy segment name array
    m_vSegmentNameArray.clear();
    for (int i=0; i<pMesh->NSegments(); i++) {
        m_vSegmentNameArray.push_back(pMesh->KthSegmentName(i));
    }
    // set segment index
    for (int i=0; i<NPoints(); i++) {
        R3MeshFace* pFace = pMesh->Face(Vertex(i)->FaceIndex());
        Vertex(i)->SetSegmentIndex(pMesh->FaceSegment(pFace));
    }
}

int PointCloud::
NSegments(void) const
{
    return (int)m_vSegmentNameArray.size();
}

const string& PointCloud::
KthSegmentName(int k) const
{
    assert(k>=0 && k<NSegments());
    return m_vSegmentNameArray[k];
}

void PointCloud::
SetName(const string& name)
{
    m_sName = name;
}

const string& PointCloud::
Name(void) const
{
    return m_sName;
}

// set segmentation indices given a property set that encodes segmentation
void PointCloud::
SetSegmentation(PointCloudPropertySet* pSegmentation)
{
    if (pSegmentation->GetPointCloud() != this) {
        fprintf(stderr, "PointCloud::SetSegmentation point cloud must match!\n");
        exit(-1);
    }
    m_vSegmentNameArray.clear();
    for (int i=0; i<pSegmentation->NProperties(); i++) {
        PointCloudProperty* pSeg = pSegmentation->Property(i);
        if (pSeg->Name() == "all") continue;
        int segIndex = m_vSegmentNameArray.size();
        m_vSegmentNameArray.push_back(pSeg->Name());
        for (int j=0; j<NPoints(); j++) {
            if (pSeg->VertexValue(j) > 0.5) {
                Vertex(j)->SetSegmentIndex(segIndex);
            }
        }
    }
}

// decompose this point cloud according to segmentation
const vector<PointCloud*> PointCloud::
SegmentPointCloud(PointCloudPropertySet* pSegmentation) const
{
    if (pSegmentation->GetPointCloud() != this) {
        fprintf(stderr, "PointCloud::SegmentPointCloud point cloud must match!\n");
        exit(-1);
    }
    vector<PointCloud*> segmentArray;
    for (int i=0; i<pSegmentation->NProperties(); i++) {
        PointCloudProperty* pSeg = pSegmentation->Property(i);
        if (pSeg->Name() == "all") continue;
        PointCloud* pSegCloud = new PointCloud;
        pSegCloud->SetName(pSeg->Name());
        for (int j=0; j<pSeg->NVertexValues(); j++) {
            if (pSeg->VertexValue(j) < 0.5) continue;
            pSegCloud->AddVertex(Vertex(j)->Position(), Vertex(j)->Normal());
        }
        segmentArray.push_back(pSegCloud);
    }
    return segmentArray;
}


// Alignment
const R3Affine PointCloud::
ICPAlign(PointCloudProperty* pMask,
         PointCloud* pTransformedCloud,
         PointCloudProperty* pTransformedMask,
         bool isCorresWithinSegment,
         double* pMinDistance) const
{
    R3Affine totalTransform = R3identity_affine;
    // initialize - scale
    R3Box box1 = pTransformedCloud->BBox();
    R3Box box2 = this->BBox();
    R3Point c1 = box1.Centroid();
    R3Point c2 = box2.Centroid();
    double scale[3];
    scale[0] = box2.XLength() / box1.XLength();
    scale[1] = box2.YLength() / box1.YLength();
    scale[2] = box2.ZLength() / box1.ZLength();
    R3Affine transformation = R3identity_affine;
    transformation.Scale(R3Vector(scale[0], scale[1], scale[2]));
    pTransformedCloud->Transform(transformation);
    totalTransform.Transform(transformation);
    c1.Transform(transformation);
    // initialize - translate
    R3Vector offset = c2 - c1;
    transformation = R3identity_affine;
    transformation.Translate(offset);
    pTransformedCloud->Transform(transformation);
    totalTransform.Transform(transformation);
#ifdef DEBUG
    printf("    Initialize\n"
           "        scale = [%g, %g, %g], offset = [%g, %g, %g]\n",
           scale[0], scale[1], scale[2], offset[0], offset[1], offset[2]);
    FileSystemUtil::WaitInput();
#endif
    // iteration
    int count = 0;
    double minDistance = FLT_MAX;
    while (true) {
#ifdef DEBUG
        printf("Iteration %d\n", count);
#endif
        count ++;
        // find correspondence both sides
        vector<PointCloudVertex*> correspondences[2];
        for (int i=0; i<pTransformedCloud->NPoints(); i++) {
            if (pTransformedMask->VertexValue(i) < 0.5) continue;
            PointCloudVertex* pVtx1 = pTransformedCloud->Vertex(i);
            PointCloudVertex* pVtx2 = this->ClosestVertex(pVtx1, pMask, isCorresWithinSegment);
            if (!pVtx2) continue;
            correspondences[0].push_back(pVtx1);
            correspondences[1].push_back(pVtx2);
        }
        for (int i=0; i<this->NPoints(); i++) {
            if (pMask->VertexValue(i) < 0.5) continue;
            PointCloudVertex* pVtx2 = this->Vertex(i);
            PointCloudVertex* pVtx1 = pTransformedCloud->ClosestVertex(pVtx2, pTransformedMask, isCorresWithinSegment);
            if (!pVtx1) continue;
            correspondences[0].push_back(pVtx1);
            correspondences[1].push_back(pVtx2);
        }
#ifdef DEBUG
        printf("        found %d pairs of correspondences\n",
               (int)correspondences[0].size());
        double squareSumDist = 0;
        for (int i=0; i<(int)correspondences[0].size(); i++) {
            double dist = R3Distance(correspondences[0][i]->Position(),
                                     correspondences[1][i]->Position());
            squareSumDist += dist * dist;
        }
        printf("        before alignment square-sum-distance = %g\n", squareSumDist);
        FileSystemUtil::WaitInput();
#endif
        // scale
        double yxSum[3] = {0, 0, 0};
        double xxSum[3] = {0, 0, 0};
        double yAvg[3] = {0, 0, 0};
        double xAvg[3] = {0, 0, 0};
        double x[3], y[3];
        for (int i=0; i<(int)correspondences[0].size(); i++) {
            R3Point pos[2];
            pos[0] = correspondences[0][i]->Position();
            pos[1] = correspondences[1][i]->Position();
            x[0] = pos[0][0];
            x[1] = pos[0][1];
            x[2] = pos[0][2];
            y[0] = pos[1][0];
            y[1] = pos[1][1];
            y[2] = pos[1][2];
            yxSum[0] += y[0]*x[0];
            xxSum[0] += x[0]*x[0];
            yAvg[0] += y[0];
            xAvg[0] += x[0];
            yxSum[1] += y[1]*x[1];
            xxSum[1] += x[1]*x[1];
            yAvg[1] += y[1];
            xAvg[1] += x[1];
            yxSum[2] += y[2]*x[2];
            xxSum[2] += x[2]*x[2];
            yAvg[2] += y[2];
            xAvg[2] += x[2];
        }
        yAvg[0] /= correspondences[0].size();
        yAvg[1] /= correspondences[0].size();
        yAvg[2] /= correspondences[0].size();
        xAvg[0] /= correspondences[0].size();
        xAvg[1] /= correspondences[0].size();
        xAvg[2] /= correspondences[0].size();
        double scale[3];
        scale[0] = yxSum[0] / xxSum[0];
        scale[1] = yxSum[1] / xxSum[1];
        scale[2] = yxSum[2] / xxSum[2];
        // offset
        double offset[3];
        offset[0] = yAvg[0]/scale[0] - xAvg[0];
        offset[1] = yAvg[1]/scale[1] - xAvg[1];
        offset[2] = yAvg[2]/scale[2] - xAvg[2];
        // transform pTransformedCloud: -> scale -> translate
        // scale
        R3Affine transformation = R3identity_affine;
        transformation.Scale(R3Vector(scale[0], scale[1], scale[2]));
        pTransformedCloud->Transform(transformation);
        totalTransform.Transform(transformation);
        // translation
        transformation = R3identity_affine;
        transformation.Translate(R3Vector(offset[0], offset[1], offset[2]));
        pTransformedCloud->Transform(transformation);
        totalTransform.Transform(transformation);
#ifdef DEBUG
        printf("        scale = [%g, %g, %g], offset = [%g, %g, %g]\n",
               scale[0], scale[1], scale[2], offset[0], offset[1], offset[2]);
        // square sum distance
        squareSumDist = 0;
        for (int i=0; i<(int)correspondences[0].size(); i++) {
            double dist = R3Distance(correspondences[0][i]->Position(),
                                     correspondences[1][i]->Position());
            squareSumDist += dist * dist;
        }
        printf("        after alignment square-sum-distance = %g\n", squareSumDist);
        FileSystemUtil::WaitInput();
        printf("\n");
#endif
        if (fabs(scale[0]-1.0)+fabs(scale[1]-1.0)+fabs(scale[2]-1.0)+fabs(offset[0])+fabs(offset[1])+fabs(offset[2]) < 1e-6) {
            minDistance = 0;
            for (int i=0; i<(int)correspondences[0].size(); i++) {
                double dist = R3Distance(correspondences[0][i]->Position(),
                                         correspondences[1][i]->Position());
                minDistance += dist * dist;
            }
            break;
        }
    
    }
    if (pMinDistance) *pMinDistance = minDistance;
    printf("    finished in %d iterations!\n", count);
    printf("    total scaling = [%g, %g, %g]\n",
           totalTransform.Matrix()[0][0],
           totalTransform.Matrix()[1][1],
           totalTransform.Matrix()[2][2]);
    return totalTransform;
}

R3Triad PointCloud::
PrincipleAxes(const R3Point* center, RNScalar* variances) const
{
    // Get centroid
    R3Point centroid = (center) ? *center : Centroid();
    
    // Compute covariance matrix
    RNArea area = 0;
    RNScalar m[9] = { 0 };
    for (int i=0; i<NPoints(); i++) {
        R3Point pos = Vertex(i)->Position();
        RNScalar x = pos[0] - centroid[0];
        RNScalar y = pos[1] - centroid[1];
        RNScalar z = pos[2] - centroid[2];
        m[0] += x*x;
        m[4] += y*y;
        m[8] += z*z;
        m[1] += x*y;
        m[3] += x*y;
        m[2] += x*z;
        m[6] += x*z;
        m[5] += y*z;
        m[7] += y*z;
        area += 1.0;
    }
    
    // Normalize covariance matrix
    if (area == 0) return R3xyz_triad;
    for (int i = 0; i < 9; i++) m[i] /= area;
    
    // Compute eigenvalues and eigenvectors
    RNScalar U[9];
    RNScalar W[3];
    RNScalar Vt[9];
    RNSvdDecompose(3, 3, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
    
    // Copy principle axes into more convenient form
    // W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
    R3Vector axes[3];
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            axes[i][j] = Vt[3*i+j];
        }
    }
    
    // Flip axes so that "heavier" on positive side for first two dimensions
    int positive_count[3] = { 0, 0, 0 };
    int negative_count[3] = { 0, 0, 0 };
    for (int i = 0; i < NPoints(); i++) {
        R3Vector vertex_vector = Vertex(i)->Position() - centroid;
        for (int j = 0; j < 2; j++) {
            RNScalar dot = axes[j].Dot(vertex_vector);
            if (dot > 0.0) positive_count[j]++;
            else negative_count[j]++;
        }
    }
    for (int j =0; j < 2; j++) {
        if (positive_count[j] < negative_count[j]) {
            axes[j].Flip();
        }
    }
    
    // Set third axis to form orthonormal triad with other two
    axes[2] = axes[0] % axes[1];
    
    // Just checking
    assert(RNIsEqual(axes[0].Length(), 1.0, RN_BIG_EPSILON));
    assert(RNIsEqual(axes[1].Length(), 1.0, RN_BIG_EPSILON));
    assert(RNIsEqual(axes[2].Length(), 1.0, RN_BIG_EPSILON));
    assert(RNIsZero(axes[0].Dot(axes[1]), RN_BIG_EPSILON));
    assert(RNIsZero(axes[1].Dot(axes[2]), RN_BIG_EPSILON));
    assert(RNIsZero(axes[0].Dot(axes[2]), RN_BIG_EPSILON));
    
    // Return variances (eigenvalues)
    if (variances) {
        variances[0] = W[0];
        variances[1] = W[1];
        variances[2] = W[2];
    }
    
    // Return triad
    return R3Triad(axes[0], axes[1], axes[2]);
}


R3Affine PointCloud::
PCATransformation(void) const
{
    R3Affine affine(R3identity_affine);
    R3Point centroid = Centroid();
    RNScalar variances[3];
    R3Triad triad = PrincipleAxes(&centroid, variances);
    affine.Transform(R3Affine(triad.InverseMatrix()));
    affine.Translate(-(centroid.Vector()));
    return affine;
}
