#ifndef MY_POINT_CLOUD_H
#define MY_POINT_CLOUD_H
#include <vector>
class PointCloud;
class PointCloudVertex {
public:
    PointCloudVertex(const R3Point& position, const R3Vector& normal, PointCloud* pCloud);
    const R3Point& Position(void) const;
    const R3Vector& Normal(void) const;
    int IndexInCloud(void) const;
    int SegmentIndex(void) const;
    int FaceIndex(void) const;
    const std::string& SegmentName(void) const;
    void Transform(const R3Transformation& transformation);
    void SetPosition(const R3Point& position);
    void SetNormal(const R3Vector& normal);
    void SetIndexInCloud(int index);
    void SetSegmentIndex(int index);
    void SetFaceIndex(int index);
protected:
    R3Point position;
    R3Vector normal;
    int indexInCloud;
    int segIndex;
    PointCloud* m_pCloud;
    // for multiple segmentation based on face index
    int faceIndex;
};

class PointCloud {
public:
    // Constructor functions
    PointCloud(void);
    virtual ~PointCloud(void);
    
    // Properties
    R3Box BBox(void) const;
    R3Point Centroid(void) const;
    int NSegments(void) const;
    const std::string& KthSegmentName(int k) const;
    const std::string& Name(void) const;
    
    // Transform
    virtual void Transform(const R3Transformation& transformation);
    R3Triad PrincipleAxes(const R3Point* center, RNScalar* variances) const;
    R3Affine PCATransformation(void) const; // only translation and rotation
    // Access
    int NPoints(void) const;
    PointCloudVertex* Vertex(int k) const;
    int VertexSegmentID(int k) const;
    PointCloudVertex* ClosestVertex(PointCloudVertex* pVtx, PointCloudProperty* pMask, bool isCorresWithinSegment) const;
    
    // I/O functions
    int ReadFile(const char* filename);
    int ReadPtsFile(const char* filename);
    int WriteFile(const char* filename) const;
    int WritePtsFile(const char* filename) const;
    void Clear(void);
    void AddVertex(const R3Point& position, const R3Vector& normal, int faceIdx=-1);
    void SetName(const std::string& name);
    
    // Convertion functions
    int Import(R3Mesh* pMesh, int numSamples, bool isPhongShading);
    void UpdateSegmentationBasedOnTriangleIndices(R3Mesh* pMesh);
    void SetSegmentation(PointCloudPropertySet* pSegmentation);
    const std::vector<PointCloud*> SegmentPointCloud(PointCloudPropertySet* pSegmentation) const;
    // Alignment
    const R3Affine ICPAlign(PointCloudProperty* pMask,
                            PointCloud* pTransformedCloud,
                            PointCloudProperty* pTransformedMask,
                            bool isCorresWithinSegment,
                            double* pMinDistance=NULL) const;
protected:
    std::vector<PointCloudVertex*> m_vPointArray;
    std::vector<std::string> m_vSegmentNameArray;
    std::string m_sName;
};

#endif