#include "R3MyGraphics/R3MyGraphics.h"
#include "Utils/Utils.h"
using namespace std;

PointCloudProperty::
PointCloudProperty(PointCloud* pCloud, const string& name)
:m_pPointCloud(pCloud),
m_sName(name),
m_vValues(pCloud->NPoints(), 0.0),
m_fMean(RN_UNKNOWN),
m_fStddev(RN_UNKNOWN),
m_fMin(RN_UNKNOWN),
m_fMax(RN_UNKNOWN)
{}

PointCloudProperty::
~PointCloudProperty(void)
{}

PointCloud* PointCloudProperty::
GetPointCloud(void) const
{
    return m_pPointCloud;
}

const string& PointCloudProperty::
Name(void) const
{
    return m_sName;
}

int PointCloudProperty::
NVertexValues(void) const
{
    return (int)m_vValues.size();
}

double PointCloudProperty::
VertexValue(PointCloudVertex* pVertex) const
{
    return VertexValue(pVertex->IndexInCloud());
}

double PointCloudProperty::
VertexValue(int vertexIdx) const
{
    assert(vertexIdx>=0 && vertexIdx<NVertexValues());
    return m_vValues[vertexIdx];
}

double PointCloudProperty::
Percentile(double percentile) const
{
    if (NVertexValues() == 0) return 0;
    // copy values
    vector<double> values = m_vValues;
    MathUtil::Sort(values);
    int index = (int) (values.size() * percentile / 100.0);
    if (index >= values.size()) index = values.size() - 1;
    return values[index];
}

void PointCloudProperty::
SetVertexValue(PointCloudVertex* pVertex, double value)
{
    SetVertexValue(pVertex->IndexInCloud(), value);
}

void PointCloudProperty::
SetVertexValue(int vertexIdx, double value)
{
    assert(vertexIdx>=0 && vertexIdx<NVertexValues());
    m_vValues[vertexIdx] = value;
}

void PointCloudProperty::
RemoveOutliers(double percentile)
{
    m_vValues = MathUtil::RemoveOutliers(m_vValues, percentile);
}

void PointCloudProperty::
NormalizeByMeanAndStd(void)
{
    // Get mean and stddev
    double m = Mean();
    double s = StandardDeviation();
    // Check stddev
    if (s == 0) {
        // Set everything to zero
        for (int i = 0; i < NVertexValues(); i++) {
            m_vValues[i] = 0;
        }
    }
    else {
        // Normalize by mean and stddev
        for (int i = 0; i < NVertexValues(); i++) {
            m_vValues[i] = (m_vValues[i] - m) / s;
        }
    }
    // Reset statistics
    ResetStatistics();
}

void PointCloudProperty::
Normalize(void)
{
    double minVal = Minimum();
    double maxVal = Maximum();
    for (int i=0; i<NVertexValues(); i++) {
        m_vValues[i] = (m_vValues[i] - minVal) / (maxVal - minVal);
    }
    ResetStatistics();
}

double PointCloudProperty::
Mean(void) const
{
    // Compute mean of property values
    if (m_fMean == RN_UNKNOWN) {
        PointCloudProperty *property = (PointCloudProperty *) this;
        property->m_fMean = MathUtil::Average(property->m_vValues);
    }
    // Return mean
    return m_fMean;
}

double PointCloudProperty::
StandardDeviation(void) const
{
    if (m_fStddev == RN_UNKNOWN) {
        PointCloudProperty *property = (PointCloudProperty *) this;
        property->m_fStddev = MathUtil::StandardDeviation(property->m_vValues);
    }
    return m_fStddev;
}

double PointCloudProperty::
MeanGivenMask(PointCloudProperty* pMask, double minVal, double maxVal) const
{
    vector<double> valueArray;
    for (int i=0; i<(int)m_vValues.size(); i++) {
        if (pMask->VertexValue(i) < 0.5) continue;
        if (m_vValues[i]<minVal) valueArray.push_back(minVal);
        else if (m_vValues[i]>maxVal) valueArray.push_back(maxVal);
        else valueArray.push_back(m_vValues[i]);
    }
    return MathUtil::Average(valueArray);
}

double PointCloudProperty::
StandardDeviationGivenMask(PointCloudProperty* pMask, double minVal, double maxVal) const
{
    vector<double> valueArray;
    for (int i=0; i<(int)m_vValues.size(); i++) {
        if (pMask->VertexValue(i) < 0.5) continue;
        if (m_vValues[i]<minVal) valueArray.push_back(minVal);
        else if (m_vValues[i]>maxVal) valueArray.push_back(maxVal);
        else valueArray.push_back(m_vValues[i]);
    }
    return MathUtil::StandardDeviation(valueArray);
}

double PointCloudProperty::
Minimum(void) const
{
    if (m_fMin == RN_UNKNOWN) {
        PointCloudProperty *property = (PointCloudProperty *) this;
        property->m_fMin = MathUtil::Minimum(property->m_vValues);
    }
    return m_fMin;
}

double PointCloudProperty::
Maximum(void) const
{
    if (m_fMax == RN_UNKNOWN) {
        PointCloudProperty *property = (PointCloudProperty *) this;
        property->m_fMax = MathUtil::Maximum(property->m_vValues);
    }
    return m_fMax;
}

const vector<double> PointCloudProperty::
Histogram(int numBins, double minVal, double maxVal) const
{
    return MathUtil::GenerateHistogram(m_vValues, numBins, minVal, maxVal);
}

const vector<double> PointCloudProperty::
HistogramGivenMask(int numBins, double minVal, double maxVal, PointCloudProperty* pMask) const
{
    vector<double> valueArray;
    // if ith vertex is included, pMask->VertexValue(i) = 1.0
    // else pMask->VertexValue(i) = 0.0
    for (int i=0; i<(int)m_vValues.size(); i++) {
        if (pMask->VertexValue(i) > 0.5) valueArray.push_back(m_vValues[i]);
    }
    return MathUtil::GenerateHistogram(valueArray, numBins, minVal, maxVal);
}

void PointCloudProperty::
ResetStatistics(void)
{
    m_fMean = RN_UNKNOWN;
    m_fStddev = RN_UNKNOWN;
    m_fMin = RN_UNKNOWN;
    m_fMax = RN_UNKNOWN;
}

// Create interpolated property
// can be used to create volume property
PointCloudProperty* PointCloudProperty::
InterpolateProperty(PointCloud* pAlterCloud, double sigma) const
{
    // w(p,x) = exp(-(p-x)^2/sigma^2), where p is a point on the current point cloud,
    // x is a point on pAlterCloud
    // f(x) = \sum_p w(p,x)f(p) / \sum w(p,x)
    PointCloudProperty* pAlterPrp = new PointCloudProperty(pAlterCloud, Name());
    const PointCloud* pCloud = m_pPointCloud;
    // go through all points on pAlterCloud
    for (int i=0; i<pAlterCloud->NPoints(); i++) {
        R3Point xposition = pAlterCloud->Vertex(i)->Position();
        double weightSum = 0;
        double weightedValueSum = 0;
        // go through all points on pCloud
        for (int j=0; j<pCloud->NPoints(); j++) {
            R3Point pposition = pCloud->Vertex(j)->Position();
            double dist = R3Distance(pposition, xposition);
            double w = exp(-dist*dist/sigma/sigma);
            weightSum += w;
            weightedValueSum += w*VertexValue(j);
        }
        if (weightSum == 0) pAlterPrp->SetVertexValue(i, 0);
        else pAlterPrp->SetVertexValue(i, weightedValueSum / weightSum);
    }
    return pAlterPrp;
}

void PointCloudProperty::
SetAllOnes(void)
{
    for (int i=0; i<(int)m_vValues.size(); i++) m_vValues[i] = 1.0;
    ResetStatistics();
}