#include "R3MyGraphics/R3MyGraphics.h"
#include <fstream>
using namespace std;

PointCloudPropertySet::
PointCloudPropertySet(PointCloud* pCloud)
:m_pPointCloud(pCloud),
m_vProperties()
{}

PointCloudPropertySet::
~PointCloudPropertySet(void)
{
    Clear();
}

PointCloud* PointCloudPropertySet::
GetPointCloud(void) const
{
    return m_pPointCloud;
}

int PointCloudPropertySet::
NProperties(void) const
{
    return (int)m_vProperties.size();
}

PointCloudProperty* PointCloudPropertySet::
Property(int k) const
{
    assert(k>=0 && k<NProperties());
    return m_vProperties[k];
}

PointCloudProperty* PointCloudPropertySet::
Property(const string& name) const
{
    for (int i=0; i<(int)m_vProperties.size(); i++) {
        if (m_vProperties[i]->Name() == name) return m_vProperties[i];
    }
    return NULL;
}

int PointCloudPropertySet::
ReadFile(const char* filename)
{
    const char *extension;
    if (!(extension = strrchr(filename, '.'))) {
        printf("Filename %s has no extension (e.g., .arff)\n", filename);
        return 0;
    }
    
    // Read file of appropriate type
    if (!strncmp(extension, ".arff", 5)) {
        if (!ReadARFF(filename)) return 0;
    }
    else {
        RNFail("Unable to read file %s (unrecognized extension: %s)\n", filename, extension);
        return 0;
    }
    return 1;
}

int PointCloudPropertySet::
ReadARFF(const char* filename)
{
    Clear();
    // Open file
    FILE *fp = fopen(filename, "r");
    if (!fp) {
        fprintf(stderr, "Unable to open ARFF file: %s\n", filename);
        return 0;
    }
    
    // Read attribute names and allocate property for each one
    const int max_line_size = 32 * 1024;
    char buffer[max_line_size], token[1024], name[1024];
    while (fgets(buffer, max_line_size, fp)) {
        if (strstr(buffer, "@data") || strstr(buffer, "@DATA")) break;
        else if (strstr(buffer, "@attribute") || strstr(buffer, "@ATTRIBUTE")) {
            if (sscanf(buffer, "%s%s", token, name) == (unsigned int) 2) {
                if (!strcmp(token, "@attribute") || !strcmp(token, "@ATTRIBUTE")) {
                    PointCloudProperty *pProperty = new PointCloudProperty(m_pPointCloud, name);
                    Insert(pProperty);
                }
            }
        }
    }
    
    // Check number of properties
    if (NProperties() == 0) return 1;
    
    // Read data and assign property values
    int vertex_index = 0;
    RNScalar property_value = 0;
    while (fgets(buffer, max_line_size, fp)) {
        // Check for header line
        if (buffer[0] == '#') continue;
        if (buffer[0] == '@') continue;
        if (buffer[0] == '\0') continue;
        if (buffer[0] == '\n') continue;
        char *bufferp = strtok(buffer, "\n\t ,");
        for (int i = 0; i < NProperties(); i++) {
            PointCloudProperty *pProperty = Property(i);
            if (!bufferp) {
                fprintf(stderr, "Unable to read property value %d for vertex %d\n", i, vertex_index);
                return 0;
            }
            if (sscanf(bufferp, "%lf", &property_value) != (unsigned int) 1) {
                fprintf(stderr, "Unable to read property value %d for vertex %d\n", i, vertex_index);
                return 0;
            }
            if (vertex_index >= m_pPointCloud->NPoints()) {
                fprintf(stderr, "Too many data lines in %s\n", filename);
                return 0;
            }
            pProperty->SetVertexValue(vertex_index, property_value);
            bufferp = strtok(NULL, "\n\t ,");
        }
        vertex_index++;
    }
    
    // Check if read value for every vertex
    if (vertex_index != m_pPointCloud->NPoints()) {
       fprintf(stderr, "Mismatching number of data lines (%d) and vertices (%d) in %s\n",
               vertex_index, m_pPointCloud->NPoints(), filename);
       return 0;
    }
    
    // Close file
    fclose(fp);
    
    // Return success
    return 1;
}

int PointCloudPropertySet::
WriteFile(const char* filename) const
{
    const char* extension;
    if (!(extension = strrchr(filename, '.'))) {
        printf("Filename %s has no extension (e.g., .arff)\n", filename);
        return 0;
    }
    
    if (!strncmp(extension, ".arff", 5)) {
        if (!WriteARFF(filename)) return 0;
    }
    else {
        RNFail("Unable to read file %s (unrecognized extension: %s)\n",
               filename, extension);
        return 0;
    }
    return 1;
}

int PointCloudPropertySet::
WriteARFF(const char* filename) const
{
    ofstream fout(filename);
    if (!fout) {
        fprintf(stderr, "PointCloudPropertySet::WriteARFF unable to find %s\n", filename);
        return 0;
    }
    // header
    /*relation*/
    fout<<"@RELATION default"<<std::endl<<std::endl;
    for (int i=0; i<NProperties(); i++) {
        fout<<"@ATTRIBUTE "<<Property(i)->Name()<<" NUMERIC"<<std::endl;
    }
    fout<<std::endl;
    // data
    fout<<"@DATA"<<std::endl;
    for (int i=0; i<m_pPointCloud->NPoints(); i++) {
        for (int j=0; j<NProperties(); j++) {
            fout<<Property(j)->VertexValue(i);
            if (j!=NProperties()-1) fout<<",";
            else fout<<std::endl;
        }
    }
    fout.close();
    return 1;
}

void PointCloudPropertySet::
CreateSegmentMasks(void)
{
    // clean up
    Clear();
    // create a mask for 'all'
    PointCloudProperty* pProperty = new PointCloudProperty(m_pPointCloud, "all");
    for (int i=0; i<m_pPointCloud->NPoints(); i++)
        pProperty->SetVertexValue(i, 1);
    Insert(pProperty);
    // create a mask for each segment
    for (int k=0; k<m_pPointCloud->NSegments(); k++) {
        pProperty = new PointCloudProperty(m_pPointCloud, m_pPointCloud->KthSegmentName(k));
        for (int i=0; i<m_pPointCloud->NPoints(); i++) {
            if (m_pPointCloud->VertexSegmentID(i) == k)
                pProperty->SetVertexValue(i, 1);
            else
                pProperty->SetVertexValue(i, 0);
        }
        Insert(pProperty);
    }
}

void PointCloudPropertySet::
Insert(PointCloudProperty* pProperty)
{
    m_vProperties.push_back(pProperty);
}

void PointCloudPropertySet::
Clear(void)
{
    for (int i=0; i<(int)m_vProperties.size(); i++) {
        delete m_vProperties[i];
    }
    m_vProperties.clear();
}
