#ifndef MY_POINT_CLOUD_PROPERTY_SET_H
#define MY_POINT_CLOUD_PROPERTY_SET_H
#include <vector>
#include <string>

class PointCloudProperty {
public:
    PointCloudProperty(PointCloud* pCloud, const std::string& name);
    ~PointCloudProperty(void);
    // Access
    PointCloud* GetPointCloud(void) const;
    const std::string& Name(void) const;
    const std::vector<double> ValueArray(PointCloudProperty* pMask = NULL) const;
    // Value access
    int NVertexValues(void) const;
    double VertexValue(PointCloudVertex* pVertex) const;
    double VertexValue(int vertexIdx) const;
    // Statistics functions
    double Percentile(double percentile) const;
    double Mean(void) const;
    double StandardDeviation(void) const;
    double MeanGivenMask(PointCloudProperty* pMask, double minVal, double maxVal) const;
    double StandardDeviationGivenMask(PointCloudProperty* pMask, double minVal, double maxVal) const;
    double Minimum(void) const;
    double Maximum(void) const;
    const std::vector<double> Histogram(int numBins, double minVal, double maxVal) const;
    const std::vector<double> HistogramGivenMask(int numBins, double minVal, double maxVal, PointCloudProperty* pMask) const;
    // Manipulation functions
    void SetVertexValue(PointCloudVertex* pVertex, double value);
    void SetVertexValue(int vertexIdx, double value);
    void RemoveOutliers(double percentile);
    void NormalizeByMeanAndStd(void);
    void Normalize(void); // to [0, 1]
    // Create interpolated property
    // can be used to create volume property
    PointCloudProperty* InterpolateProperty(PointCloud* pAlterCloud, double sigma) const;
    void SetAllOnes(void);
protected:
    void ResetStatistics(void);
protected:
    PointCloud* m_pPointCloud;
    std::string m_sName;
    std::vector<double> m_vValues;
    double m_fMean;
    double m_fStddev;
    double m_fMin;
    double m_fMax;
};

class PointCloudPropertySet {
public:
    PointCloudPropertySet(PointCloud* pCloud);
    ~PointCloudPropertySet(void);
    // Access
    PointCloud* GetPointCloud(void) const;
    // Property access functions
    int NProperties(void) const;
    PointCloudProperty* Property(int k) const;
    PointCloudProperty* Property(const std::string& name) const;
    // I/O
    int ReadFile(const char* filename);
    int ReadARFF(const char* filename);
    int WriteFile(const char* filename) const;
    int WriteARFF(const char* filename) const;
    // Connect segmentation to property set
    void CreateSegmentMasks(void);
    // Manipulation functions
    void Insert(PointCloudProperty* pProperty);
    void Clear(void);
protected:
    PointCloud* m_pPointCloud;
    std::vector<PointCloudProperty*> m_vProperties;
};
#endif