#ifndef MY_R3_MY_CAMERA_H
#define MY_R3_MY_CAMERA_H
/* Include file for R3 camera class */



/* Initialization functions */

int R3InitMyCamera();
void R3StopMyCamera();



/* Class definition */

class R3MyCamera {
public:
    // Constructor functions
	R3MyCamera(void);
	R3MyCamera(const R3CoordSystem &cs,
               RNAngle xfov, RNAngle yfov, RNLength neardist, RNLength fardist);
	R3MyCamera(const R3Point& origin, const R3Triad& triad,
               RNAngle xfov, RNAngle yfov, RNLength neardist, RNLength fardist);
	R3MyCamera(const R3Point& origin, const R3Vector& towards, const R3Vector& up,
               RNAngle xfov, RNAngle yfov, RNLength neardist, RNLength fardist);
	R3MyCamera(const R3Point& origin, RNAngle pitch, RNAngle yaw, RNAngle roll,
               RNAngle xfov, RNAngle yfov, RNLength neardist, RNLength fardist);
    R3MyCamera(const R3MyCamera& camera);
    // Asymmetric camera
    R3MyCamera(const R3Point& origin, const R3Vector& towards, const R3Vector& up,
               RNAngle xfov0, RNAngle xfov1,
               RNAngle yfov0, RNAngle yfov1,
               RNLength neardist, RNLength fardist);
    
	// Property functions/operators
    const R3Point& Origin(void) const;
    const R3Vector Towards(void) const;
    const R3Vector& Backwards(void) const;
    const R3Vector& Up(void) const;
    const R3Vector Down(void) const;
    const R3Vector& Right(void) const;
    const R3Vector Left(void) const;
	const RNAngle Pitch(void) const;
	const RNAngle Yaw(void) const;
	const RNAngle Roll(void) const;
//    const RNAngle XFOV(void) const;
//    const RNAngle YFOV(void) const;
    const RNAngle LeftXFOV(void) const;
    const RNAngle RightXFOV(void) const;
    const RNAngle TopYFOV(void) const;
    const RNAngle BottomYFOV(void) const;
    const RNLength Near(void) const;
    const RNLength Far(void) const;
	const R3Triad& Triad(void) const;
	const R3CoordSystem& CoordSystem(void) const;
	const R3Halfspace Halfspace(RNDirection dir, RNDimension dim) const;
	const R4Matrix Perspective(void) const;
	const RNBoolean operator==(const R3MyCamera& camera) const;
	const RNBoolean operator!=(const R3MyCamera& camera) const;
    
    // Manipulation functions/operators
    void Reset(const R3MyCamera& _camera);
    void Reset(const R3Point& origin, const R3Vector& towards, const R3Vector& up);
    void Reposition(const R3Point& origin);
    void Reorient(const R3Vector& towards, const R3Vector& up);
    void Reorient(RNAngle pitch, RNAngle yaw, RNAngle roll);
    void Mirror(const R3Plane& plane);
    void SetCoordSystem(const R3CoordSystem& cs);
    void SetOrigin(const R3Point& origin);
    void SetTowards(const R3Vector& towards);
    void SetBackwards(const R3Vector& backwards);
	void SetUp(const R3Vector& up);
    void SetDown(const R3Vector& down);
    void SetRight(const R3Vector& right);
    void SetLeft(const R3Vector& left);
    void SetPitch(RNAngle pitch);
    void SetYaw(RNAngle pitch);
    void SetRoll(RNAngle pitch);
    void SetFOV(RNAngle xfov, RNAngle yfov);
    void SetXFOV(RNAngle xfov);
    void SetYFOV(RNAngle yfov);
    void SetFOV(RNAngle xfov0, RNAngle xfov1, RNAngle yfov0, RNAngle yfov1);
    void SetXFOV(RNAngle xfov0, RNAngle xfov1);
    void SetYFOV(RNAngle yfov0, RNAngle yfov1);
    void SetNear(RNLength neardist);
    void SetFar(RNLength fardist);
    
    // More manipulation functions/operators
    void Translate(const R3Vector& translation);
    void Rotate(const R3Vector& axis, RNAngle dtheta);
    void RotatePitch(RNAngle dpitch);
    void RotateYaw(RNAngle dyaw);
    void RotateRoll(RNAngle droll);
	void Transform(const R3Transformation& transformation);
    
    // Draw functions/operators
    void Load(RNBoolean select_mode = FALSE) const;
    void Draw(void) const;
    void Outline(void) const;
    
private:
    R3CoordSystem cs;
    RNAngle xfov[2], yfov[2];
    RNScalar neardist, fardist;
};



/* Public variables */

extern RNLength R3default_mycamera_xfov;
extern RNLength R3default_mycamera_yfov;
extern R3Vector R3default_mycamera_towards_vector;
extern R3Vector R3default_mycamera_up_vector;
extern R3Vector R3default_mycamera_right_vector;
extern R3MyCamera R3default_mycamera;



/* Inline functions */

inline const R3Point& R3MyCamera::
Origin(void) const
{
    // Return origin
    return cs.Origin();
}



inline const R3Vector R3MyCamera::
Towards(void) const
{
    // Return towards vector
    return -(cs.Axes().Axis(RN_Z));
}



inline const R3Vector& R3MyCamera::
Backwards(void) const
{
    // Return backwards vector
    return cs.Axes().Axis(RN_Z);
}



inline const R3Vector& R3MyCamera::
Up(void) const
{
    // Return up vector
    return cs.Axes().Axis(RN_Y);
}



inline const R3Vector R3MyCamera::
Down(void) const
{
    // Return down vector
    return -Up();
}



inline const R3Vector& R3MyCamera::
Right(void) const
{
    // Return right vector
    return cs.Axes().Axis(RN_X);
}



inline const R3Vector R3MyCamera::
Left(void) const
{
    // Return left vector
    return -Right();
}



//inline const RNAngle R3MyCamera::
//XFOV(void) const
//{
//    // Return x field of view
//    return xfov;
//}
//
//
//
//inline const RNAngle R3MyCamera::
//YFOV(void) const
//{
//    // Return y field of view
//    return yfov;
//}

inline const RNAngle R3MyCamera::
LeftXFOV(void) const
{
    return xfov[0];
}

inline const RNAngle R3MyCamera::
RightXFOV(void) const
{
    return xfov[1];
}

inline const RNAngle R3MyCamera::
BottomYFOV(void) const
{
    return yfov[0];
}

inline const RNAngle R3MyCamera::
TopYFOV(void) const
{
    return yfov[1];
}


inline const RNLength R3MyCamera::
Near(void) const
{
    // Return near plane distance
    return neardist;
}



inline const RNLength R3MyCamera::
Far(void) const
{
    // Return far plane distance
    return fardist;
}



inline const R3Triad& R3MyCamera::
Triad(void) const
{
    // Return triad
    return cs.Axes();
}



inline const R3CoordSystem& R3MyCamera::
CoordSystem(void) const
{
    // Return coordinate system
    return cs;
}



inline const RNBoolean R3MyCamera::
operator!=(const R3MyCamera& camera) const
{
    // Return whether camera is not equal
    return !(*this == camera);
}



inline void R3MyCamera::
SetBackwards(const R3Vector& backwards)
{
    // Set backwards vector
    SetTowards(-backwards);
}



inline void R3MyCamera::
SetDown(const R3Vector& down)
{
    // Set down vector
    SetUp(-down);
}



inline void R3MyCamera::
SetLeft(const R3Vector& left)
{
    // Set left vector
    SetRight(-left);
}



inline void R3MyCamera::
RotatePitch(RNAngle dpitch)
{
    // Rotate around up vector
    // Rotate(Up(), dpitch);
    Rotate(R3default_camera_up_vector, dpitch);
}



inline void R3MyCamera::
RotateYaw(RNAngle dyaw)
{
    // Rotate around left vector
    Rotate(Right(), dyaw);
}



inline void R3MyCamera::
RotateRoll(RNAngle droll)
{
    // Rotate around towards vector
    Rotate(Towards(), droll);
}



inline void R3MyCamera::
Outline(void) const
{
    // Draw camera
    Draw();
}

#endif


