#include "R3MyGraphics/R3MyGraphics.h"
#include <map>
using namespace std;
/* Member functions */

R3MyModel::
R3MyModel(void)
:triangles(NULL),
material_list(),
display_list(0),
model_name(""),
vertices_list(NULL),
texcoords_list(NULL),
normals_list(NULL),
vertices_indices_list(NULL),
per_material_indices_list(NULL)
{}

R3MyModel::
~R3MyModel(void)
{
    // delete display list
    if (display_list > 0) {
        glDeleteLists(display_list, 1);
    }
    // delete per_material_indices_list
    for (int i=0; i<NMaterials(); i++) {
        delete[] per_material_indices_list[i];
    }
    delete[] per_material_indices_list;
    // delete triangles
    if (triangles) delete triangles;
    // delete materials
    // ClearMaterials();
    // delete lists
    DeleteLists();
}

RNBoolean R3MyModel::
Intersects(const R3Ray& ray, 
           R3Point *hit_point, R3Vector *hit_normal, 
           RNScalar *hit_t) const
{
    // Check triangles
    if (!triangles) return FALSE;
    
    // Intersect with shape
    return triangles->Intersects(ray, hit_point, hit_normal, hit_t);
}

void R3MyModel::
Draw(const R3DrawFlags draw_flags) const
{
    // Check triangles
    if (!triangles) return;
    if (triangles->NTriangles() == 0) return;
	glEnable(GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    // Draw by materials
    for (int i=0; i<NMaterials(); i++)
    {
        Material(i)->Draw();
        int len = num_indices_per_material[i];
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        glVertexPointer(3, GL_FLOAT, 0, vertices_list);
        glTexCoordPointer(2, GL_FLOAT, 0, texcoords_list);
        glNormalPointer(GL_FLOAT, 0, normals_list);
        
        glDrawElements(GL_TRIANGLES, len, GL_UNSIGNED_INT, per_material_indices_list[i]);
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
    }
}

void R3MyModel::
DrawWithoutMaterial(const R3DrawFlags draw_flags) const
{
    // Check triangles
    if (!triangles) return;
    if (triangles->NTriangles() == 0) return;
	glEnable(GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    // Draw by materials
    static float material[] = {1, 1, 1, 1};
    material[0] = 0;
    material[1] = 0.5;
    material[2] = 1.0;
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
    for (int i=0; i<NMaterials(); i++)
    {
        int len = num_indices_per_material[i];
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        glVertexPointer(3, GL_FLOAT, 0, vertices_list);
        glTexCoordPointer(2, GL_FLOAT, 0, texcoords_list);
        glNormalPointer(GL_FLOAT, 0, normals_list);
        
        glDrawElements(GL_TRIANGLES, len, GL_UNSIGNED_INT, per_material_indices_list[i]);
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
    }
}

void R3MyModel::
Transform(const R3Transformation& transformation)
{
    triangles->Transform(transformation);
}


////////////////////////////////////////////////////////////////////////
// I/O FUNCTIONS
////////////////////////////////////////////////////////////////////////
void R3MyModel::
UpdateQuickLists(void)
{
    DeleteLists();
    GenerateLists();
    LoadMaterials();
}

int R3MyModel::
ReadFile(const char *filename)
{
    // Parse input filename extension
    const char *extension;
    if (!(extension = strrchr(filename, '.'))) {
        printf("Filename %s has no extension (e.g., .obj, .off)\n", filename);
        return 0;
    }
    
    // Read file of appropriate type
    if (!strncmp(extension, ".obj", 4)) {
        if (!ReadObjFile(filename)) return 0;
    }
    else if (!strncmp(extension, ".off", 4)) {
        if (!ReadOffFile(filename)) return 0;
    }
    else {
        RNFail("Unable to read file %s (unrecognized extension: %s)\n", filename, extension);
        return 0;
    }
    UpdateQuickLists();
    // Return success
    return 1;
}

int R3MyModel::
WriteFile(const char *filename) const
{
    // Parse input filename extension
    const char *extension;
    if (!(extension = strrchr(filename, '.'))) {
        printf("Filename %s has no extension (e.g., .obj)\n", filename);
        return 0;
    }
    
    // Write file of appropriate type
    if (!strncmp(extension, ".obj", 4)) {
        if (!WriteObjFile(filename)) return 0;
    }
    else {
        RNFail("Unable to write file %s (unrecognized extension: %s)\n", filename, extension);
        return 0;
    }
    
    // Return success
    return 1;
}

int R3MyModel::
ReadObjMtlFile(const char* filename, const char* texture_dirname)
{
    printf("R3MyModel::ReadObjMtlFile import %s, with texture in %s ...\n",
           filename, texture_dirname);
    fflush(stdout);
    // clear material list
    ClearMaterials();
    // Open file
    FILE *fp = fopen(filename, "r");
    if (!fp) {
        RNFail("Unable to open file %s", filename);
        return 0;
    }
    // Parse file
    char buffer[1024];
    int line_count = 0;
    R3Brdf *brdf = NULL;
    R2MyTexture *texture = NULL;
    R3MyMaterial *material = NULL;
    while (fgets(buffer, 1023, fp)) {
        line_count++;
        // skip white space
        char *bufferp = buffer;
        while (isspace(*bufferp)) bufferp++;
        // skip blank lines and comments
        if (*bufferp == '#') continue;
        if (*bufferp == '\0') continue;
        // get keyword
        char keyword[80];
        if (sscanf(bufferp, "%s", keyword) != 1) {
            RNFail("Syntax error on line %d in file %s", line_count, filename);
            return 0;
        }
        // check keyword
        if (!strcmp(keyword, "newmtl")) {
            char name[1024];
            if (sscanf(bufferp, "%s%s", keyword, name) != (unsigned int) 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // create new material
            texture = NULL;
            brdf = new R3Brdf();
            material = new R3MyMaterial(brdf, texture, name);
            material_list.push_back(material);
        }
        else if (!strcmp(keyword, "Ka")) {
            double r, g, b;
            if (sscanf(bufferp, "%s%lf%lf%lf", keyword, &r, &g, &b) != (unsigned int) 4) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // ambient component
            if (material && brdf) {
                brdf->SetAmbient(RNRgb(r, g, b));
                material->Update();
            }
        }
        else if (!strcmp(keyword, "Kd")) {
            double r, g, b;
            if (sscanf(bufferp, "%s%lf%lf%lf", keyword, &r, &g, &b) != (unsigned int) 4) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // diffuse component
            if (material && brdf) {
                brdf->SetDiffuse(RNRgb(r, g, b));
                material->Update();
            }
        }
        else if (!strcmp(keyword, "Ks")) {
            double r, g, b;
            if (sscanf(bufferp, "%s%lf%lf%lf", keyword, &r, &g, &b) != (unsigned int) 4) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // specular component
            if (material && brdf) {
                brdf->SetSpecular(RNRgb(r, g, b));
                material->Update();
            }
        }
        else if (!strcmp(keyword, "Ns")) {
            double ns;
            if (sscanf(bufferp, "%s%lf", keyword, &ns) != (unsigned int) 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // shiness componnet
            if (material && brdf) {
                brdf->SetShininess(ns);
                material->Update();
            }
        }
        else if (!strcmp(keyword, "Ni")) {
            double index_of_refraction;
            if (sscanf(bufferp, "%s%lf", keyword, &index_of_refraction) != (unsigned int) 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // index of reflection
            if (material && brdf) {
                brdf->SetIndexOfRefraction(index_of_refraction);
                material->Update();
            }
        }
        // the stanford's file has a wrong d.
        else if (!strcmp(keyword, "d")) {
            double transparency;
            if (sscanf(bufferp, "%s%lf", keyword, &transparency) != (unsigned int) 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // set opacity/transparency
            if (material && brdf) {
                brdf->SetOpacity(1-transparency);
                material->Update();
            }
        }
        else if (!strcmp(keyword, "map_Kd")) {
            char texture_name[1024];
            if (sscanf(bufferp, "%s%s", keyword, texture_name) != (unsigned int) 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            // texture for diffuse component
            if (material) {
                string texture_filename = string(texture_dirname) + "/" + texture_name;
                R2MyImage *image = new R2MyImage();
                // check for tiff
                char * ext = strrchr(texture_name, '.');
                if (ext && !strcmp(ext+1, "tif")) {
                    printf("Ignore *.tif (%s)!\n", texture_name);
                }
                else {
                    if (!image->Read(texture_filename.c_str())) return 0;
                }
                R2MyTexture *texture = new R2MyTexture(image);
                material->SetTexture(texture);
                material->Update();
            }
        }
    }
    
    // Close file
    fclose(fp);
    
    // Return success
    return 1;
}

// parse integers from a/b/c
static void ParseTriplets(char* s, int* i1, int* i2, int* i3)
{
    char * p1 = strchr(s, '/');
    if (!p1)
    {
        *i1 = atoi(s);
        *i2 = -1;
        *i3 = -1;
        return;
    }
    *p1 = 0; *i1 = atoi(s);
    *p1 = '/';
    char * p2 = strchr(p1+1, '/');
    if (!p2)
    {
        *i2 = atoi(p1+1);
        *i3 = -1;
        return;
    }
    *p2 = 0; *i2 = atoi(p1+1);
    *p2 = '/';
    *i3 = atoi(p2+1);
}

int R3MyModel::
ReadOffFile(const char* filename)
{
    // Open file
    FILE *fp;
    if (!(fp = fopen(filename, "r"))) {
        fprintf(stderr, "Unable to open file %s\n", filename);
        return 0;
    }
    
    // Read file
    int nverts = 0;
    int nfaces = 0;
    int nedges = 0;
    int line_count = 0;
    int face_count = 0;
    char buffer[1024];
    char header[64];
    R3MyMaterial *material = &R3Mydefault_material;
    R3TriangleVertex* vertex = NULL;
    RNArray<R3TriangleVertex*> verts;
    RNArray<R3Triangle*> tris;
    // single material
    material_list.push_back(material);
    while (fgets(buffer, 1023, fp)) {
        // Increment line counter
        line_count++;
        
        // Skip white space
        char *bufferp = buffer;
        while (isspace(*bufferp)) bufferp++;
        
        // Skip blank lines and comments
        if (*bufferp == '#') continue;
        if (*bufferp == '\0') continue;
        
        // Check section
        if (nverts == 0) {
            // Read header keyword
            if (strstr(bufferp, "OFF")) {
                // Check if counts are on first line
                int tmp;
                if (sscanf(bufferp, "%s%d%d%d", header, &tmp, &nfaces, &nedges) == 4) {
                    nverts = tmp;
                }
            }
            else {
                // Read counts from second line
                if ((sscanf(bufferp, "%d%d%d", &nverts, &nfaces, &nedges) != 3) || (nverts == 0)) {
                    RNFail("Syntax error reading header on line %d in file %s\n", line_count, filename);
                    fclose(fp);
                    return 0;
                }
            }
        }
        else if (verts.NEntries() < nverts) {
            
            // Read vertex coordinates
            double x, y, z;
            if (sscanf(bufferp, "%lf%lf%lf", &x, &y, &z) != 3) {
                RNFail("Syntax error with vertex coordinates on line %d in file %s\n", line_count, filename);
                fclose(fp);
                return 0;
            }
            
            // Create vertex
            vertex = new R3TriangleVertex(R3Point(x, y, z));
            verts.Insert(vertex);
        }
        else if (face_count < nfaces) {
            // Read number of vertices in face
            int face_nverts = 0;
            bufferp = strtok(bufferp, " \t");
            if (bufferp) face_nverts = atoi(bufferp);
            else {
                RNFail("Syntax error with face on line %d in file %s\n", line_count, filename);
                fclose(fp);
                return 0;
            }
            
            // Read vertex indices for face
            R3TriangleVertex *v1 = NULL;
            R3TriangleVertex *v2 = NULL;
            R3TriangleVertex *v3 = NULL;
            for (int i = 0; i < face_nverts; i++) {
                bufferp = strtok(NULL, " \t");
                if (bufferp) {
                    R3TriangleVertex *v = verts.Kth(atoi(bufferp));
                    if (!v1) v1 = v;
                    else v3 = v;
                }
                else {
                    RNFail("Syntax error with face on line %d in file %s\n", line_count, filename);
                    fclose(fp);
                    return 0;
                }
                
                // Create triangle
                if (v1 && v2 && v3 && (v1 != v2) && (v2 != v3) && (v1 != v3)) {
//                    if (!CreateFace(v1, v2, v3)) {
//                        // Must have been degeneracy (e.g., flips or three faces sharing an edge)
//                        // Remember for later processing (to preserve vertex indices)
//                        degenerate_triangle_vertices.Insert(v1);
//                        degenerate_triangle_vertices.Insert(v2);
//                        degenerate_triangle_vertices.Insert(v3);
//                    }
                    // Create first triangle
                    R3Triangle *triangle = new R3Triangle(v1, v2, v3);
                    tris.Insert(triangle);
                    triangle_materials.push_back(material);
                }
                
                // Move to next triangle
                v2 = v3;
            }
            
            // Increment counter
            face_count++;
        }
        else {
            // Should never get here
            RNFail("Found extra text starting at line %d in file %s\n", line_count, filename);
            break;
        }
    }
    
//    // Create degenerate triangles
//    for (int i = 0; i <= degenerate_triangle_vertices.NEntries()-3; i+=3) {
//        R3MeshVertex *v1 = degenerate_triangle_vertices.Kth(i+0);
//        R3MeshVertex *v2 = degenerate_triangle_vertices.Kth(i+1);
//        R3MeshVertex *v3 = degenerate_triangle_vertices.Kth(i+2);
//        if (!CreateFace(v1, v2, v3)) {
//            if (!CreateFace(v1, v3, v2)) {
//                // Note: these vertices are allocated separately, and so they will not be deleted (memory leak)
//                R3MeshVertex *v1a = CreateVertex(VertexPosition(v1));
//                R3MeshVertex *v2a = CreateVertex(VertexPosition(v2));
//                R3MeshVertex *v3a = CreateVertex(VertexPosition(v3));
//                CreateFace(v1a, v2a, v3a);
//            }
//        }
//    }
    
    // Create triangle array
    triangles = new R3TriangleArray(verts, tris);
    // Close file
    fclose(fp);
    // Return success
    return 1;
}

int R3MyModel::
ReadObjFile(const char *filename)
{
    // Shouldn't really be here, but needed to have default material
    R3InitGraphics();
    R3StopGraphics();
    
    // Open file
    FILE *fp = fopen(filename, "r");
    if (!fp) {
        RNFail("Unable to open file %s", filename);
        return 0;
    }
    
    // Determine directory name (for texture image files)
    char dirname[1024];
    strncpy(dirname, filename, 1024);
    char *endp = strrchr(dirname, '/');
    if (!endp) endp = strrchr(dirname, '\\');
    if (!endp) strcpy(dirname, ".");
    else *endp = '\0';
    
    // Read body
    char buffer[1024];
    int line_count = 0;
    R3MyMaterial *material = &R3Mydefault_material;
    R3TriangleVertex *vertex = NULL;
    RNArray<R3TriangleVertex *> verts;
    RNArray<R3TriangleVertex *> vertsActual;
    RNArray<R3Triangle *> tris;
    RNArray<R2Point *> texcoords;
    // single material
    material_list.push_back(material);
    while (fgets(buffer, 1023, fp)) {
        // Increment line counter
        line_count++;
        
        // Skip white space
        char *bufferp = buffer;
        while (isspace(*bufferp)) bufferp++;
        
        // Skip blank lines and comments
        if (*bufferp == '#') continue;
        if (*bufferp == '\0') continue;
        
        // Get keyword
        char keyword[80];
        if (sscanf(bufferp, "%s", keyword) != 1) {
            RNFail("Syntax error on line %d in file %s", line_count, filename);
            return 0;
        }
        
        // Check keyword
        if (!strcmp(keyword, "v")) {
            // Read vertex coordinates
            double x, y, z;
            if (sscanf(bufferp, "%s%lf%lf%lf", keyword, &x, &y, &z) != 4) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            
            // Create vertex
            vertex = new R3TriangleVertex(R3Point(x, y, z));
            verts.Insert(vertex);
        }
        else if (!strcmp(keyword, "vt")) {
            // Read texture coordinates
            double u, v;
            if (sscanf(bufferp, "%s%lf%lf", keyword, &u, &v) != 3) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            
            // Create tex coords
            R2Point* tex = new R2Point(u, v);
            texcoords.Insert(tex);
        }
        else if (!strcmp(keyword, "f")) {
            // Read vertex indices
            int quad = 1;
            char s1[128], s2[128], s3[128], s4[128] = { '\0' };
            if (sscanf(bufferp, "%s%s%s%s%s", keyword, s1, s2, s3, s4) != 5) {
                quad = 0;;
                if (sscanf(bufferp, "%s%s%s%s", keyword, s1, s2, s3) != 4) {
                    RNFail("Syntax error on line %d in file %s", line_count, filename);
                    continue;
                }
            }
            
            // Parse vertex indices
            int i1[3], i2[3], i3[3], i4[3] = {-1, -1, -1};
            ParseTriplets(s1, &i1[0], &i1[1], &i1[2]);
            ParseTriplets(s2, &i2[0], &i2[1], &i2[2]);
            ParseTriplets(s3, &i3[0], &i3[1], &i3[2]);
            if (quad) ParseTriplets(s4, &i4[0], &i4[1], &i4[2]);
            
            // Get vertices (new vertices instead of reusing old vertices)
            R3TriangleVertex *v1 = new R3TriangleVertex(*(verts.Kth(i1[0]-1)));
            R3TriangleVertex *v2 = new R3TriangleVertex(*(verts.Kth(i2[0]-1)));
            R3TriangleVertex *v3 = new R3TriangleVertex(*(verts.Kth(i3[0]-1)));
            R3TriangleVertex *v4 = (quad) ? new R3TriangleVertex(*(verts.Kth(i4[0]-1))) : NULL;
            // Get tex coords
            if (i1[1] >= 1 && i1[1] <= texcoords.NEntries()) v1->SetTextureCoords(*(texcoords.Kth(i1[1]-1)));
            if (i2[1] >= 1 && i2[1] <= texcoords.NEntries()) v2->SetTextureCoords(*(texcoords.Kth(i2[1]-1)));
            if (i3[1] >= 1 && i3[1] <= texcoords.NEntries()) v3->SetTextureCoords(*(texcoords.Kth(i3[1]-1)));
            if (quad && i4[1] >= 1 && i4[1] <= texcoords.NEntries()) v4->SetTextureCoords(*(texcoords.Kth(i4[1]-1)));
            
            // Check vertices
            if ((v1 == v2) || (v2 == v3) || (v1 == v3)) continue;
            if ((quad) && ((v4 == v1) || (v4 == v2) || (v4 == v3))) quad = 0;
            
            // Create first triangle
            R3Triangle *triangle = new R3Triangle(v1, v2, v3);
            tris.Insert(triangle);
            triangle_materials.push_back(material);
            vertsActual.Insert(v1);
            vertsActual.Insert(v2);
            vertsActual.Insert(v3);
            
            // Create second triangle
            if (quad) {
                R3Triangle *triangle = new R3Triangle(v1, v3, v4);
                tris.Insert(triangle);
                triangle_materials.push_back(material);
                vertsActual.Insert(v4);
            }
        }
        else if (!strcmp(keyword, "mtllib")) {
            // Read fields
            char mtllib_relative_path[1024];
            if (sscanf(bufferp, "%s%s", keyword, mtllib_relative_path) != 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            string mtllib_path = string(dirname)+"/"+mtllib_relative_path;
            string texture_path = string(dirname)+"/../textures";
            // import materials
            if (!ReadObjMtlFile(mtllib_path.c_str(), texture_path.c_str())) return 0;
        }
        else if (!strcmp(keyword, "usemtl")) {
            // Read fields
            char mtlname[1024];
            if (sscanf(bufferp, "%s%s", keyword, mtlname) != 2) {
                RNFail("Syntax error on line %d in file %s", line_count, filename);
                return 0;
            }
            
            // Find material
            material = FindMaterial(mtlname);
            if (!material) {
                fprintf(stderr, "Unable to find material %s at on line %d in file %s", mtlname, line_count, filename);
                return 0;
            }
        }
    }
    
    // Create triangle array
    triangles = new R3TriangleArray(vertsActual, tris);
    // Close file
    fclose(fp);
    // Return success
    return 1;
}

int R3MyModel::
WriteObjFile(const char *filename) const
{
    fprintf(stderr, "R3MyModel::WriteObjFile not implemented yet!\n");
    // Return success
    return 0;
}

void R3MyModel::
DeleteLists(void)
{
    delete[] vertices_list;
    delete[] texcoords_list;
    delete[] normals_list;
    delete[] vertices_indices_list;
}

void R3MyModel::
GenerateLists(void)
{
    GenerateVerticesList();
    GenerateTexcoordsList();
    GenerateNormalsList();
    GenerateVerticesIndicesList();
}

void R3MyModel::
GenerateVerticesList(void)
{
    int nfaces = NTriangles();
    vertices_list = new GLfloat[nfaces *3 * 3];
    for (int i=0; i<nfaces; i++)
    {
        R3TriangleVertex* v[3];
        v[0] = Triangle(i)->V0();
        v[1] = Triangle(i)->V1();
        v[2] = Triangle(i)->V2();
        int index[3];
        index[0] = 3*i + 0;
        index[1] = 3*i + 1;
        index[2] = 3*i + 2;
        for (int j=0; j<3; j++)
        {
            vertices_list[3*index[j] + 0] = v[j]->Position()[0];
            vertices_list[3*index[j] + 1] = v[j]->Position()[1];
            vertices_list[3*index[j] + 2] = v[j]->Position()[2];
        }
    }
}

void R3MyModel::
GenerateTexcoordsList(void)
{
    int nfaces = NTriangles();
    texcoords_list = new GLfloat[nfaces*3 * 2];
    for (int i=0; i<nfaces; i++)
    {
        R3TriangleVertex* v[3];
        v[0] = Triangle(i)->V0();
        v[1] = Triangle(i)->V1();
        v[2] = Triangle(i)->V2();
        int index[3];
        index[0] = 3*i + 0;
        index[1] = 3*i + 1;
        index[2] = 3*i + 2;
        for (int j=0; j<3; j++)
        {
            texcoords_list[2*index[j] + 0] = v[j]->TextureCoords()[0];
            texcoords_list[2*index[j] + 1] = v[j]->TextureCoords()[1];
        }
    }
}

void R3MyModel::
GenerateNormalsList(void)
{
    int nfaces = NTriangles();
    normals_list = new GLfloat[nfaces *3 * 3];
    for (int i=0; i<nfaces; i++)
    {
        R3TriangleVertex* v[3];
        v[0] = Triangle(i)->V0();
        v[1] = Triangle(i)->V1();
        v[2] = Triangle(i)->V2();
        int index[3];
        index[0] = 3*i + 0;
        index[1] = 3*i + 1;
        index[2] = 3*i + 2;
        for (int j=0; j<3; j++)
        {
            normals_list[3*index[j] + 0] = Triangle(i)->Normal()[0];
            normals_list[3*index[j] + 1] = Triangle(i)->Normal()[1];
            normals_list[3*index[j] + 2] = Triangle(i)->Normal()[2];
        }
    }
}

void R3MyModel::
GenerateVerticesIndicesList(void)
{
    int nfaces = triangles->NTriangles();
    vertices_indices_list = new GLuint[nfaces * 3];
    for (int i=0; i<nfaces; i++)
    {
        vertices_indices_list[3*i + 0] = 3*i + 0;
        vertices_indices_list[3*i + 1] = 3*i + 1;
        vertices_indices_list[3*i + 2] = 3*i + 2;
    }
}

void R3MyModel::
LoadMaterials(void)
{
    if ((int)triangle_materials.size() != NTriangles()) {
        fprintf(stderr, "R3MyModel::LoadMaterials triangle_materials.size() != NTriangles()\n");
        exit(-1);
    }
    // mapping from material to group of faces
    map<R3MyMaterial*, vector<int> > mapping;
    for (int i=0; i<NTriangles(); i++)
    {
        mapping[triangle_materials[i]].push_back(i);
    }
    
    per_material_indices_list = new GLuint*[NMaterials()];
    // follow the order of materials
    num_indices_per_material.clear();
    for (int i=0; i<NMaterials(); i++)
    {
        const vector<int>& array = mapping[Material(i)];
        per_material_indices_list[i] = new GLuint[3*array.size()];
        num_indices_per_material.push_back(3*array.size());
        for (int j=0; j<int(array.size()); j++)
        {
            per_material_indices_list[i][3*j + 0] = vertices_indices_list[3*array[j] + 0];
            per_material_indices_list[i][3*j + 1] = vertices_indices_list[3*array[j] + 1];
            per_material_indices_list[i][3*j + 2] = vertices_indices_list[3*array[j] + 2];
        }
    }
    
}

void R3MyModel::
ClearMaterials(void)
{
    for (int i=0; i<NMaterials(); i++) {
        delete material_list[i];
    }
    material_list.clear();
    triangle_materials.clear();
}

R3MyMaterial * R3MyModel::
FindMaterial(const char *name) const
{
    for (int i=0; i<NMaterials(); i++) {
        if (!strcmp(Material(i)->Name(), name)) return Material(i);
    }
    return NULL;
}








