#ifndef R3_MY_MODEL_H
#define R3_MY_MODEL_H
#include <vector>
#include <string>
class R3MyModel {
public:
    // Constructor functions
    R3MyModel(void);
    ~R3MyModel(void);
    
    // Properties
    R3Box BBox(void) const;
    R3Point Centroid(void) const;
    
    // Access 
    int NVertices(void) const;
    R3TriangleVertex *Vertex(int k) const;
    int NTriangles(void) const;
    R3Triangle *Triangle(int k) const;
    int NMaterials(void) const;
    R3MyMaterial *Material(int k) const;
    
    // Ray intersection functions
    RNBoolean Intersects(const R3Ray& ray, 
                         R3Point *hit_point = NULL, R3Vector *hit_normal = NULL, 
                         RNScalar *hit_t = NULL) const;
    
    // Draw functions
    void Draw(const R3DrawFlags draw_flags = R3_DEFAULT_DRAW_FLAGS) const;
    void DrawWithoutMaterial(const R3DrawFlags draw_flags = R3_DEFAULT_DRAW_FLAGS) const;
    void Transform(const R3Transformation& transformation);
    
    // I/O functions
    int ReadFile(const char *filename);
    int ReadOffFile(const char* filename);
    int ReadObjFile(const char *filename);
    int ReadObjMtlFile(const char *filename, const char* texture_dirname);
    int WriteFile(const char *filename) const;
    int WriteObjFile(const char *filename) const;
    void UpdateQuickLists(void);
protected:
    void DeleteLists(void);
    void GenerateLists(void);
    void GenerateVerticesList(void);
    void GenerateTexcoordsList(void);
    void GenerateNormalsList(void);
    void GenerateVerticesIndicesList(void);
    void LoadMaterials(void);
    void ClearMaterials(void);
    R3MyMaterial* FindMaterial(const char* mtlname) const;
protected:
    // triangles
    R3TriangleArray *triangles;
    // material list
    std::vector<R3MyMaterial*> material_list;
    std::vector<R3MyMaterial*> triangle_materials;
    unsigned int display_list;
    std::string model_name;
    // opengl quick reference
    GLfloat* vertices_list;
    GLfloat* texcoords_list;
    GLfloat* normals_list;
    GLuint* vertices_indices_list;
    GLuint** per_material_indices_list;
    std::vector<int> num_indices_per_material;
};



/* Inline functions */

inline R3Box R3MyModel::
BBox(void) const
{
    // Return bounding box
    if (!triangles) return R3null_box;
    return triangles->BBox();
}



inline R3Point R3MyModel::
Centroid(void) const
{
    // Return centroid
    if (!triangles) return R3zero_point;
    return triangles->Centroid();
}



inline int R3MyModel::
NVertices(void) const
{
    // Return number of vertices
    if (!triangles) return 0;
    return triangles->NVertices();
}



inline R3TriangleVertex *R3MyModel::
Vertex(int k) const
{
    // Return Kth vertex
    if (!triangles) return NULL;
    return triangles->Vertex(k);
}



inline int R3MyModel::
NTriangles(void) const
{
    // Return number of triangles
    if (!triangles) return 0;
    return triangles->NTriangles();
}



inline R3Triangle *R3MyModel::
Triangle(int k) const
{
    // Return Kth triangle
    if (!triangles) return NULL;
    return triangles->Triangle(k);
}



inline int R3MyModel::
NMaterials(void) const
{
    // Return number of materials
    return (int)material_list.size();
}


inline R3MyMaterial *R3MyModel::
Material(int k) const
{
    // Return Kth material
    return material_list[k];
}
#endif


