#include "R3MyGraphics/R3MyGraphics.h"
#include "Utils/Utils.h"
#include <fstream>
#define POS_X 0
#define NEG_X 1
#define POS_Y 2
#define NEG_Y 3
#define POS_Z 4
#define NEG_Z 5

using namespace std;

R3IntBox::
R3IntBox(int minx, int miny, int minz, int maxx, int maxy, int maxz)
:minX(minx), minY(miny), minZ(minz), maxX(maxx), maxY(maxy), maxZ(maxz)
{}

int R3IntBox::
Volume(void) const
{
    return (maxX-minX+1)*(maxY-minY+1)*(maxZ-minZ+1);
}

Voxelization::
Voxelization(void)
:m_iXRange(0),m_iYRange(0),m_iZRange(0), m_fVoxelSize(0),
m_fXMin(0), m_fYMin(0), m_fZMin(0)
{}

void Voxelization::
Set(const R3Box& box, double voxelSize)
{
    m_fVoxelSize = voxelSize;
    m_fXMin = box.XMin();
    m_fYMin = box.YMin();
    m_fZMin = box.ZMin();
    m_iXRange = int(box.XLength()/m_fVoxelSize)+1;
    m_iYRange = int(box.YLength()/m_fVoxelSize)+1;
    m_iZRange = int(box.ZLength()/m_fVoxelSize)+1;
    InitializeOccupyFlags();
}

void Voxelization::
SetOccupyFlag(int x, int y, int z)
{
    MarkOccupyFlag(x, y, z, 1);
}

void Voxelization::
MarkOccupyFlag(int x, int y, int z, int flag)
{
    assert(x>=0&&x<m_iXRange&&y>=0&&y<m_iYRange&&z>=0&&z<m_iZRange);
    m_aOccupyFlags[(x*m_iYRange+y)*m_iZRange+z] = flag;
}

int Voxelization::
GetOccupyFlag(int x, int y, int z) const
{
    assert(x>=0&&x<m_iXRange&&y>=0&&y<m_iYRange&&z>=0&&z<m_iZRange);
    return m_aOccupyFlags[(x*m_iYRange+y)*m_iZRange+z];
}

void Voxelization::
InitializeOccupyFlags(void)
{
    m_aOccupyFlags = vector<int>(m_iXRange*m_iYRange*m_iZRange, 0);
}

bool Voxelization::
IsEmpty(void) const
{
    return (m_iXRange == 0);
}

void Voxelization::
Draw(void) const
{
    glDisable(GL_LIGHTING);
    glLineWidth(m_fVoxelSize*0.05);
    glColor3f(0, 0, 0);
    for (int x=0; x<m_iXRange; x++) {
        for (int y=0; y<m_iYRange; y++) {
            for (int z=0; z<m_iZRange; z++) {
                if (GetOccupyFlag(x, y, z)) GetCellBox(x, y, z).Outline();
            }
        }
    }
}

const R3Box Voxelization::
GetCellBox(int x, int y, int z) const
{
    double xmin = m_fXMin+x*m_fVoxelSize;
    double ymin = m_fYMin+y*m_fVoxelSize;
    double zmin = m_fZMin+z*m_fVoxelSize;
    return R3Box(xmin, ymin, zmin, xmin+m_fVoxelSize, ymin+m_fVoxelSize, zmin+m_fVoxelSize);
}

// this is buggy
int Voxelization::
Import(R3Mesh* pMesh, double voxelSize)
{
    printf("Voxelization::Import %s ...\n", pMesh->Name());
    // set basics
    Set(pMesh->BBox(), voxelSize);
    // set surface occupancy
    char prefix[] = "    Get surface occupancy";
    for (int faceIdx=0; faceIdx<pMesh->NFaces(); faceIdx++) {
        FileSystemUtil::DrawProgressBar(faceIdx*1.0/pMesh->NFaces(), 20, prefix);
        R3MeshFace* pFace = pMesh->Face(faceIdx);
        R3Box faceBox = pMesh->FaceBBox(pFace);
        R3Plane facePlane = pMesh->FacePlane(pFace);
        // triangle
        R3MeshVertex* pVertex[3];
        pVertex[0] = pMesh->VertexOnFace(pFace, 0);
        R3TriangleVertex v0(pMesh->VertexPosition(pVertex[0]));
        pVertex[1] = pMesh->VertexOnFace(pFace, 1);
        R3TriangleVertex v1(pMesh->VertexPosition(pVertex[1]));
        pVertex[2] = pMesh->VertexOnFace(pFace, 2);
        R3TriangleVertex v2(pMesh->VertexPosition(pVertex[2]));
        R3Triangle faceTri(&v0, &v1, &v2);
        for (int xi=int((faceBox.XMin()-m_fXMin)/m_fVoxelSize-1);
             xi<=int((faceBox.XMax()-m_fXMin)/m_fVoxelSize);
             xi++) {
            if (xi<0 || xi>=m_iXRange) continue;
            for (int yi=int((faceBox.YMin()-m_fYMin)/m_fVoxelSize-1);
                 yi<=int((faceBox.YMax()-m_fYMin)/m_fVoxelSize);
                 yi++) {
                if (yi<0 || yi>=m_iYRange) continue;
                for (int zi=int((faceBox.ZMin()-m_fZMin)/m_fVoxelSize-1);
                     zi<=int((faceBox.ZMax()-m_fZMin)/m_fVoxelSize);
                     zi++) {
                    if (zi<0 || zi>=m_iZRange) continue;
                    R3Box cell = GetCellBox(xi, yi, zi);
                    if (!R3Intersects(facePlane, cell)) continue;
                    // if cell contains any vertex or the triangle intersects the box
                    if (R3Contains(cell, pMesh->VertexPosition(pVertex[0])) ||
                        R3Contains(cell, pMesh->VertexPosition(pVertex[1])) ||
                        R3Contains(cell, pMesh->VertexPosition(pVertex[2])) ||
                        R3Intersects(faceTri, cell)) {
                        SetOccupyFlag(xi, yi, zi);
                    }
                }
            }
        }
    }
    FileSystemUtil::DrawProgressBar(1.0, 20, prefix);
    printf("\n");
    // set interior occupancy using connected components
    SetInteriorOccupancy();
    return 1;
}

int Voxelization::
Import(PointCloud* pCloud, double voxelSize)
{
    printf("Voxelization::Import %s ...\n", pCloud->Name().c_str());
    // set basics
    Set(pCloud->BBox(), voxelSize);
    // set surface occupancy
    char prefix[] = "    Get surface occupancy";
    for (int i=0; i<pCloud->NPoints(); i++) {
        FileSystemUtil::DrawProgressBar(i*1.0/pCloud->NPoints(), 20, prefix);
        R3Point pos = pCloud->Vertex(i)->Position();
        SetOccupyFlag( (pos[0] - m_fXMin)/m_fVoxelSize,
                      (pos[1] - m_fYMin)/m_fVoxelSize,
                      (pos[2] - m_fZMin)/m_fVoxelSize);
    }
    FileSystemUtil::DrawProgressBar(1.0, 20, prefix);
    printf("\n");
    // set interior occupancy using connected components
    printf("    Get interior occupancy ...");
    SetInteriorOccupancy();
    printf("\n");
    return 1;
}

void Voxelization::
SetInteriorOccupancy(void)
{
    for (int y=0; y<m_iYRange; y++) {
        for (int z=0; z<m_iZRange; z++) {
            // if the cell has been occupied or visited before
            if (GetOccupyFlag(0, y, z) > 0) continue;
            FloodFill(0, y, z);
        }
    }
    for (int x=0; x<m_iXRange; x++) {
        for (int z=0; z<m_iZRange; z++) {
            if (GetOccupyFlag(x, 0, z) > 0) continue;
            FloodFill(x, 0, z);
        }
    }
    for (int x=0; x<m_iXRange; x++) {
        for (int y=0; y<m_iYRange; y++) {
            if (GetOccupyFlag(x, y, 0) > 0) continue;
            FloodFill(x, y, 0);
        }
    }
    // change all 0 to 1 and all 2 to 0
    int zeroCount = 0;
    for (int i=0; i<(int)m_aOccupyFlags.size(); i++) {
        if (m_aOccupyFlags[i] == 0) {
            m_aOccupyFlags[i] = 1;
            zeroCount ++;
        }
        else if (m_aOccupyFlags[i] == 2) m_aOccupyFlags[i] = 0;
    }
    printf("    %d/%d = %g%% interior cells are newly added!\n", zeroCount, (int)m_aOccupyFlags.size(),
           zeroCount*100.0/m_aOccupyFlags.size());
}

void Voxelization::
FloodFill(int x, int y, int z)
{
    vector<int> queue[3];
    queue[0].push_back(x);
    queue[1].push_back(y);
    queue[2].push_back(z);
    int pointer = 0;
    while (pointer < (int)queue[0].size()) {
        int cx = queue[0][pointer];
        int cy = queue[1][pointer];
        int cz = queue[2][pointer];
        pointer ++;
        if (GetOccupyFlag(cx, cy, cz) > 0) continue;
        // mark the current cell
        MarkOccupyFlag(cx, cy, cz, 2);
        // add neighbors to queue
        if (cx-1>=0) {
            queue[0].push_back(cx-1);
            queue[1].push_back(cy);
            queue[2].push_back(cz);
        }
        if (cx+1<m_iXRange) {
            queue[0].push_back(cx+1);
            queue[1].push_back(cy);
            queue[2].push_back(cz);
        }
        if (cy-1>=0) {
            queue[0].push_back(cx);
            queue[1].push_back(cy-1);
            queue[2].push_back(cz);
        }
        if (cy+1<m_iYRange) {
            queue[0].push_back(cx);
            queue[1].push_back(cy+1);
            queue[2].push_back(cz);
        }
        if (cz-1>=0) {
            queue[0].push_back(cx);
            queue[1].push_back(cy);
            queue[2].push_back(cz-1);
        }
        if (cz+1<m_iZRange) {
            queue[0].push_back(cx);
            queue[1].push_back(cy);
            queue[2].push_back(cz+1);
        }
    }
}

// format
// line 1: xmin, ymin, zmin
// line 2: voxel-size
// line 3: xmax-int, ymax-int, zmax-int
// Line 4 to 3+N: XINT YINT ZINT (the 3D index of n-th occupied cell)
int Voxelization::
ReadFile(const char* filename)
{
    ifstream fin(filename);
    if (!fin) {
        fprintf(stderr, "Voxelization::ReadFile unable to open %s!\n", filename);
        return 0;
    }
    printf("Voxelization::ReadFile %s ...\n", filename);
    // line 1
    fin >> m_fXMin >> m_fYMin >> m_fZMin;
    // line 2
    fin >> m_fVoxelSize;
    // line 3
    fin >> m_iXRange >> m_iYRange >> m_iZRange;
    InitializeOccupyFlags();
    // line 4 and more
    int x, y, z;
    int count = 0;
    while (fin >> x >> y >> z) {
        SetOccupyFlag(x, y, z);
        count ++;
    }
    fin.close();
    printf("    occupy rate %d/%d = %g%%\n", count, (int)m_aOccupyFlags.size(), 100.0*count/m_aOccupyFlags.size());
    return 1;
}

int Voxelization::
WriteFile(const char* filename) const
{
    ofstream fout(filename);
    if (!fout) {
        fprintf(stderr, "Voxelization::WriteFile unable to open %s!\n", filename);
        return 0;
    }
    printf("Voxelization::WriteFile %s ...\n", filename);
    // line 1
    fout << m_fXMin << " " << m_fYMin << " " << m_fZMin << endl;
    // line 2
    fout << m_fVoxelSize << endl;
    // line 3
    fout << m_iXRange << " " << m_iYRange << " " << m_iZRange << endl;
    // line 4 and more
    int count = 0;
    for (int x=0; x<m_iXRange; x++) {
        for (int y=0; y<m_iYRange; y++) {
            for (int z=0; z<m_iZRange; z++) {
                if (GetOccupyFlag(x, y, z)) {
                    fout << x << " " << y << " " << z << endl;
                    count ++;
                }
            }
        }
    }
    fout.close();
    printf("    occupy rate %d/%d = %g%%\n", count, (int)m_aOccupyFlags.size(), 100.0*count/m_aOccupyFlags.size());
    return 1;
}

void Voxelization::
InitializeCuboids(double occupyRate, double stopOccupyRatio)
{
    printf("Voxelization::InitializeCuboids ...\n");
    // clear cuboid array and occupy flags
    m_aCuboidArray.clear();
    // count how many cells are occupied
    int occupiedVolume = 0;
    for (int i=0; i<(int)m_aOccupyFlags.size(); i++) {
        if (m_aOccupyFlags[i]>0) occupiedVolume ++;
    }
    int stopVolume = stopOccupyRatio*occupiedVolume;
    // each iteration only selects the largest cuboid
    while (true) {
        // flag = 1 means the occupied cell hasn't been visited yet
        int index = FindMarkedItem(1, 0);
        R3IntBox largestBox(0, 0, 0, 0, 0, 0);
        int largestVolume = 0;
        while (index >= 0) {
            R3IntBox newCand = GrowCuboid(index, occupyRate);
            index = FindMarkedItem(1, index+1);
            if (newCand.Volume() > largestVolume) {
                largestVolume = newCand.Volume();
                largestBox = newCand;
            }
        }
        // if largestVolume is too small, then stop
        if (largestVolume < stopVolume) break;
        m_aCuboidArray.push_back(largestBox);
        printf("    select a cuboid with volume = %d\n", largestVolume);
        // mark all cells taken by largestBox as 0
        MarkOccupyFlagForCuboid(largestBox, 0);
        // mark the rest occupied cells as 1
        for (int i=0; i<(int)m_aOccupyFlags.size(); i++) {
            if (m_aOccupyFlags[i]>0) m_aOccupyFlags[i] = 1;
        }
    }
    
}

const vector<R3Box> Voxelization::
GetCuboids(void) const
{
    vector<R3Box> cuboids;
    for (int i=0; i<(int)m_aCuboidArray.size(); i++)
        cuboids.push_back(GetCuboid(m_aCuboidArray[i]));
    return cuboids;
}

const R3Box Voxelization::
GetCuboid(const R3IntBox& cuboid) const
{
    R3Box minBox = GetCellBox(cuboid.minX, cuboid.minY, cuboid.minZ);
    R3Box maxBox = GetCellBox(cuboid.maxX, cuboid.maxY, cuboid.maxZ);
    return R3Box(minBox.XMin(), minBox.YMin(), minBox.ZMin(),
                 maxBox.XMax(), maxBox.YMax(), maxBox.ZMax());
}

int Voxelization::
FindMarkedItem(int flag, int startIdx) const
{
    for (int i=startIdx; i<(int)m_aOccupyFlags.size(); i++) {
        if (m_aOccupyFlags[i] == flag) return i;
    }
    return -1;
}

void Voxelization::
ConvertIndexTo3DIndices(int index, int* px, int* py, int* pz) const
{
    *pz = index % m_iZRange;
    index /= m_iZRange;
    *py = index % m_iYRange;
    index /= m_iYRange;
    *px = index;
}

const R3IntBox Voxelization::
GrowCuboid(int index, double occupyRate)
{
    int occupyCount = 0;
    int totalCount = 0;
    // initialization
    int minX=0, minY=0, minZ=0;
    ConvertIndexTo3DIndices(index, &minX, &minY, &minZ);
    int maxX=minX, maxY=minY, maxZ=minZ;
    occupyCount ++; totalCount ++;
    while (true) {
        int bestFinalOccupyCount = occupyCount;
        int bestFinalTotalCount = totalCount;
        // we pick the attempt that is over occupyRate and produce the highest occupyCount
        int bestMoveIndex = -1;
        // attempt expansion
        for (int i=0; i<6; i++) {
            int newOccupyCount = 0;
            int newTotalCount = 0;
            GetChangeOfMove(minX, minY, minZ, maxX, maxY, maxZ,
                            i, &newOccupyCount, &newTotalCount);
            double attemptRate = double(newOccupyCount+occupyCount)/(newTotalCount+totalCount);
            if (attemptRate < occupyRate) continue;
            if (newOccupyCount+occupyCount > bestFinalOccupyCount) {
                bestFinalOccupyCount = newOccupyCount+occupyCount;
                bestFinalTotalCount = newTotalCount+totalCount;
                bestMoveIndex = i;
            }
        }
        if (bestMoveIndex == -1) break;
        // update
        occupyCount = bestFinalOccupyCount;
        totalCount = bestFinalTotalCount;
        if (bestMoveIndex == POS_X) {
            maxX ++;
            for (int y=minY; y<=maxY; y++) {
                for (int z=minZ; z<=maxZ; z++) {
                    if (GetOccupyFlag(maxX, y, z) == 1)
                        MarkOccupyFlag(maxX, y, z, 2);
                }
            }
        }
        else if (bestMoveIndex == NEG_X) {
            minX --;
            for (int y=minY; y<=maxY; y++) {
                for (int z=minZ; z<=maxZ; z++) {
                    if (GetOccupyFlag(minX, y, z) == 1)
                        MarkOccupyFlag(minX, y, z, 2);
                }
            }
        }
        else if (bestMoveIndex == POS_Y) {
            maxY ++;
            for (int x=minX; x<=maxX; x++) {
                for (int z=minZ; z<=maxZ; z++) {
                    if (GetOccupyFlag(x, maxY, z) == 1)
                        MarkOccupyFlag(x, maxY, z, 2);
                }
            }
        }
        else if (bestMoveIndex == NEG_Y) {
            minY --;
            for (int x=minX; x<=maxX; x++) {
                for (int z=minZ; z<=maxZ; z++) {
                    if (GetOccupyFlag(x, minY, z) == 1)
                        MarkOccupyFlag(x, minY, z, 2);
                }
            }
        }
        else if (bestMoveIndex == POS_Z) {
            maxZ ++;
            for (int x=minX; x<=maxX; x++) {
                for (int y=minY; y<=maxY; y++) {
                    if (GetOccupyFlag(x, y, maxZ) == 1)
                        MarkOccupyFlag(x, y, maxZ, 2);
                }
            }
        }
        else {
            minZ --;
            for (int x=minX; x<=maxX; x++) {
                for (int y=minY; y<=maxY; y++) {
                    if (GetOccupyFlag(x, y, minZ) == 1)
                        MarkOccupyFlag(x, y, minZ, 2);
                }
            }
        }
    }
    return R3IntBox(minX, minY, minZ, maxX, maxY, maxZ);
}

void Voxelization::
GetChangeOfMove(int minX, int minY, int minZ, int maxX, int maxY, int maxZ,
                int expansionDir, int* pNewOccupy, int* pNewTotal) const
{
    (*pNewOccupy) = 0;
    (*pNewTotal) = 0;
    int bigValue = 100000000;
    if (expansionDir == POS_X) {
        if (maxX+1>=m_iXRange) {
            (*pNewTotal) = bigValue;
            return;
        }
        for (int y=minY; y<=maxY; y++) {
            for (int z=minZ; z<=maxZ; z++) {
                (*pNewTotal) ++;
                if (GetOccupyFlag(maxX+1, y, z)) (*pNewOccupy) ++;
            }
        }
    }
    else if (expansionDir == NEG_X) {
        if (minX-1<0) {
            (*pNewTotal) = bigValue;
            return;
        }
        for (int y=minY; y<=maxY; y++) {
            for (int z=minZ; z<=maxZ; z++) {
                (*pNewTotal) ++;
                if (GetOccupyFlag(minX-1, y, z)) (*pNewOccupy) ++;
            }
        }
    }
    else if (expansionDir == POS_Y) {
        if (maxY+1>=m_iYRange) {
            (*pNewTotal) = bigValue;
            return;
        }
        for (int x=minX; x<=maxX; x++) {
            for (int z=minZ; z<=maxZ; z++) {
                (*pNewTotal) ++;
                if (GetOccupyFlag(x, maxY+1, z)) (*pNewOccupy) ++;
            }
        }
    }
    else if (expansionDir == NEG_Y) {
        if (minY-1<0) {
            (*pNewTotal) = bigValue;
            return;
        }
        for (int x=minX; x<=maxX; x++) {
            for (int z=minZ; z<=maxZ; z++) {
                (*pNewTotal) ++;
                if (GetOccupyFlag(x, minY-1, z)) (*pNewOccupy) ++;
            }
        }
    }
    else if (expansionDir == POS_Z) {
        if (maxZ+1>=m_iZRange) {
            (*pNewTotal) = bigValue;
            return;
        }
        for (int x=minX; x<maxX; x++) {
            for (int y=minY; y<=maxY; y++) {
                (*pNewTotal) ++;
                if (GetOccupyFlag(x, y, maxZ+1)) (*pNewOccupy) ++;
            }
        }
    }
    else if (expansionDir == NEG_Z) {
        if (minZ-1<0) {
            (*pNewTotal) = bigValue;
            return;
        }
        for (int x=minX; x<maxX; x++) {
            for (int y=minY; y<=maxY; y++) {
                (*pNewTotal) ++;
                if (GetOccupyFlag(x, y, minZ-1)) (*pNewOccupy) ++;
            }
        }
    }
    else {
        fprintf(stderr, "Voxelization::GetChangeOfMove expansionDir cannot be %d!\n", expansionDir);
        exit(-1);
    }
}

void Voxelization::
MarkOccupyFlagForCuboid(const R3IntBox& cuboid, int flag)
{
    for (int x=cuboid.minX; x<=cuboid.maxX; x++) {
        for (int y=cuboid.minY; y<=cuboid.maxY; y++) {
            for (int z=cuboid.minZ; z<=cuboid.maxZ; z++) {
                MarkOccupyFlag(x, y, z, flag);
            }
        }
    }
}

const vector<int> Voxelization::
GetNumberOfUnoccupiedCellsPerCuboids(void) const
{
    vector<int> numArray;
    for (int i=0; i<(int)m_aCuboidArray.size(); i++) {
        const R3IntBox& cuboid = m_aCuboidArray[i];
        int count = 0;
        for (int x=cuboid.minX; x<=cuboid.maxX; x++) {
            for (int y=cuboid.minY; y<=cuboid.maxY; y++) {
                for (int z=cuboid.minZ; z<=cuboid.maxZ; z++) {
                    if (GetOccupyFlag(x, y, z) == 0) count ++;
                }
            }
        }
        numArray.push_back(count);
    }
    return numArray;
}

//void Voxelization::
//SelectCuboids(double stopVolumeFraction, double minOverlapFraction)
//{
//    // count how many cells are occupied and set all flags to 0
//    int numOccupied = 0;
//    for (int i=0; i<(int)m_aOccupyFlags.size(); i++) {
//        if (m_aOccupyFlags[i] > 0) {
//            numOccupied ++;
//            m_aOccupyFlags[i] = 0;
//        }
//    }
//    int stopVolume = stopVolumeFraction*numOccupied;
//    printf("Voxelization::SelectCuboids stop volume = %dX%g = %d\n", numOccupied, stopVolumeFraction, stopVolume);
//    vector<int> numUnoccupiedCellsArray;
//    int maxUnOccupiedCells = 0;
//    vector<R3IntBox> selectedCuboidArray;
//    do {
//        numUnoccupiedCellsArray = GetNumberOfUnoccupiedCellsPerCuboids();
//        // prune the cells that do not reach minOverlapFraction
//        for (int i=0; i<(int)numUnoccupiedCellsArray.size(); i++) {
//            if (double(numUnoccupiedCellsArray[i])/m_aCuboidArray[i].Volume() < 1.0-minOverlapFraction)
//                numUnoccupiedCellsArray[i] = -1;
//        }
//        int maxIndex = -1;
//        maxUnOccupiedCells = MathUtil::Maximum(numUnoccupiedCellsArray, &maxIndex);
//        if (maxUnOccupiedCells < stopVolume) break;
//        selectedCuboidArray.push_back(m_aCuboidArray[maxIndex]);
//        SetOccupyFlagForCuboid(m_aCuboidArray[maxIndex]);
//        printf("Select cuboid %d, with num-newly-occupied-cells = %d, with non-overlapped-fraction = %g\n", maxIndex, maxUnOccupiedCells, double(maxUnOccupiedCells)/m_aCuboidArray[maxIndex].Volume());
//    } while (true);
//    // update m_aCuboidArray
//    m_aCuboidArray = selectedCuboidArray;
//}