#include "CompatibilityRanker.h"
#include "Utils/Utils.h"
#include "RNBasics/RNBasics.h"
#include <algorithm>
#include "ConfigParser/ConfigParser.h"
using namespace std;
Combination::
Combination(SceneGraph* pGraph)
:compatibilityScore(FLT_MAX),
selection(pGraph->NNodes(), 0),
m_pSceneGraph(pGraph)
{}

Combination::
Combination(const Combination& comb)
:compatibilityScore(comb.compatibilityScore),
selection(comb.selection),
m_pSceneGraph(comb.m_pSceneGraph),
stackedFeatureVector(comb.stackedFeatureVector),
edgeCompatibilityArray(comb.edgeCompatibilityArray)
{}

int Combination::
ImportSelection(const string& filename)
{
    // being very random
    for (int i=0; i<7; i++) {
        MathUtil::RandomInt(10);
    }
    
    vector<string> lines = FileUtil::ReadLines(filename);
    for (int i=0; i<(int)lines.size(); i++) {
        string line = StringUtil::Trim(lines[i], " ");
        if (line == "") continue;
        vector<string> segs = StringUtil::Split(line, ' ');
        if ((int)segs.size() != 2) {
            fprintf(stderr, "CompatibilityRanker::InitializeSelection unable to parse line %d!\n", i);
            return 0;
        }
        SceneGraphNode* pNode = m_pSceneGraph->GetNodeByCategory(segs[0]);
        if (!pNode) {
            fprintf(stderr, "CompatibilityRanker::InitializeSelection unable to find category %s!\n", segs[0].c_str());
            return 0;
        }
        int instanceIdx = -1;
        // random
        if (segs[1] == "random") {
            instanceIdx = MathUtil::RandomInt(pNode->NInstances());
        }
        else {
            instanceIdx = pNode->GetInstanceIndexByName(segs[1]);
            if (instanceIdx == -1) {
                fprintf(stderr, "CompatibilityRanker::InitializeSelection unable to find instance %s in category %s!\n", segs[1].c_str(), segs[0].c_str());
                return 0;
            }
        }
        selection[pNode->GraphIndex()] = instanceIdx;
    }
    Update();
    return 1;
}

void Combination::
UpdateSelection(int modelIndex, int newInstanceIdx)
{
    assert(modelIndex>=0 && modelIndex<m_pSceneGraph->NNodes());
    selection[modelIndex] = newInstanceIdx;
    Update();
}

void Combination::
UpdateSelection(const vector<int>& _selection)
{
    assert(_selection.size() == m_pSceneGraph->NNodes());
    selection = _selection;
    Update();
}

int Combination::
ExportSelection(const string& filename) const
{
    FILE* fout = fopen(filename.c_str(), "w");
    if (!fout) {
        fprintf(stderr, "Combination::ExportSelection unable to open %s!\n", filename.c_str());
        return 0;
    }
    assert((int)selection.size() == m_pSceneGraph->NNodes());
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        SceneGraphNode* pNode = m_pSceneGraph->Node(i);
        int instanceIdx = selection[i];
        fprintf(fout, "%s %s\n", pNode->CategoryName().c_str(),
                pNode->GetKthInstanceModelName(instanceIdx).c_str());
    }
    fclose(fout);
    return 1;
}

void Combination::
Update(void)
{
    UpdateFeatureVector();
    ComputeCompatibilityScore();
}

void Combination::
UpdateFeatureVector(void)
{
    stackedFeatureVector.clear();
    for (int i=0; i<(int)selection.size(); i++) {
        const vector<double>& v = m_pSceneGraph->Node(i)->GetKthInstanceDescriptors(selection[i]);
        stackedFeatureVector.insert(stackedFeatureVector.end(), v.begin(), v.end());
    }
}

void Combination::
ComputeCompatibilityScore(void)
{
    compatibilityScore = 0;
    edgeCompatibilityArray.clear();
    for (int i=0; i<m_pSceneGraph->NCompatibilityPairs(); i++) {
        SceneGraphNode* pFirst = m_pSceneGraph->GetKthCompatibilityPairFirst(i);
        SceneGraphNode* pSecond = m_pSceneGraph->GetKthCompatibilityPairSecond(i);
        const vector<double>& v1 = pFirst->GetKthInstanceDescriptors(selection[pFirst->GraphIndex()]);
        const vector<double>& v2 = pSecond->GetKthInstanceDescriptors(selection[pSecond->GraphIndex()]);
        double dist = MathUtil::EuclideanDistance(v1, v2);
        compatibilityScore += dist;
        edgeCompatibilityArray.push_back(dist);
    }
}

void Combination::
UpdateCompatibilityScoreUsingFullyConnectedGraph(void)
{
    compatibilityScore = 0;
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        SceneGraphNode* pFirst = m_pSceneGraph->Node(i);
        const vector<double>& v1 = pFirst->GetKthInstanceDescriptors(selection[i]);
        for (int j=i+1; j<m_pSceneGraph->NNodes(); j++) {
            SceneGraphNode* pSecond = m_pSceneGraph->Node(j);
            const vector<double>& v2 = pSecond->GetKthInstanceDescriptors(selection[j]);
            double dist = MathUtil::EuclideanDistance(v1, v2);
            compatibilityScore += dist;
        }
    }
}

double Combination::
GetKthEdgeCompatibility(int k) const
{
    assert(k>=0&&k<(int)edgeCompatibilityArray.size());
    return edgeCompatibilityArray[k];
}

void Combination::
SyncSceneGraphInstances(SceneGraph* pGraph) const
{
    for (int i=0; i<pGraph->NNodes(); i++) {
        SceneGraphNode* pNode = pGraph->Node(i);
        //if (pNode->IsLocked()) {
        //    assert(pNode->GetCurrentModelIndex() == m_vSuggestionArray[m_iCurrentIdx]->selection[i]);
        //}
        pNode->SelectKthModel(selection[i]);
    }
}

double Combination::
GetVisualizationScore(void) const
{
    return m_pSceneGraph->NCompatibilityPairs()/compatibilityScore;
}

const string Combination::
GetLockedNodeNames(void) const
{
    string list = "";
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        SceneGraphNode* pNode = m_pSceneGraph->Node(i);
        if (pNode->IsLocked()) {
            list = list + pNode->CategoryName() + ";";
        }
    }
    return list;
}

CompatibilityRanker::
CompatibilityRanker(SceneGraph* pSceneGraph)
:m_pSceneGraph(pSceneGraph),
m_pSavedSelection(NULL),
m_iCurrentIdx(-2),
m_aScoreTable(pSceneGraph->NNodes()),
m_aBackTrackTable(pSceneGraph->NNodes()),
m_bGlobalSuggestions(false)
{}

CompatibilityRanker::
~CompatibilityRanker(void)
{
    ClearSuggestionArray();
    delete m_pSavedSelection;
}

void CompatibilityRanker::
InitializeSelection(const string& filename)
{
    Combination* pComb = new Combination(m_pSceneGraph);
    int suc = pComb->ImportSelection(filename);
    if (!suc) {
        fprintf(stderr, "CompatibilityRanker::InitializeSelection unable to import from %s!\n",
                filename.c_str());
        exit(-1);
    }
    m_iCurrentIdx = m_vSuggestionArray.size();
    m_vSuggestionArray.push_back(pComb);
    SyncSceneGraphInstances();
}

double CompatibilityRanker::
CompatibilityDistance(SceneGraphNode* pNode1, int instanceIdx1,
                      SceneGraphNode* pNode2, int instanceIdx2) const
{
    const vector<double>& v1 = pNode1->GetKthInstanceDescriptors(instanceIdx1);
    const vector<double>& v2 = pNode2->GetKthInstanceDescriptors(instanceIdx2);
    return MathUtil::EuclideanDistance(v1, v2);
}

void CompatibilityRanker::
InitializeDPTables(void)
{
    assert((int)m_aScoreTable.size() == m_pSceneGraph->NNodes());
    assert(m_aScoreTable.size() == m_aBackTrackTable.size());
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        m_aScoreTable[i] = vector< vector<double> >(m_pSceneGraph->Node(i)->NInstances());
        m_aBackTrackTable[i] = vector< vector< vector<BackTrackInfo> > >(m_pSceneGraph->Node(i)->NInstances());
    }
}

void CompatibilityRanker::
BranchAndBoundSearchSumRecursive(int curPseudoIdx, vector<int>& workingIdxArray,
                                 double& partialScore,
                                 const vector<int>& childOrderedIndices,
                                 const vector<double>& bounds,
                                 vector<double>& scoreArray, double& maxScore,
                                 vector< vector<int> >& indexArray,
                                 const vector< vector<double> >& scoreTable,
                                 int num)
{
    // save to scoreArray and indexArray
    if (curPseudoIdx>=(int)childOrderedIndices.size()) {
        if ((int)scoreArray.size()<num) {
            scoreArray.push_back(partialScore);
            indexArray.push_back(workingIdxArray);
        }
        else if (partialScore < maxScore) {
            int replacedIdx = -1;
            double _maxScore = MathUtil::Maximum(scoreArray, &replacedIdx);
            assert(partialScore < _maxScore);
            scoreArray[replacedIdx] = partialScore;
            indexArray[replacedIdx] = workingIdxArray;
        }
        maxScore = MathUtil::Maximum(scoreArray);
        return;
    }
    // pruning
    if ((int)scoreArray.size()>=num && partialScore+bounds[curPseudoIdx]>=maxScore) return;
    // enumerate all solutions at this cell
    int childIdx = childOrderedIndices[curPseudoIdx];
    for (int i=0; i<(int)scoreTable[childIdx].size(); i++) {
        double partialScoreSaved = partialScore;
        partialScore += scoreTable[childIdx][i];
        workingIdxArray[childIdx] = i;
        BranchAndBoundSearchSumRecursive(curPseudoIdx+1, workingIdxArray,
                                         partialScore,
                                         childOrderedIndices, bounds,
                                         scoreArray, maxScore,
                                         indexArray, scoreTable, num);
        partialScore = partialScoreSaved;
    }
}

void CompatibilityRanker::
BranchAndBoundSearchSum(vector<double>& scoreTable,
                        vector< vector<BackTrackInfo> >& backTrackTable,
                        const vector< vector<double> >& scoreChildArray,
                        const vector< vector<BackTrackInfo> >& backTrackChildArray,
                        int num)
{
    // assume scoreChildArray[i] are already ranked
    // so we rank the children by max-min in scoreChildArray[childIdx]
    vector<double> diffs;
    for (int i=0; i<(int)scoreChildArray.size(); i++) {
        const vector<double>& array = scoreChildArray[i];
        assert((int)array.size() > 0);
        double diff = array[array.size()-1] - array[0];
        assert(diff >= 0);
        diffs.push_back(diff);
    }
    vector<int> childOrderedIndices;
    // descending order
    MathUtil::Sort(diffs, childOrderedIndices, false);
    // bound at each cell
    vector<double> bounds(childOrderedIndices.size(), 0);
    bounds[bounds.size()-1] = scoreChildArray[childOrderedIndices[bounds.size()-1]][0];
    for (int i=bounds.size()-2; i>=0; i--) {
        bounds[i] = bounds[i+1] + scoreChildArray[childOrderedIndices[i]][0];
    }
    vector<double> scoreTableTemp;
    // pair = solution index ordered by real order of children
    vector< vector<int> > indexTableTemp;
    vector<int> workingIdxArray(childOrderedIndices.size(), 0);
    double partialScore = 0;
    double maxScore = FLT_MAX;
    BranchAndBoundSearchSumRecursive(0, workingIdxArray, partialScore,
                                     childOrderedIndices, bounds,
                                     scoreTableTemp, maxScore, indexTableTemp,
                                     scoreChildArray, num);
    // convert to scoreTable and backTrackTable
    vector<int> sortedIndices;
    MathUtil::Sort(scoreTableTemp, sortedIndices);
    // score and solution table
    for (int idx=0; idx<(int)sortedIndices.size(); idx++) {
        int i = sortedIndices[idx];
        scoreTable.push_back(scoreTableTemp[i]);
        vector<BackTrackInfo> backTrackArray;
        // real order of children
        for (int j=0; j<(int)indexTableTemp[i].size(); j++) {
            int childSolutionIdx = indexTableTemp[i][j];
            backTrackArray.push_back(backTrackChildArray[j][childSolutionIdx]);
        }
        backTrackTable.push_back(backTrackArray);
    }
}

void CompatibilityRanker::
BottomUpSingle(SceneGraphNode* pNode, int modelIndex, int num)
{
    // select the *num* least sum
    vector<double>& scoreTable = m_aScoreTable[pNode->GraphIndex()][modelIndex];
    vector< vector<BackTrackInfo> >& backTrackTable = m_aBackTrackTable[pNode->GraphIndex()][modelIndex];
    assert(scoreTable.size() == 0);
    assert(backTrackTable.size() == 0);
    if (pNode->NChildren() == 0) {
        scoreTable.push_back(0);
        backTrackTable.push_back(vector<BackTrackInfo>());
        return;
    }
    // go through all children, all models for each child,
    // and all candidates for each model
    // (1) children
    // (2) 10 best pairs of model index and solution index
    vector< vector<BackTrackInfo> > backTrackChildArray;
    // (2) 10 best scores (of the edge)
    vector< vector<double> > scoreChildArray;
    for (int i=0; i<pNode->NChildren(); i++) {
        vector<BackTrackInfo> backTrackArray;
        vector<double> scoreArray;
        SceneGraphNode* pChildNode = pNode->Child(i);
        // generate all possible scores
        for (int j=0; j<pChildNode->NInstances(); j++) {
            const vector<double>& childScoreArray = m_aScoreTable[pChildNode->GraphIndex()][j];
            double edgeScore = CompatibilityDistance(pNode, modelIndex, pChildNode, j);
            for (int k=0; k<(int)childScoreArray.size(); k++) {
                scoreArray.push_back(childScoreArray[k]+edgeScore);
                BackTrackInfo info;
                info.modelIndex = j;
                info.solutionIndex = k;
                backTrackArray.push_back(info);
            }
        }
        // rank and select the most promising *num*
        vector<int> indices = MathUtil::KthMinimum(scoreArray, num);
        vector<BackTrackInfo> selectedBackTrackArray;
        vector<double> selectedScoreArray;
        for (int j=0; j<(int)indices.size(); j++) {
            selectedBackTrackArray.push_back(backTrackArray[indices[j]]);
            selectedScoreArray.push_back(scoreArray[indices[j]]);
        }
        // save
        backTrackChildArray.push_back(selectedBackTrackArray);
        scoreChildArray.push_back(selectedScoreArray);
    }
    BranchAndBoundSearchSum(scoreTable, backTrackTable,
                            scoreChildArray, backTrackChildArray, num);
}

void CompatibilityRanker::
BottomUp(SceneGraphNode* pNode, int num)
{
    for (int i=0; i<pNode->NChildren(); i++) {
        BottomUp(pNode->Child(i), num);
    }
    if (pNode->IsLocked()) {
        int modelIndex = pNode->GetCurrentModelIndex();
        BottomUpSingle(pNode, modelIndex, num);
    }
    else {
        for (int i=0; i<pNode->NInstances(); i++) {
            // skip if this instance is already selected
            // (only used in global suggestions)
            if (m_aSelectionFlags[pNode->GraphIndex()][i]) continue;
            BottomUpSingle(pNode, i, num);
        }
    }
    
}

void CompatibilityRanker::
TopDown(int num)
{
    SceneGraphNode* pNode = m_pSceneGraph->Root();
    vector<BackTrackInfo> backTrackArray;
    vector<double> scoreArray;
    for (int modelIndex=0; modelIndex<pNode->NInstances(); modelIndex++) {
        const vector<double>& scoreTable = m_aScoreTable[pNode->GraphIndex()][modelIndex];
        for (int i=0; i<(int)scoreTable.size(); i++) {
            scoreArray.push_back(scoreTable[i]);
            BackTrackInfo info;
            info.modelIndex = modelIndex;
            info.solutionIndex = i;
            backTrackArray.push_back(info);
        }
    }
    // find the best num
    vector<int> indices = MathUtil::KthMinimum(scoreArray, num);
    vector<BackTrackInfo> selectedBackTrackArray;
    vector<double> selectedScoreArray;
    for (int i=0; i<(int)indices.size(); i++) {
        selectedBackTrackArray.push_back(backTrackArray[indices[i]]);
        selectedScoreArray.push_back(scoreArray[indices[i]]);
    }
    // convert selectedBackTrackArray to m_vSuggestionArray
    for (int i=0; i<(int)selectedBackTrackArray.size(); i++) {
        Combination* pComb = new Combination(m_pSceneGraph);
        TopDownSingle(pComb, pNode, selectedBackTrackArray[i]);
        pComb->Update();
        m_vSuggestionArray.push_back(pComb);
    }
}

void CompatibilityRanker::
TopDownSingle(Combination* pComb, SceneGraphNode* pNode, const BackTrackInfo& info)
{
    int nodeIndex = pNode->GraphIndex();
    pComb->selection[nodeIndex] = info.modelIndex;
    const vector<BackTrackInfo>& childInfoArray = m_aBackTrackTable[nodeIndex][info.modelIndex][info.solutionIndex];
    if ((int)childInfoArray.size() != pNode->NChildren()) {
        printf("at %s %d != %d\n", pNode->CategoryName().c_str(),
               (int)childInfoArray.size(), pNode->NChildren());
        assert(false);
    }
    for (int i=0; i<pNode->NChildren(); i++) {
        SceneGraphNode* pChildNode = pNode->Child(i);
        TopDownSingle(pComb, pChildNode, childInfoArray[i]);
    }
}

void CompatibilityRanker::
ClearSuggestionArray(void)
{
    for (int i=0; i<(int)m_vSuggestionArray.size(); i++) {
        delete m_vSuggestionArray[i];
    }
    m_vSuggestionArray.clear();
    m_iCurrentIdx = -1;
}

void CompatibilityRanker::
ProduceSuggestions(int num)
{
    SaveCurrentSelection();
    if (m_bGlobalSuggestions)
        ProduceGlobalSuggestions(num);
    else
        ProduceLocalSuggestions(num);
}

bool CompatibilityRanker::
IsIdenticalToASuggestion(Combination* pComb) const
{
    for (int i=0; i<(int)m_vSuggestionArray.size(); i++) {
        if (pComb->selection == m_vSuggestionArray[i]->selection) return true;
    }
    return false;
}

void CompatibilityRanker::
ProduceRandomSuggestions(int num)
{
    SaveCurrentSelection();
    ClearSuggestionArray();
    //InitializeSelectionFlags();
    
    // check which nodes are unlocked
    vector<int> unlockedNodeIndices;
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        if (!m_pSceneGraph->Node(i)->IsLocked())
            unlockedNodeIndices.push_back(i);
    }
    // all nodes are locked -> no suggestions
    if ((int)unlockedNodeIndices.size() == 0) return;
    // >= 2 nodes are locked -> sampling
    if ((int)unlockedNodeIndices.size() >= 2) {
        for (int i=0; i<num; i++) {
            Combination* pComb = new Combination(*m_pSavedSelection);
            do {
                for (int j=0; j<(int)unlockedNodeIndices.size(); j++) {
                    int nodeIndex = unlockedNodeIndices[j];
                    pComb->selection[nodeIndex] = MathUtil::RandomInt(m_pSceneGraph->Node(nodeIndex)->NInstances());
                }
            } while (IsIdenticalToASuggestion(pComb) ||
                     pComb->selection == m_pSavedSelection->selection);
            pComb->Update();
            m_vSuggestionArray.push_back(pComb);
        }
    }
    // == 1 node is locked -> permutation
    else {
        vector<int> array;
        int nodeIndex = unlockedNodeIndices[0];
        for (int i=0; i<m_pSceneGraph->Node(nodeIndex)->NInstances(); i++)
            array.push_back(i);
        random_shuffle(array.begin(), array.end());
        for (int i=0; i<(int)array.size() && (int)m_vSuggestionArray.size()<num; i++) {
            if (array[i] == m_pSavedSelection->selection[nodeIndex]) continue;
            Combination* pComb = new Combination(*m_pSavedSelection);
            pComb->selection[nodeIndex] = array[i];
            pComb->Update();
            m_vSuggestionArray.push_back(pComb);
        }
    }
    m_iCurrentIdx = -1;
    SyncSceneGraphInstances();
}

// save the current selection to m_pSavedSelection
void CompatibilityRanker::
SaveCurrentSelection(void)
{
    // if m_pSavedSelection is currently selected, then do nothing
    if (m_iCurrentIdx == -1) return;
    // destroy the current saved selection
    if (m_pSavedSelection) delete m_pSavedSelection;
    m_pSavedSelection = new Combination(*(GetCurrentSuggestion()));
}

void CompatibilityRanker::
RemoveDuplicatedSelection(Combination* pComb, int num)
{
    assert(pComb);
    for (int i=0; i<(int)m_vSuggestionArray.size(); i++) {
        if (pComb->selection == m_vSuggestionArray[i]->selection) {
            delete m_vSuggestionArray[i];
            m_vSuggestionArray.erase(m_vSuggestionArray.begin()+i);
            break;
        }
    }
    // in case nothing is removed and m_vSuggestionArray.size() > num
    if (m_vSuggestionArray.size() > num) {
        assert(m_vSuggestionArray.size() == num+1);
        delete m_vSuggestionArray[num];
        m_vSuggestionArray.erase(m_vSuggestionArray.begin()+num);
    }
}

void CompatibilityRanker::
InitializeSelectionFlags(void)
{
    m_aSelectionFlags = vector< vector<bool> >(m_pSceneGraph->NNodes());
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        m_aSelectionFlags[i] = vector<bool>(m_pSceneGraph->Node(i)->NInstances(), false);
    }
}

// return 0, #candidates == 0
// return 1, >, only one node is unlocked (has to enumerate all possibilities)
// return 2, >>, more than one node are unlocked (dont need to enumerate all possibilities)
int CompatibilityRanker::
AreThereEnoughCandidatesLeft(void) const
{
    vector<int> remainedExamples; // include locked nodes
    int numUnlocked = 0;
    for (int i=0; i<m_pSceneGraph->NNodes(); i++) {
        if (m_pSceneGraph->Node(i)->IsLocked()) {
            remainedExamples.push_back(1);
            continue;
        }
        int sum = 0;
        numUnlocked ++;
        for (int j=0; j<(int)m_aSelectionFlags[i].size(); j++) {
            if (!m_aSelectionFlags[i][j]) sum ++;
        }
        remainedExamples.push_back(sum);
    }
    // find 0
    int product = 1;
    for (int i=0; i<(int)remainedExamples.size(); i++) {
        if (remainedExamples[i] == 0) {
            product = 0;
            break;
        }
    }
    if (product == 0) {
        return 0;
    }
    // exception: if all nodes are locked
    if (numUnlocked == 0) return 0;
    // if more than one node is unlocked, then return 2
    // if only one node is unlocked, then return 1
    if (numUnlocked == 1) return 1;
    return 2;
}

void CompatibilityRanker::
ProduceLocalSuggestions(int num)
{
    // step 0: remember the current combination
    ClearSuggestionArray();
    // step 1: we use the structure in scene graph
    InitializeSelectionFlags();
    InitializeDPTables();
    // step 2: from the leaf nodes, we remember the best *num* solutions for each assignment of the node
    // and we trace up into the tree, and at each node, I only remember 10 best solutions for each possible assignment
    BottomUp(m_pSceneGraph->Root(), num+1);
    // step 3: create m_vSuggestionArray
    TopDown(num+1);
    // step 4: remove current combination from suggestion array (if exists)
    RemoveDuplicatedSelection(m_pSavedSelection, num);
    m_iCurrentIdx = -1;
    SyncSceneGraphInstances();
}

void CompatibilityRanker::
ProduceGlobalSuggestions(int num)
{
    // step 0: remember the current combination
    ClearSuggestionArray();
    // step 1: initialize selection flags
    InitializeSelectionFlags();
    if (m_pSavedSelection) {
        for (int i=0; i<(int)m_pSavedSelection->selection.size(); i++)
            m_aSelectionFlags[i][m_pSavedSelection->selection[i]] = true;
    }
    // step 2: repeatedly generate the combination with lowest score
    for (int i=0; i<num+1; i++) {
        if (AreThereEnoughCandidatesLeft() == 0) break;
        // step 1: we use the structure in scene graph
        InitializeDPTables();
        BottomUp(m_pSceneGraph->Root(), 1);
        // push a new item to m_vSuggestionArray
        TopDown(1);
        // set selection flags
        Combination* pComb = m_vSuggestionArray[m_vSuggestionArray.size()-1];
        for (int j=0; j<(int)pComb->selection.size(); j++) {
            m_aSelectionFlags[j][pComb->selection[j]] = true;
        }
    }
    // step 4: remove current combination from suggestion array (if exists)
    RemoveDuplicatedSelection(m_pSavedSelection, num);
    m_iCurrentIdx = -1;
    SyncSceneGraphInstances();
}

bool CompatibilityRanker::
GoToNextCombination(void)
{
    if (m_iCurrentIdx>=NSuggestions()-1) return false;
    GoToKthCombination(m_iCurrentIdx+1);
    return true;
}

bool CompatibilityRanker::
GoToPrevCombination(void)
{
    if (m_iCurrentIdx<=-1) return false;
    GoToKthCombination(m_iCurrentIdx-1);
    return true;
}

void CompatibilityRanker::
GoToKthCombination(int k)
{
    assert(k>=-1 && k<NSuggestions());
    m_iCurrentIdx = k;
    SyncSceneGraphInstances();
}

void CompatibilityRanker::
SyncSceneGraphInstances(void)
{
    // assert((int)m_vSuggestionArray[m_iCurrentIdx]->selection.size() == m_pSceneGraph->NNodes());
    GetCurrentSuggestion()->SyncSceneGraphInstances(m_pSceneGraph);
}

int CompatibilityRanker::
NSuggestions(void) const
{
    return (int)m_vSuggestionArray.size();
}

Combination* CompatibilityRanker::
GetKthSuggestion(int k) const
{
    if (k == -1) return m_pSavedSelection;
    assert(k>=0 && k<NSuggestions());
    return m_vSuggestionArray[k];
}

int CompatibilityRanker::
CurrentSuggestionIndex(void) const
{
    return m_iCurrentIdx;
}

Combination* CompatibilityRanker::
GetCurrentSuggestion(void) const
{
    return GetKthSuggestion(m_iCurrentIdx);
}

Combination* CompatibilityRanker::
GetSavedSelection(void) const
{
    return m_pSavedSelection;
}

void CompatibilityRanker::
EnableGlobalSuggestions(void)
{
    m_bGlobalSuggestions = true;
}

void CompatibilityRanker::
DisableGlobalSuggestions(void)
{
    m_bGlobalSuggestions = false;
}

bool CompatibilityRanker::
IsGlobalSuggestionEnabled(void) const
{
    return m_bGlobalSuggestions;
}