#ifndef MY_COMPATIBILITY_RANKER_H__
#define MY_COMPATIBILITY_RANKER_H__
#include "SceneGraph.h"
#include <vector>
#include <string>

class Combination
{
public:
    Combination(SceneGraph* pGraph);
    Combination(const Combination& comb);
    int ImportSelection(const std::string& filename);
    int ExportSelection(const std::string& filename) const;
    // double FeatureVectorDistance(const Combination& comb) const;
    double compatibilityScore;
    std::vector<int> selection;
    void Update(void);
    double GetKthEdgeCompatibility(int k) const;
    void SyncSceneGraphInstances(SceneGraph* pGraph) const;
    double GetVisualizationScore(void) const;
    const std::string GetLockedNodeNames(void) const;
    void UpdateSelection(int modelIndex, int newInstanceIdx);
    void UpdateSelection(const std::vector<int>& _selection);
    void UpdateCompatibilityScoreUsingFullyConnectedGraph(void);
protected:
    void UpdateFeatureVector(void);
    void ComputeCompatibilityScore(void);
protected:
    SceneGraph* m_pSceneGraph;
    std::vector<double> stackedFeatureVector;
    std::vector<double> edgeCompatibilityArray;
};

struct BackTrackInfo
{
    int modelIndex;
    int solutionIndex;
};

class CompatibilityRanker
{
public:
    CompatibilityRanker(SceneGraph* pSceneGraph);
    ~CompatibilityRanker(void);
    void InitializeSelection(const std::string& filename);
    void ProduceRandomSuggestions(int num);
    void ProduceSuggestions(int num);
    bool GoToNextCombination(void);
    bool GoToPrevCombination(void);
    void GoToKthCombination(int k);
    int NSuggestions(void) const;
    Combination* GetKthSuggestion(int k) const;
    int CurrentSuggestionIndex(void) const;
    Combination* GetCurrentSuggestion(void) const;
    Combination* GetSavedSelection(void) const;
    void EnableGlobalSuggestions(void);
    void DisableGlobalSuggestions(void);
    bool IsGlobalSuggestionEnabled(void) const;
protected:
    void ProduceLocalSuggestions(int num);
    void ProduceGlobalSuggestions(int num);
    void SyncSceneGraphInstances(void);
    void SaveCurrentSelection(void);
    void RemoveDuplicatedSelection(Combination* pCurrentComb, int num);
    // compatibility
    double CompatibilityDistance(SceneGraphNode* pNode1, int instanceIdx1,
                                 SceneGraphNode* pNode2, int instanceIdx2) const;
    // dynamic programming related
    void InitializeDPTables(void);
    void ClearSuggestionArray(void);
    void BottomUp(SceneGraphNode* pNode, int num);
    void BottomUpSingle(SceneGraphNode* pNode, int modelIndex, int num);
    void TopDown(int num);
    void TopDownSingle(Combination* pComb, SceneGraphNode* pNode,
                       const BackTrackInfo& info);
    // branch and bound for finding optimal
    void BranchAndBoundSearchSum(std::vector<double>& scoreTable,
                                 std::vector< std::vector<BackTrackInfo> >& backTrackTable,
                                 const std::vector< std::vector<double> >& scoreChildArray,
                                 const std::vector< std::vector<BackTrackInfo> >& backTrackChildArray,
                                 int num);
    void BranchAndBoundSearchSumRecursive(int curPseudoIdx, std::vector<int>& workingIdxArray,
                                          double& partialScore,
                                          const std::vector<int>& childOrderedIndices,
                                          const std::vector<double>& bounds,
                                          std::vector<double>& scoreArray, double& maxScore,
                                          std::vector< std::vector<int> >& indexArray,
                                          const std::vector< std::vector<double> >& scoreTable,
                                          int num);
    // selection flags
    void InitializeSelectionFlags(void);
    int AreThereEnoughCandidatesLeft(void) const;
    bool IsIdenticalToASuggestion(Combination* pComb) const;
protected:
    SceneGraph* m_pSceneGraph;
    std::vector<Combination*> m_vSuggestionArray;
    Combination* m_pSavedSelection;
    int m_iCurrentIdx;
    
    // dynamic programming states
    // (1) scene graph node index
    // (2) model index
    // (3) ordered multiple optimal results
    std::vector< std::vector< std::vector<double> > > m_aScoreTable;
    // (4) model index/solution of children (in the order of SceneGraphNode->Child(k))
    std::vector< std::vector< std::vector< std::vector<BackTrackInfo> > > > m_aBackTrackTable;
    // global or local
    bool m_bGlobalSuggestions;
    // selected flags
    std::vector< std::vector<bool> > m_aSelectionFlags;
};
#endif