#include "SceneGraph.h"
#include "Utils/Utils.h"
#include "ConfigParser/ConfigParser.h"
using namespace std;

SceneGraph::
SceneGraph(void)
:m_pRootNode(NULL)
{}

SceneGraph::
~SceneGraph(void)
{
    for (int i=0; i<(int)m_aNodeArray.size(); i++) {
        delete m_aNodeArray[i];
    }
    m_aNodeArray.clear();
}

static string
GetModelDirname(const string& modelReponame, const string& categoryName)
{
    return modelReponame+"/"+categoryName+"/obj";
}

int SceneGraph::
ImportModelDatabase(const string& modelReponame, const string& descReponame)
{
    //printf("SceneGraph::ImportModelDatabase from %s and %s ...\n", modelReponame.c_str(), descReponame.c_str());
    for (int i=0; i<(int)m_aNodeArray.size(); i++) {
        //printf("    %s ... ", m_aNodeArray[i]->CategoryName().c_str());
        fflush(stdout);
        m_aNodeArray[i]->ImportModelDatabase(GetModelDirname(modelReponame, m_aNodeArray[i]->CategoryName()),
                                             descReponame);
        //printf("%d models!\n", m_aNodeArray[i]->NInstances());
    }
    return 1;
}

int SceneGraph::
ImportCompatibilityPairs(const string& filename)
{
    m_vCompatibilityPairs.clear();
    printf("SceneGraph::ImportCompatibilityPairs from %s ...\n", filename.c_str());
    vector<string> lines = FileUtil::ReadLines(filename);
    for (int i=0; i<(int)lines.size(); i++) {
        string line = StringUtil::Trim(lines[i], " ");
        if (line == "") continue;
        vector<string> segs = StringUtil::Split(line, ' ');
        if ((int)segs.size() != 2) {
            fprintf(stderr, "SceneGraph::ImportCompatibilityPairs unable to parse line %d!\n", i);
            return 0;
        }
        SceneGraphNode* pNode1 = GetNodeByCategory(segs[0]);
        if (!pNode1) {
            fprintf(stderr, "SceneGraph::ImportCompatibilityPairs unable to find category %s!\n", segs[0].c_str());
            return 0;
        }
        SceneGraphNode* pNode2 = GetNodeByCategory(segs[1]);
        if (!pNode2) {
            fprintf(stderr, "SceneGraph::ImportCompatibilityPairs unable to find category %s!\n", segs[1].c_str());
            return 0;
        }
        m_vCompatibilityPairs.push_back(make_pair(pNode1, pNode2));
    }
    printf("    find %d compatibility pairs!\n", (int)m_vCompatibilityPairs.size());
    return 1;
}

int SceneGraph::
InitializePermanentLocks(const std::string& filename)
{
    printf("SceneGraph::InitializePermanentLocks from %s ...\n", filename.c_str());
    vector<string> lines = FileUtil::ReadLines(filename);
    for (int i=0; i<(int)lines.size(); i++) {
        string line = StringUtil::Trim(lines[i], " ");
        if (line == "") continue;
        SceneGraphNode* pNode = GetNodeByCategory(line);
        if (!pNode) {
            fprintf(stderr, "SceneGraph::InitializePermanentLocks unable to find category %s!\n", line.c_str());
            return 0;
        }
        pNode->SetPermanentLockStatus(true);
        pNode->m_bLock = true;
    }
    return 1;
}

int SceneGraph::
ImportPositions(const string& filename)
{
    FILE* fp;
    if (!(fp = fopen(filename.c_str(), "r"))) {
        fprintf(stderr, "SceneGraph::ImportLayout unable to open %s!\n",
                filename.c_str());
        return 0;
    }
    char buffer[1024];
    int lineCount = 0;
    SceneGraphNode* pCurNode = NULL;
    while (fgets(buffer, 1023, fp)) {
        lineCount ++;
        // Skip white space
        char *bufferp = buffer;
        while (isspace(*bufferp)) bufferp++;
        
        // Skip blank lines and comments
        if (*bufferp == '#') continue;
        if (*bufferp == '\0') continue;
        
        // Get keyword
        char keyword[80];
        if (sscanf(bufferp, "%s", keyword) != 1) {
            fprintf( stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
            return 0;
        }
        // Check keyword
        if (!strcmp(keyword, "node")) {
            int nodeIndex;
            char categoryName[1024];
            if (sscanf(bufferp, "%s%d%s", keyword, &nodeIndex, categoryName) != 3) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
                return 0;
            }
            if (nodeIndex != NNodes()) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
                return 0;
            }
            pCurNode = new SceneGraphNode(this);
            pCurNode->m_sCategoryName = categoryName;
        }
        else if (!strcmp(keyword, "parent")) {
            if (!pCurNode) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s (no current node)\n", lineCount, filename.c_str());
                return 0;
            }
            int parentIndex;
            if (sscanf(bufferp, "%s%d", keyword, &parentIndex) != 2) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
                return 0;
            }
            if (parentIndex != -1) {
                if (parentIndex<0||parentIndex>=NNodes()) {
                    fprintf(stderr, "SceneGraph::ImportLayout unable to find nodeIndex=%d on line %d in file %s\n",
                            parentIndex, lineCount, filename.c_str());
                    return 0;
                }
                Node(parentIndex)->InsertChild(pCurNode);
            }
        }
        else if (!strcmp(keyword, "relative_direction")) {
            if (!pCurNode) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s (no current node)\n", lineCount,
                        filename.c_str());
                return 0;
            }
            double x,y,z;
            if (sscanf(bufferp, "%s%lg%lg%lg", keyword, &x, &y, &z) != 4) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
                return 0;
            }
            pCurNode->m_vRelativeOffsetDirection = R3Vector(x, y, z);
        }
        else if (!strcmp(keyword, "relative_offset")) {
            if (!pCurNode) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s (no current node)\n", lineCount,
                        filename.c_str());
                return 0;
            }
            double offset;
            if (sscanf(bufferp, "%s%lg", keyword, &offset) != 2) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
                return 0;
            }
            pCurNode->m_fSplitDist = offset;
        }
        else if (!strcmp(keyword, "self_rotation")) {
            if (!pCurNode) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s (no current node)\n", lineCount,
                        filename.c_str());
                return 0;
            }
            double radian = 0;
            if (sscanf(bufferp, "%s%lg", keyword, &radian) != 2) {
                fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s\n", lineCount, filename.c_str());
                return 0;
            }
            pCurNode->m_fSelfRotationInRadian = radian;
        }
        else {
            fprintf(stderr, "SceneGraph::ImportLayout syntax error on line %d in file %s (unable to parse keyword %s)\n", lineCount, filename.c_str(), keyword);
            return 0;
        }
    }
    fclose(fp);
    // find m_pRootNode
    m_pRootNode = NULL;
    for (int i=0; i<(int)m_aNodeArray.size(); i++) {
        if (!m_aNodeArray[i]->Parent()) {
            if (m_pRootNode) {
                fprintf(stderr, "SceneGraph::ImportLayout node %d and node %d are both root!\n", m_pRootNode->GraphIndex(), m_aNodeArray[i]->GraphIndex());
                return 0;
            }
            m_pRootNode = m_aNodeArray[i];
        }
    }
    return 1;
}

int SceneGraph::
ExportPositions(const string& filename) const
{
    FILE* fp;
    if (!(fp = fopen(filename.c_str(), "w"))) {
        fprintf(stderr, "SceneGraph::ExportPositions unable to open %s!\n",
                filename.c_str());
        return 0;
    }
    for (int i=0; i<(int)m_aNodeArray.size(); i++) {
        SceneGraphNode* pNode = m_aNodeArray[i];
        fprintf(fp, "node %d %s\n", i, pNode->CategoryName().c_str());
        if (pNode->Parent() == NULL) {
            fprintf(fp, "parent -1\n");
            continue;
        }
        fprintf(fp, "parent %d\n", pNode->Parent()->GraphIndex());
        fprintf(fp, "relative_direction %g %g %g\n", pNode->m_vRelativeOffsetDirection[0],
                pNode->m_vRelativeOffsetDirection[1], pNode->m_vRelativeOffsetDirection[2]);
        fprintf(fp, "relative_offset %g\n", pNode->m_fSplitDist);
        fprintf(fp, "self_rotation %g\n", pNode->m_fSelfRotationInRadian);
    }
    fclose(fp);
    return 1;
}

void SceneGraph::
Draw(bool drawOnlyFreeObjects) const
{
    for (int i=0; i<NNodes(); i++) {
        if (!drawOnlyFreeObjects || !Node(i)->IsLocked())
            Node(i)->Draw();
    }
}


SceneGraphNode* SceneGraph::
Root(void) const
{
    return m_pRootNode;
}

int SceneGraph::
NNodes(void) const
{
    return m_aNodeArray.size();
}

SceneGraphNode* SceneGraph::
Node(int idx) const
{
    assert(idx>=0 && idx<NNodes());
    return m_aNodeArray[idx];
}

const vector<R3Point> SceneGraph::
GetBBoxCornerPoints(void) const
{
    ((SceneGraph*)this)->UpdateTransformations();
    vector<R3Point> pointArray;
    for (int i=0; i<NNodes(); i++) {
        R3Box bbox = Node(i)->BBox();
        for (int j=0; j<8; j++) {
            R3Point point = bbox.Corner(j);
            point.Transform(Node(i)->Transformation());
            pointArray.push_back(point);
        }
    }
    return pointArray;
}

const vector<R3Point> SceneGraph::
GetBBoxCornerPointsOfFreeObjects(void) const
{
    ((SceneGraph*)this)->UpdateTransformations();
    vector<R3Point> pointArray;
    for (int i=0; i<NNodes(); i++) {
        if (Node(i)->IsLocked()) continue;
        R3Box bbox = Node(i)->BBox();
        for (int j=0; j<8; j++) {
            R3Point point = bbox.Corner(j);
            point.Transform(Node(i)->Transformation());
            pointArray.push_back(point);
        }
    }
    return pointArray;
}

R3Box SceneGraph::
BBox(void) const
{
    R3Box bbox = R3null_box;
    vector<R3Point> pointArray = GetBBoxCornerPoints();
    for (int i=0; i<(int)pointArray.size(); i++) {
        bbox.Union(pointArray[i]);
    }
    return bbox;
}

R3Box SceneGraph::
BBoxOfFreeObjects(void) const
{
    R3Box bbox = R3null_box;
    vector<R3Point> pointArray = GetBBoxCornerPointsOfFreeObjects();
    for (int i=0; i<(int)pointArray.size(); i++) {
        bbox.Union(pointArray[i]);
    }
    return bbox;
}

SceneGraphNode* SceneGraph::
GetNodeByCategory(const string& categoryName) const
{
    for (int i=0; i<(int)m_aNodeArray.size(); i++) {
        if (m_aNodeArray[i]->CategoryName() == categoryName)
            return m_aNodeArray[i];
    }
    return NULL;
}

void SceneGraph::
AddNode(SceneGraphNode* pNode)
{
    pNode->m_pSceneGraph = this;
    pNode->m_iIndex = m_aNodeArray.size();
    m_aNodeArray.push_back(pNode);
}

void SceneGraph::
UpdateTransformations(void)
{
    m_pRootNode->UpdateTransformation();
}

int SceneGraph::
NCompatibilityPairs(void) const
{
    return (int)m_vCompatibilityPairs.size();
}

SceneGraphNode* SceneGraph::
GetKthCompatibilityPairFirst(int k) const
{
    assert(k>=0 && k<NCompatibilityPairs());
    return m_vCompatibilityPairs[k].first;
}

SceneGraphNode* SceneGraph::
GetKthCompatibilityPairSecond(int k) const
{
    assert(k>=0 && k<NCompatibilityPairs());
    return m_vCompatibilityPairs[k].second;
}

SceneGraphNode* SceneGraph::
Intersect(const R3Ray& ray, double* pHit) const
{
    // update transformation
    m_pRootNode->UpdateTransformation();
    // intersects
    SceneGraphNode* pHitNode = NULL;
    double hitT = FLT_MAX;
    for (int i=0; i<NNodes(); i++) {
        double t = FLT_MAX;
        bool hitFlag = Node(i)->Intersect(ray, &t);
        if (hitFlag && t<hitT) {
            hitT = t;
            pHitNode = Node(i);
        }
    }
    if (pHit) *pHit = hitT;
    return pHitNode;
}