#ifndef MY_SCENE_GRAPH_H__
#define MY_SCENE_GRAPH_H__
#include "R3MyGraphics/R3MyGraphics.h"
#include <vector>
#include <string>
class SceneGraph;

class SceneModel
{
public:
    SceneModel(void);
    int ImportMesh(const std::string& filename);
    int ImportDescriptors(const std::string& filename);
    R3Mesh* GetMesh(void) const;
    const std::vector<double>& GetDescriptors(void) const;
    const std::string& GetModelName(void) const;
protected:
    // geometry
    R3Mesh* m_pMesh;
    // embedded descriptor
    std::vector<double> m_vDesc;
    // name
    std::string m_sModelName;
};

class SceneGraphNode
{
    friend class SceneGraph;
public:
    SceneGraphNode(SceneGraph* pSceneGraph);
    int ImportModelDatabase(const std::string& meshDirname, const std::string& descDirname);
    void Draw(void) const;
    void DrawWhenDragging(const R3Affine& affine) const;
    bool Intersect(const R3Ray& ray, double* pHit) const;
    // category
    void SetCategoryName(const std::string& name);
    const std::string& CategoryName(void) const;
    // scene graph structure
    int NChildren(void) const;
    SceneGraphNode* Child(int idx) const;
    SceneGraphNode* Parent(void) const;
    void InsertChild(SceneGraphNode* pNode);
    int GraphIndex(void) const;
    // transformation
    bool IsGeometryUpdated(void) const;
    void UpdateTransformation(void);
    void UpdatePositions(void);
    void Transform(const R3Affine& affine);
    // instances
    int NInstances(void) const;
    R3Mesh* GetKthInstance(int kth) const;
    const std::vector<double>& GetKthInstanceDescriptors(int kth) const;
    const std::string& GetKthInstanceModelName(int kth) const;
    R3Mesh* GetInstanceByName(const std::string& name) const;
    int GetInstanceIndexByName(const std::string& name) const;
    R3Mesh* GetCurrentModel(void) const;
    int GetCurrentModelIndex(void) const;
    void SelectKthModel(int kth);
    // bbox
    R3Box BBox(void) const;
    R3Point BottomCenter(void) const;
    // transformation
    const R3Affine& Transformation(void) const;
    // visibility
    void SetVisibility(bool isVisible);
    // lock
    void SwitchLockStatus(void);
    void SetLockStatus(bool status);
    bool IsLocked(void) const;
    void SetPermanentLockStatus(bool lock);
    bool IsPermanentlyLocked(void) const;
protected:
    // scene graph structure
    SceneGraph* m_pSceneGraph;
    int m_iIndex;
    SceneGraphNode* m_pParent;
    std::vector<SceneGraphNode*> m_aChildArray;
    // layout parameters
    double m_fSplitDist;
    double m_fSelfRotationInRadian;
    R3Vector m_vRelativeOffsetDirection;
    // global transformation
    R3Affine m_tTransformation;
    bool m_bGeomUpdated; // indicate if the transformation needs to be updated
    // category and models
    std::string m_sCategoryName;
    std::vector<SceneModel*> m_aModelArray;
    int m_iCurrentModelIndex;
    // visibility
    bool m_bVisible;
    // lock status
    bool m_bLock;
    // permanent lock status
    bool m_bPermLock;
};

class SceneGraph
{
public:
    SceneGraph(void);
    ~SceneGraph(void);
    // import and export
    /*model database*/
    int ImportModelDatabase(const std::string& meshReponame, const std::string& descReponame);
    /*compatibilty pairs*/
    int ImportCompatibilityPairs(const std::string& pairFilename);
    /*permanent locks*/
    int InitializePermanentLocks(const std::string& filename);
    /*layout*/
    int ImportPositions(const std::string& filename);
    int ExportPositions(const std::string& filename) const;
    // draw
    void Draw(bool drawOnlyFreeModels = false) const;
    // intersection
    SceneGraphNode* Intersect(const R3Ray& ray, double* pHit) const;
    // access
    SceneGraphNode* Root(void) const;
    int NNodes(void) const;
    SceneGraphNode* Node(int idx) const;
    const std::vector<R3Point> GetBBoxCornerPoints(void) const;
    const std::vector<R3Point> GetBBoxCornerPointsOfFreeObjects(void) const;
    R3Box BBox(void) const;
    R3Box BBoxOfFreeObjects(void) const;
    SceneGraphNode* GetNodeByCategory(const std::string& categoryName) const;
    void AddNode(SceneGraphNode* pNode);
    void UpdateTransformations(void);
    // compatibility pairs
    int NCompatibilityPairs(void) const;
    SceneGraphNode* GetKthCompatibilityPairFirst(int k) const;
    SceneGraphNode* GetKthCompatibilityPairSecond(int k) const;
protected:
    std::vector<SceneGraphNode*> m_aNodeArray;
    SceneGraphNode* m_pRootNode;
    std::vector< std::pair<SceneGraphNode*, SceneGraphNode*> > m_vCompatibilityPairs;
};
#endif