#include "SceneGraph.h"
#include "Utils/Utils.h"
#include "ConfigParser/ConfigParser.h"
#include <fstream>
using namespace std;

SceneModel::
SceneModel(void)
:m_pMesh(NULL),
m_sModelName("unknown")
{}

int SceneModel::
ImportMesh(const string& filename)
{
    delete m_pMesh;
    m_pMesh = new R3Mesh;
    m_pMesh->ReadFile(filename.c_str());
    m_sModelName = FileSystemUtil::GetBasename(filename, false);
    return 1;
}

int SceneModel::
ImportDescriptors(const string& filename)
{
    m_vDesc = vector<double>();
    ifstream fin(filename.c_str());
    double value;
    while (fin >> value)
        m_vDesc.push_back(value);
    fin.close();
    assert(m_sModelName == FileSystemUtil::GetBasename(filename, false));
    return 1;
}

R3Mesh* SceneModel::
GetMesh(void) const
{
    return m_pMesh;
}

const vector<double>& SceneModel::
GetDescriptors(void) const
{
    return m_vDesc;
}

const string& SceneModel::
GetModelName(void) const
{
    return m_sModelName;
}

SceneGraphNode::
SceneGraphNode(SceneGraph* pSceneGraph)
:m_pSceneGraph(NULL),
m_iIndex(-1),
m_pParent(NULL),
m_fSplitDist(0),
m_fSelfRotationInRadian(0),
m_vRelativeOffsetDirection(1, 0, 0),
m_tTransformation(R3identity_affine),
m_bGeomUpdated(false),
m_sCategoryName(""),
m_iCurrentModelIndex(-1),
m_bVisible(true),
m_bLock(false),
m_bPermLock(false)
{
    pSceneGraph->AddNode(this);
}

int SceneGraphNode::
ImportModelDatabase(const string& meshDirname, const string& descDirname)
{
    vector<string> filenameArray = FileSystemUtil::ListFiles(meshDirname, "obj");
    for (int i=0; i<(int)filenameArray.size(); i++) {
        SceneModel* pModel = new SceneModel;
        // mesh
        pModel->ImportMesh(filenameArray[i].c_str());
        // descriptor
        string basename = FileSystemUtil::GetBasename(filenameArray[i], false);
        string descFilename = descDirname+"/"+basename+".sd";
        if (!FileSystemUtil::PathExist(descFilename)) {
            fprintf(stderr, "SceneGraphNode::ImportModelDatabase unable to find descriptor file %s!\n",
                    descFilename.c_str());
        }
        else {
            pModel->ImportDescriptors(descFilename);
        }
        m_aModelArray.push_back(pModel);
    }
    m_iCurrentModelIndex = 0;
    m_bGeomUpdated = true;
    return 1;
}


int SceneGraphNode::
NInstances(void) const
{
    return (int)m_aModelArray.size();
}

R3Mesh* SceneGraphNode::
GetKthInstance(int k) const
{
    assert(k>=0 && k<NInstances());
    return m_aModelArray[k]->GetMesh();
}

const vector<double>& SceneGraphNode::
GetKthInstanceDescriptors(int k) const
{
    assert(k>=0 && k<NInstances());
    return m_aModelArray[k]->GetDescriptors();
}

const string& SceneGraphNode::
GetKthInstanceModelName(int k) const
{
    assert(k>=0 && k<NInstances());
    return m_aModelArray[k]->GetModelName();
}

R3Mesh* SceneGraphNode::
GetInstanceByName(const string& name) const
{
    int index = GetInstanceIndexByName(name);
    if (index == -1) return NULL;
    return GetKthInstance(index);
}

int SceneGraphNode::
GetInstanceIndexByName(const string& name) const
{
    for (int i=0; i<(int)m_aModelArray.size(); i++) {
        if (m_aModelArray[i]->GetModelName() == name) return i;
    }
    return -1;
}

R3Mesh* SceneGraphNode::
GetCurrentModel(void) const
{
    if (m_iCurrentModelIndex == -1) return NULL;
    return GetKthInstance(m_iCurrentModelIndex);
}

int SceneGraphNode::
GetCurrentModelIndex(void) const
{
    return m_iCurrentModelIndex;
}

void SceneGraphNode::
SelectKthModel(int k)
{
    assert(k>=0 && k<NInstances());
    m_iCurrentModelIndex = k;
    m_bGeomUpdated = true;
}

void SceneGraphNode::
Draw(void) const
{
    // push transformation
    m_tTransformation.Push();
    if (GetCurrentModel() && m_bVisible) {
        static GLfloat materialPermlock[] = {0.0, 0.5, 1.0, 1.0};
        static GLfloat materialLock[] = {1, 0.5, 0, 1.0};
        static GLfloat materialUnlock[] = {0.5, 0.5, 0.5, 1.0};
        if (IsPermanentlyLocked())
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialPermlock);
        else if (IsLocked())
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialLock);
        else
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialUnlock);
        GetCurrentModel()->Draw();
    }
    // pop transformations
    m_tTransformation.Pop();
}

void SceneGraphNode::
DrawWhenDragging(const R3Affine& affine) const
{
    m_tTransformation.Push();
    affine.Push();
    if (GetCurrentModel() && m_bVisible) GetCurrentModel()->Draw();
    affine.Pop();
    m_tTransformation.Pop();
}

void SceneGraphNode::
SetCategoryName(const string& name)
{
    m_sCategoryName = name;
}

const string& SceneGraphNode::
CategoryName(void) const
{
    return m_sCategoryName;
}

int SceneGraphNode::
GraphIndex(void) const
{
    return m_iIndex;
}

int SceneGraphNode::
NChildren(void) const
{
    return (int)m_aChildArray.size();
}

SceneGraphNode* SceneGraphNode::
Child(int idx) const
{
    assert(idx>=0&&idx<NChildren());
    return m_aChildArray[idx];
}

SceneGraphNode* SceneGraphNode::
Parent(void) const
{
    return m_pParent;
}

void SceneGraphNode::
InsertChild(SceneGraphNode* pNode)
{
    pNode->m_pParent = this;
    m_aChildArray.push_back(pNode);
}

bool SceneGraphNode::
IsGeometryUpdated(void) const
{
    return m_bGeomUpdated;
}

static const R3Vector
GetSegmentOnBox(const R3Box& box, const R3Vector& direction)
{
    // assume the center is at the bottom center
    R3Vector uniDirection = direction;
    uniDirection.Normalize();
    double sx = box.XLength()*0.5/(1e-3+fabs(uniDirection[0]));
    double sy = 0;
    if (uniDirection[1] >= 0) {
        sy = box.YLength()/(1e-3+fabs(uniDirection[1]));
    }
    double sz = box.ZLength()*0.5/(1e-3+fabs(uniDirection[2]));
    double mins = 0;
    if (sx<sy && sx<sz) mins = sx;
    else if (sy < sz) mins = sy;
    else mins = sz;
    return direction*mins;
}

void SceneGraphNode::
UpdateTransformation(void)
{
    // assume parent node's transformation has been fixed
    if (m_pParent && GetCurrentModel() && (m_pParent->IsGeometryUpdated() || IsGeometryUpdated())) {
        // rotate -> compute new bbox -> rotate
        R3Affine rotation = R3identity_affine;
        rotation.Rotate(R3posy_vector, m_fSelfRotationInRadian);
        // relative position (everthing is computed in parent's coordinate system)
        // segment in parent's bbox
        R3Vector segmentOnParent = GetSegmentOnBox(m_pParent->BBox(), m_vRelativeOffsetDirection);
        // segment in my bbox (local coordinate)
        R3Vector offsetDirection = m_vRelativeOffsetDirection;
        offsetDirection.InverseTransform(rotation);
        R3Vector segmentOnMe = GetSegmentOnBox(BBox(), -offsetDirection);
        // segment in my bbox (parent's coordinate)
        segmentOnMe.Transform(rotation);
        R3Vector totalSegment = segmentOnParent - segmentOnMe + m_vRelativeOffsetDirection*m_fSplitDist;
        R3Vector translationVec = m_pParent->BottomCenter() + totalSegment - BottomCenter();
        // total transformation
        m_tTransformation = m_pParent->m_tTransformation;
        // second: translation
        m_tTransformation.Translate(translationVec);
        // first: rotation
        m_tTransformation.Rotate(R3posy_vector, m_fSelfRotationInRadian);
    }
    // recursing
    for (int i=0; i<NChildren(); i++) {
        Child(i)->UpdateTransformation();
    }
    m_bGeomUpdated = false;
}

//static void
//DecomposeTransformation(const R3Affine& affine, const R3Vector& center_offset, R3Affine* translation, R3Affine* rotation)
//{
//    // rotation part
//    R3Vector v = R3posx_vector;
//    v.Transform(affine);
//    *rotation = R3identity_affine;
//    rotation->Rotate(R3posx_vector, v);
//    
//    // translation part
//    *translation = R3identity_affine;
//    translation->Transform(affine);
//    translation->Translate(center_offset);
//    translation->InverseTransform(*rotation);
//    translation->Translate(-center_offset);
//}


// use m_tTransformation to update m_fSplitDist, m_vRelativeOffsetDirection, m_fSelfRotationInRadian
void SceneGraphNode::
UpdatePositions(void)
{
    if (m_pParent && GetCurrentModel()) {
        // compute the relative transformation
        R3Affine relTransformation = R3identity_affine;
        relTransformation.InverseTransform(m_pParent->m_tTransformation);
        relTransformation.Transform(m_tTransformation);
        // convert rotation to radian
        R3Vector v = R3posx_vector;
        v.Transform(relTransformation);
        m_fSelfRotationInRadian = acos(v.Dot(R3posx_vector));
        if ((R3posx_vector%v).Dot(R3posy_vector) < 0)
            m_fSelfRotationInRadian = -m_fSelfRotationInRadian;
        // rotation
        R3Affine rotation = R3identity_affine;
        rotation.Rotate(R3posy_vector, m_fSelfRotationInRadian);
        // translation
        R3Point zero(0, 0, 0);
        zero.Transform(relTransformation);
        R3Vector translationVec = zero - R3Point(0, 0, 0);
        m_vRelativeOffsetDirection = translationVec;
        m_vRelativeOffsetDirection.Normalize();
        // total segment
        R3Vector totalSegment = BottomCenter() + translationVec - m_pParent->BottomCenter();
        // segment in parent's bbox
        R3Vector segmentOnParent = GetSegmentOnBox(m_pParent->BBox(), m_vRelativeOffsetDirection);
        // segment in my bbox (local coordinate)
        R3Vector offsetDirection = m_vRelativeOffsetDirection;
        offsetDirection.InverseTransform(rotation);
        R3Vector segmentOnMe = GetSegmentOnBox(BBox(), -offsetDirection);
        // segment in my bbox (parent's coordinate)
        segmentOnMe.Transform(rotation);
        
        R3Vector xDir = totalSegment + segmentOnMe - segmentOnParent;
        m_fSplitDist = xDir.Dot(m_vRelativeOffsetDirection);
    }
    m_bGeomUpdated = true;
}

void SceneGraphNode::
Transform(const R3Affine& affine)
{
    // update m_tTransformation
    UpdateTransformation();
    // update transformation of current node
    m_tTransformation.Transform(affine);
    UpdatePositions();
    // update transformation of child nodes
    // (the global transformation of children shouldn't change)
    for (int i=0; i<NChildren(); i++) {
        Child(i)->UpdatePositions();
    }
}

R3Box SceneGraphNode::
BBox(void) const
{
    if (GetCurrentModel()) return GetCurrentModel()->BBox();
    else return R3null_box;
}

R3Point SceneGraphNode::
BottomCenter(void) const
{
    R3Point centroid = BBox().Centroid();
    return R3Point(centroid[0], 0, centroid[2]);
}

const R3Affine& SceneGraphNode::
Transformation(void) const
{
    return m_tTransformation;
}

bool SceneGraphNode::
Intersect(const R3Ray& rayIn, double* pHit) const
{
    R3Ray ray = rayIn;
    ray.InverseTransform(m_tTransformation);
    // hit the current mesh
    bool hitMe = false;
    double hitT = FLT_MAX;
    if (GetCurrentModel() && m_bVisible) {
        R3MeshIntersection intersection;
        hitMe = GetCurrentModel()->Intersection(ray, &intersection);
        if (hitMe) {
            hitT = intersection.t;
        }
    }
    if (pHit) *pHit = hitT;
    return hitMe;
}

void SceneGraphNode::
SetVisibility(bool isVisible)
{
    m_bVisible = isVisible;
}

void SceneGraphNode::
SwitchLockStatus(void)
{
    m_bLock = !m_bLock;
}

void SceneGraphNode::
SetLockStatus(bool status)
{
    m_bLock = status;
}

bool SceneGraphNode::
IsLocked(void) const
{
    return m_bLock;
}

void SceneGraphNode::
SetPermanentLockStatus(bool lock)
{
    m_bPermLock = lock;
}

bool SceneGraphNode::
IsPermanentlyLocked(void) const
{
    return m_bPermLock;
}
