#include <cmath>
#include <set>
#include <map>
#include <limits>
#include <algorithm>
#include "Utils.h"
#include "R3MyGraphics/R3MyGraphics.h"
using namespace std;

vector<double> MathUtil::stored_random_number_array;

typedef vector<RNScalar> ScalarArray;
RNScalar MathUtil::
Sum(const vector<RNScalar>& array)
{
    RNScalar sum = 0;
    for (int i=0; i<int(array.size()); i++)
    {
        sum += array[i];
    }
    return sum;
}

int MathUtil::
Sum(const vector<int>& array)
{
    int sum = 0;
    for (int i=0; i<int(array.size()); i++)
    {
        sum += array[i];
    }
    return sum;
}

void MathUtil::
DivideBySum(vector<double>& array)
{
	double sum = Sum(array);
	if (sum == 0.0) return;
	for (int i=0; i<int(array.size()); i++) {
		array[i] /= sum;
	}
}

RNScalar MathUtil::
Average(const vector<RNScalar>& array)
{
    if (array.size() == 0) return 0.0;
    return Sum(array)/int(array.size());
}

RNScalar MathUtil::
Median(const vector<RNScalar>& array)
{
    if (array.size() == 0) return 0.0;
    vector<RNScalar> sortedArray = array;
    Sort(sortedArray);
    if ((int)sortedArray.size() % 2 == 1) {
        return sortedArray[((int)sortedArray.size() - 1)/2];
    }
    else {
        return (sortedArray[(int)sortedArray.size()/2-1] + sortedArray[(int)sortedArray.size()/2])*0.5;
    }
}

const ScalarArray MathUtil::
Average(const vector<ScalarArray>& array, int nDims)
{
    ScalarArray ans(nDims, 0);
    vector<ScalarArray> data(nDims);
    for (int i=0; i<(int)array.size(); i++) {
        assert(array[i].size() == nDims);
        for (int j=0; j<nDims; j++) {
            data[j].push_back(array[i][j]);
        }
    }
    for (int i=0; i<nDims; i++) {
        ans[i] = Average(data[i]);
    }
    return ans;
}

RNScalar MathUtil::
Maximum(const vector<RNScalar>& array, int* p)
{
    RNScalar max = -FLT_MAX;
    for (int i=0; i<int(array.size()); i++)
    {
        if (array[i] > max)
        {
            max = array[i];
            if (p) *p = i;
        }
    }
    return max;
}

RNScalar MathUtil::
MaximumAbs(const vector<RNScalar>& array, int* p)
{
    RNScalar maxAbs = -FLT_MAX;
    int maxIdx = -1;
    for (int i=0; i<(int)array.size(); i++) {
        double val = fabs(array[i]);
        if (val > maxAbs) {
            maxAbs = val;
            maxIdx = i;
        }
    }
    if (p) *p = maxIdx;
    return maxAbs;
}

int MathUtil::
Maximum(const vector<int>& array, int* p)
{
    int max = 0;
    int maxIdx = -1;
    for (int i=0; i<(int)array.size(); i++) {
        if (array[i] > max || maxIdx == -1) {
            max = array[i];
            maxIdx = i;
        }
    }
    if (p) *p = maxIdx;
    return max;
}

const vector<int> MathUtil::
KthMaximum(const vector<RNScalar>& array, int k)
{
    vector<int> maxIndices;
    vector<bool> visitedFlags(array.size(), false);
    for (int i=0; i<k; i++) {
        int maxIdx = -1;
        double maxNum = 0;
        for (int j=0; j<(int)array.size(); j++) {
            if (visitedFlags[j]) continue;
            if (array[j] > maxNum || maxIdx == -1) {
                maxNum = array[j];
                maxIdx = j;
            }
        }
        if (maxIdx == -1) return maxIndices;
        maxIndices.push_back(maxIdx);
        visitedFlags[maxIdx] = true;
    }
    return maxIndices;
}

const vector<int> MathUtil::
KthMinimum(const vector<RNScalar>& array, int k)
{
    vector<int> minIndices;
    vector<bool> visitedFlags(array.size(), false);
    for (int i=0; i<k; i++) {
        int minIdx = -1;
        double minNum = 0;
        for (int j=0; j<(int)array.size(); j++) {
            if (visitedFlags[j]) continue;
            if (array[j] < minNum || minIdx == -1) {
                minNum = array[j];
                minIdx = j;
            }
        }
        if (minIdx == -1) return minIndices;
        minIndices.push_back(minIdx);
        visitedFlags[minIdx] = true;
    }
    return minIndices;
}

RNScalar MathUtil::
PercentileMedian(const vector<RNScalar>& _array, double percentile)
{
    vector<RNScalar> array = _array;
    Sort(array);
    double idxFloat = percentile*array.size();
    int idx = (int)idxFloat;
    double t = idxFloat - idx;
    if (idx+1 >= (int)array.size()) return array[array.size()-1];
    return array[idx]*(1.0-t) + array[idx+1]*t;
}


RNScalar MathUtil::
Minimum(const vector<RNScalar>& array, int* p)
{
    RNScalar min = FLT_MAX*2.0;
    for (int i=0; i<int(array.size()); i++)
    {
        if (array[i] < min)
        {
            min = array[i];
            if (p) *p = i;
        }
    }
    return min;
}

int MathUtil::
Minimum(const vector<int>& array, int* p)
{
    int min = 1000000000;
    for (int i=0; i<int(array.size()); i++)
    {
        if (array[i] < min)
        {
            min = array[i];
            if (p) *p = i;
        }
    }
    return min;
}

RNScalar MathUtil::
StandardDeviation(const vector<RNScalar>& array)
{
    if (array.size() == 0) return 0.0;
    RNScalar avg = Average(array);
    RNScalar sum = 0;
    for (int i=0; i<int(array.size()); i++)
    {
        RNScalar diff = array[i] - avg;
        sum += diff * diff;
    }
    return sqrt(sum / int(array.size()));
}

const ScalarArray MathUtil::
StandardDeviation(const vector<ScalarArray>& array, int nDims)
{
    ScalarArray ans(nDims, 0);
    vector<ScalarArray> data(nDims);
    for (int i=0; i<(int)array.size(); i++) {
        assert(array[i].size() == nDims);
        for (int j=0; j<nDims; j++) {
            data[j].push_back(array[i][j]);
        }
    }
    for (int i=0; i<nDims; i++) {
        ans[i] = StandardDeviation(data[i]);
    }
    return ans;
}

RNScalar MathUtil::
Sum(const vector<RNScalar>& array, const vector<RNScalar>& weights)
{
    RNScalar sum = 0;
    for (int i=0; i<int(array.size()); i++)
    {
        sum += array[i]*weights[i];
    }
    return sum;
}

RNScalar MathUtil::
Average(const vector<RNScalar>& array, const vector<RNScalar>& weights)
{
    RNScalar sum = Sum(array, weights);
    RNScalar sum_w = Sum(weights);
    if (sum_w != 0)
    {
        sum /= sum_w;
    }
    return sum;
}

RNScalar MathUtil::
StandardDeviation(const vector<RNScalar>& array, const vector<RNScalar>& weights)
{
    RNScalar avg = Average(array, weights);
    vector<RNScalar> diffs;
    for (int i=0; i<int(array.size()); i++)
    {
        RNScalar diff = array[i] - avg;
        diffs.push_back(diff*diff);
    }
    return sqrt(Average(diffs, weights));
}

void MathUtil::
Sort(vector<double>& values, bool isAscending)
{
    if (isAscending) {
        AscendingComparator comparator;
        std::sort(values.begin(), values.end(), comparator);
    }
    else {
        DescendingComparator comparator;
        std::sort(values.begin(), values.end(), comparator);
    }
}

void MathUtil::
Sort(const vector<double>& values, vector<int>& ranked_index, bool isAscending)
{
    vector< pair<double, int> > valueIndexPairs(values.size());
    for (int i=0; i<(int)values.size(); i++) {
        valueIndexPairs[i].first = values[i];
        valueIndexPairs[i].second = i;
    }
    if (isAscending) {
        AscendingPairComparator comparator;
        std::sort(valueIndexPairs.begin(), valueIndexPairs.end(), comparator);
    }
    else {
        DescendingPairComparator comparator;
        std::sort(valueIndexPairs.begin(), valueIndexPairs.end(), comparator);
    }
    ranked_index = vector<int>(values.size());
    for (int i=0; i<(int)ranked_index.size(); i++) {
        ranked_index[i] = valueIndexPairs[i].second;
    }
}

void MathUtil::
BubbleSort(vector<RNScalar>& values)
{
	vector<int> ranking;
	BubbleSort(values, ranking);
	vector<RNScalar> sorted_array;
	for (int i=0; i<int(ranking.size()); i++) {
		sorted_array.push_back(values[ranking[i]]);
	}
	values = sorted_array;
}

// sort to ascending order
void MathUtil::
BubbleSort(const vector<RNScalar>& values,
		   vector<int>& ranked_index)
{
    int n = values.size();
    ranked_index.clear();
    for (int i=0; i<n; i++)
    {
        ranked_index.push_back(i);
    }
    
    for (int i=0; i<n-1; i++)
    {
        for (int j=0; j<n-1-i; j++)
        {
            if (values[ranked_index[j]] > values[ranked_index[j+1]])
            {
                // swap j and j+1
                int temp = ranked_index[j+1];
                ranked_index[j+1] = ranked_index[j];
                ranked_index[j] = temp;
            }
        }
    }
}

void MathUtil::
BubbleSort(vector<int>& values)
{
	vector<int> ranking;
	BubbleSort(values, ranking);
	vector<int> sorted_array;
	for (int i=0; i<int(ranking.size()); i++) {
		sorted_array.push_back(values[ranking[i]]);
	}
	values = sorted_array;
}

void MathUtil::
BubbleSort(const vector<int>& values,
		   vector<int>& ranked_index)
{
    int n = values.size();
    ranked_index.clear();
    for (int i=0; i<n; i++)
    {
        ranked_index.push_back(i);
    }
    
    for (int i=0; i<n-1; i++)
    {
        for (int j=0; j<n-1-i; j++)
        {
            if (values[ranked_index[j]] > values[ranked_index[j+1]])
            {
                // swap j and j+1
                int temp = ranked_index[j+1];
                ranked_index[j+1] = ranked_index[j];
                ranked_index[j] = temp;
            }
        }
    }
}


const vector<int>  MathUtil::
Intersect(const vector<int>& array1, const vector<int>& array2)
{
    set<int> setArray2;
    for (int i=0; i<(int)array2.size(); i++) {
        setArray2.insert(array2[i]);
    }
    vector<int> intersection;
    for (int i=0; i<(int)array1.size(); i++) {
        if (setArray2.find(array1[i]) != setArray2.end())
            intersection.push_back(array1[i]);
    }
    return intersection;
}

bool MathUtil::
IsSubSetOfSortedArray(const vector<int>& subset_array,
					  const vector<int>& full_array)
{
	assert(IsSemiAscendinglySorted(subset_array));
	assert(IsSemiAscendinglySorted(full_array));
	int p_in_full_array = 0;
	int p_in_subset_array = 0;
	while (p_in_full_array < int(full_array.size()) &&
		   p_in_subset_array < int(subset_array.size())) {
		if (subset_array[p_in_subset_array] == full_array[p_in_full_array]) {
			p_in_full_array ++;
			p_in_subset_array ++;
		}
		else {
			p_in_full_array ++;
		}
	}
	if (p_in_subset_array == int(subset_array.size())) return true;
	else return false;
}

bool MathUtil::
IsIdenticalToSortedArray(const vector<int>& array1,
						 const vector<int>& array2)
{
	if (array1.size() != array2.size()) return false;
	assert(IsSemiAscendinglySorted(array1));
	assert(IsSemiAscendinglySorted(array2));
	for (int i=0; i<int(array1.size()); i++) {
		if (array1[i] != array2[i]) return false;
	}
	return true;
}

bool MathUtil::
IsIdenticalToArray(const vector<int>& array1, const vector<int>& array2)
{
    if (array1.size() != array2.size()) return false;
    for (int i=0; i<(int)array1.size(); i++) {
        if (array1[i] != array2[i]) return false;
    }
    return true;
}

bool MathUtil::
IsAscendinglySorted(const vector<int>& array)
{
	// array[i] < array[i+1] should be always true
	for (int i=0; i<int(array.size())-1; i++) {
		if (array[i] >= array[i+1]) return false;
	}
	return true;
}

bool MathUtil::
IsSemiAscendinglySorted(const vector<int>& array)
{
	// array[i] < array[i+1] should be always true
	for (int i=0; i<int(array.size())-1; i++) {
		if (array[i] > array[i+1]) return false;
	}
	return true;
}

int MathUtil::
RandomInt(int range)
{
    int rand_int = rand()%range;
    return rand_int;
}

void MathUtil::
InitializeStoredRandomNumbers(int numNumbers)
{
    stored_random_number_array.clear();
    stored_random_number_array.resize(numNumbers, 0);
    for (int i=0; i<numNumbers; i++) {
        stored_random_number_array[i] = RandomNumber(1.0);
    }
}

// we precompute some random numbers from [0, 1.0],
// we fetch them when needed
double MathUtil::
RandomNumberFromStored(double range)
{
    static bool isInitialized = false;
    static int counter = 0;
    static const int numNumbers = 4096;
    if (!isInitialized) {
        InitializeStoredRandomNumbers(numNumbers);
        counter = 0;
        isInitialized = true;
    }
    counter = (counter + 1) % numNumbers;
    return stored_random_number_array[counter]*range;
}


double MathUtil::
RandomNumber(double range)
{
    int int_range = 4096;
    int rand_num = RandomInt(int_range);
    return rand_num*range/int_range;
}

int MathUtil::
RandomItemIndex(const vector<double>& probs)
{
	// linear time solution
	vector<double> norm_probs = probs;
	double sum = Sum(probs);
	if (sum != 0){
		for (int i=0; i<int(norm_probs.size()); i++) {
			norm_probs[i] /= sum;
		}
	}
	double rand = RandomNumber(1.0);
	double sum_so_far = 0;
	for (int i=0; i<int(norm_probs.size()); i++) {
		sum_so_far += norm_probs[i];
		if (sum_so_far > rand) {
			return i;
		}
	}
//	assert(false);
	return int(norm_probs.size())-1;
}

double MathUtil::
RandomFromNormalDistribution(double mean, double sigma)
{
    double u1 = RandomNumber(1.0);
    double u2 = RandomNumber(1.0);
    if (u1 == 0.0)
    {
        u1 = 1e-4;
    }
    return sqrt(-2.0*log(u1))*cos(2*RN_PI*u2)*sigma + mean;
}

bool MathUtil::
AcceptAccordingToRate(double acceptRate)
{
    double rand = RandomNumber(1.0);
    if (rand <= acceptRate) return true;
    return false;
}

int MathUtil::
Power(int base, int exp)
{
	if (base>10 && exp>10) return -1;
	int prod = 1;
	for (int i=0; i<exp; i++) {
		prod *= base;
	}
	return prod;
}

/*convert an integer (index) to an array with num_digits(N) digits, each
 digit is 0~max_per_digit(M)-1
 the range of integer is (0 ~ pow(M,N)-1)*/
const vector<int> MathUtil::
Integer2MNaryArray(int num_digits, int max_per_digit, int index)
{
	vector<int> array(num_digits, 0);
	for (int i=0; i<num_digits; i++) {
		array[i] = index % max_per_digit;
		index = index / max_per_digit;
	}
	assert(index == 0);
	return array;
}

const vector<int> MathUtil::
RandomSubset(int target_sz, int total_sz)
{
//	printf("Beigin random subset ...\n");
	assert(target_sz <= total_sz);
	vector<bool> visited_flag(total_sz, false);
	vector<int> subset;
	while (int(subset.size()) < target_sz) {
		int rand = RandomInt(total_sz);
		if (visited_flag[rand]) {
			while (visited_flag[rand]) {
				rand = (rand + 1)%total_sz;
			}
		}
		subset.push_back(rand);
		visited_flag[rand] = true;
	}
//	printf("end random subset!\n");
	return subset;
}



bool MathUtil::
IsNaN(double val)
{
    return (val != val);
}

bool MathUtil::
IsInf(double val)
{
    return std::numeric_limits<double>::has_infinity &&
    val == std::numeric_limits<double>::infinity();
}

// targetNum is the target number of items we want to end up with
const vector< vector<int> > MathUtil::
EnumerateSubset(const vector<int>& pool, int size, int targetNum, int startIdx)
{
    if (targetNum == 0) {
        vector<int> array;
        vector<vector<int> > returnArrays;
        returnArrays.push_back(array);
        return returnArrays;
    }
    int maxIdx = pool.size() - targetNum;
    assert(maxIdx >= startIdx);
    vector<vector<int> > returnArrays;
    for (int i=startIdx; i<=maxIdx; i++) {
        vector<vector<int> > ansArrays = EnumerateSubset(pool, size-1, targetNum-1, i+1);
        for (int j=0; j<(int)ansArrays.size(); j++) {
            ansArrays[j].push_back(pool[i]);
            returnArrays.push_back(ansArrays[j]);
        }
    }
    return returnArrays;
}

void MathUtil::
EnumerateSubsetInFlags(vector<int>& curFlags, vector< std::vector<int> >& subsetArrayInFlags,
                       const vector<int>& candFlags, int startIdx)
{
    if (startIdx >= (int)curFlags.size()) {
        subsetArrayInFlags.push_back(curFlags);
    }
    else {
        for (int i=0; i<(int)candFlags.size(); i++) {
            curFlags[startIdx] = candFlags[i];
            EnumerateSubsetInFlags(curFlags, subsetArrayInFlags, candFlags, startIdx+1);
        }
    }
}

int MathUtil::
Choose(int numCands, int numSelected)
{
    if (numCands<numSelected) return 0;
    int prod = 1;
    for (int x=numCands; x>numCands-numSelected; x--) {
        prod *= x;
    }
    for (int x=1; x<=numSelected; x++) {
        prod /= x;
    }
    return prod;
}

// begin with root node, and end with leaf nodes
const vector<int> MathUtil::
ComputeTopologicalOrder(const vector< pair<int,int> >& edges)
{
    vector<int> visitedNodeArray;
    set<int> visitedNodeSet;
    int pointer = 0;
    vector<bool> edgeDiscardedFlags(edges.size(), false);
    // map<pair<int,int>,int> edgeIdxMap;
    map<int, vector<int> > outgoingEdgeIDs;
    map<int, vector<int> > incomingEdgeIDs;
    set<int> nodeArray;
    for (int i=0; i<(int)edges.size(); i++) {
        // edgeIdxMap[edges[i]] = i;
        outgoingEdgeIDs[edges[i].first].push_back(i);
        incomingEdgeIDs[edges[i].second].push_back(i);
        nodeArray.insert(edges[i].first);
        nodeArray.insert(edges[i].second);
    }
    
    // initialize visited node array
    for (set<int>::const_iterator it=nodeArray.begin();
         it!=nodeArray.end(); it++) {
        int node = *it;
        if (incomingEdgeIDs.find(node) == incomingEdgeIDs.end()) {
            visitedNodeArray.push_back(node);
            visitedNodeSet.insert(node);
        }
    }
    while (pointer < visitedNodeArray.size()) {
        int node = visitedNodeArray[pointer];
        pointer ++;
        if (outgoingEdgeIDs.find(node) == outgoingEdgeIDs.end()) continue;
        // go through all outgoing edges,
        // and detect candidates to add into visited
        vector<int> edgeIDs = outgoingEdgeIDs[node];
        vector<int> toCheckNodes;
        for (int i=0; i<(int)edgeIDs.size(); i++) {
            pair<int,int> edge = edges[edgeIDs[i]];
            assert(edge.first == node);
            edgeDiscardedFlags[edgeIDs[i]] = true;
            toCheckNodes.push_back(edge.second);
        }
        for (int i=0; i<(int)toCheckNodes.size(); i++) {
            int candNode = toCheckNodes[i];
            // assert the node hasn't been added yet
            assert(visitedNodeSet.find(candNode) == visitedNodeSet.end());
            vector<int> inEdgeIDs = incomingEdgeIDs[candNode];
            bool allRemoved = true;
            for (int j=0; j<(int)inEdgeIDs.size(); j++) {
                if (!edgeDiscardedFlags[inEdgeIDs[j]]) {
                    allRemoved = false;
                    break;
                }
            }
            if (allRemoved) visitedNodeArray.push_back(candNode);
        }
    }
    bool allRemoved = true;
    for (int i=0; i<(int)edgeDiscardedFlags.size(); i++) {
        if (!edgeDiscardedFlags[i]) {
            allRemoved = false;
            break;
        }
    }
    if (!allRemoved) {
        fprintf(stderr, "MathUtil::ComputeTopologicalOrder some edges are not removed in the end!\n");
    }
    return visitedNodeArray;
}

int MathUtil::
Factorial(int k)
{
    int prod = 1;
    for (int i=2; i<=k; i++)
        prod *= i;
    return prod;
}

double MathUtil::
KernelTriangular(double u)
{
    if (u <= 1 && u >= 0) return 1-u;
    else if (u >= -1 && u < 0) return 1+u;
    else return 0;
}

double MathUtil::
KernelGaussian(double u)
{
    return 0.39894228*fastMinusExp(0.5*u*u);
}

double MathUtil::
KernelGaussianInLogSpace(double u)
{
    return 0.5*u*u + 0.9189385332;
}

/** Computes a fast approximation to exp(-x). */
double MathUtil::
fastMinusExp(double x)
{
    return exp(-x);
    //return fastMinuzExpWider_1(x);
}

// widen up fastMinuzExp
float MathUtil::
fastMinuzExpWider_1(float x)
{
    bool lessZero = false;
    
    if (x < 0)
    {
        lessZero = true;
        x = -x;
    }
    
    int mult = 0;
    
    while (x > 0.69 * 2 * 2 * 2 * 2 * 2 * 2)
    {
        mult += 6;
        x /= 64.0f;
    }
    
    while (x > 0.69 * 2 * 2 * 2)
    {
        mult += 3;
        x /= 8.0f;
    }
    
    while (x > 0.69 * 2 * 2)
    {
        mult += 2;
        x /= 4.0f;
    }
    
    while (x > 0.69)
    {
        mult++;
        x /= 2.0f;
    }
    
    x = fastMinuzExp_1(x);
    
    while (mult)
    {
        mult--;
        x = x * x;
    }
    
    if (lessZero)
    {
        return 1 / x;
    }
    else
    {
        return x;
    }
}

float MathUtil::
fastMinuzExp_1(float x)
{
    // err <= 3e-3
    return (float)(1
                   - x * (0.9664
                          - x * (0.3536)));
}

const vector<double> MathUtil::
RemoveOutliers(const vector<double>& array, double percentile)
{
    vector<double> result = array;
    std::sort(result.begin(), result.end());
    double outlierValueMax = result[(int)floor(percentile*result.size())];
    double outlierValueMin = result[(int)floor((1.0-percentile)*result.size())];
    // clamp values that > outlierValueMax or < outlierValueMin
    for (int i=0; i<(int)array.size(); i++) {
        if (array[i] < outlierValueMin) {
            result[i] = outlierValueMin;
        }
        else if (array[i] > outlierValueMax) {
            result[i] = outlierValueMax;
        }
        else {
            result[i] = array[i];
        }
    }
    return result;
}

const vector<double> MathUtil::
GenerateHistogram(const vector<double>& samples, int numBins, double minVal, double maxVal)
{
    vector<double> values(numBins, 0.0);
    double spacing = (maxVal - minVal) / numBins;
    for (int i=0; i<(int)samples.size(); i++) {
        double t = (samples[i] - minVal - spacing*0.5)/spacing;
        if (t < 0) t = 0;
        else if (t > numBins-1) t = numBins - 1;
        int floorIdx = int(t);
        int ceilingIdx = floorIdx+1;
        t = t - floorIdx;
        // floorIdx : 1 - t
        if (floorIdx>=0 && floorIdx<numBins) {
            values[floorIdx] += 1 - t;
        }
        if (ceilingIdx>=0 && ceilingIdx<numBins) {
            values[ceilingIdx] += t;
        }
    }
    // normalize
    DivideBySum(values);
    return values;
}

const vector<double> MathUtil::
GenerateCumulativeHistogram(const vector<double>& histogram)
{
    vector<double> cumHist = histogram;
    for (int i=1; i<(int)cumHist.size(); i++) {
        cumHist[i] = cumHist[i] + cumHist[i-1];
    }
    return cumHist;
}

const R3Triad MathUtil::
GenerateTriad(const R3Vector& _zDir)
{
    R3Vector zDir = _zDir;
    zDir.Normalize();
    R3Vector xDir = R3posx_vector;
    if (fabs(xDir.Dot(zDir)) > 0.99) {
        xDir = R3posy_vector;
    }
    R3Vector yDir = zDir % xDir;
    yDir.Normalize();
    xDir = yDir % zDir;
    xDir.Normalize();
    return R3Triad(xDir, yDir, zDir);
}

void MathUtil::SparseMatrix::
Write(const string& filename) const
{
    if (xs.size() != ys.size() || xs.size() != values.size()) {
        fprintf(stderr, "SparseMatrix::Write sizes of xs, ys, values don't match!\n");
        exit(-1);
    }
    ofstream fout(filename.c_str());
    if (!fout) {
        fprintf(stderr, "SparseMatrix::Write unable to open %s!\n", filename.c_str());
        exit(-1);
    }
    fout << m << " " << n << endl;
    for (int i=0; i<(int)xs.size(); i++) {
        fout << xs[i] << " " << ys[i] << " " << values[i] << endl;
    }
    fout.close();
}

const vector<double> MathUtil::SparseMatrix::
SolveEigenvaluesUsingPython(const string& solverScriptName,
                            const string& matrixFilename,
                            const string& eigenValueFilename) const
{
    // write matrix
    Write(matrixFilename);
    // execute script
    string cmd("python ");
    cmd = cmd + solverScriptName+" "+matrixFilename+" "+eigenValueFilename;
    FileSystemUtil::Execute(cmd);
    // get back eigenvalues
    vector<double> eigenvalues = ReadEigenvalues(eigenValueFilename);
    return eigenvalues;
}

const vector<double> MathUtil::SparseMatrix::
ReadEigenvalues(const string& filename) const
{
    ifstream fin(filename.c_str());
    if (!fin) {
        fprintf(stderr, "SparseMatrix::ReadEigenvalues unable to open %s!\n", filename.c_str());
        exit(-1);
    }
    vector<double> values;
    double value;
    while (fin >> value) values.push_back(value);
    fin.close();
    return values;
}

const vector<double> MathUtil::
ResampleVector(const vector<double>& orgVector, int numDims)
{
    vector<double> ratioVector(numDims);
    for (int i=0; i<numDims; i++) {
        ratioVector[i] = double(i)*(int)orgVector.size() / numDims;
    }
    for (int i=0; i<numDims; i++) {
        int idx = int(ratioVector[i]);
        double t = ratioVector[i] - idx;
        if (idx < 0) {
            ratioVector[i] = orgVector[0];
        }
        else if (idx >= (int)orgVector.size()-1) {
            ratioVector[i] = orgVector[orgVector.size()-1];
        }
        else {
            ratioVector[i] = orgVector[idx]*(1.0-t) + orgVector[idx+1]*t;
        }
    }
    return ratioVector;
}

const R2Point MathUtil::
Mean(const vector<R2Point>& positionArray)
{
    R2Point center(0, 0);
    for (int i=0; i<(int)positionArray.size(); i++) {
        center = center + positionArray[i].Vector();
    }
    center[0] /= positionArray.size();
    center[1] /= positionArray.size();
    return center;
}

// G. Taubin, "Estimation Of Planar Curves, Surfaces And Nonplanar
//             Space Curves Defined By Implicit Equations, With
//             Applications To Edge And Range Image Segmentation",
// IEEE Trans. PAMI, Vol. 13, pages 1115-1138, (1991)
double MathUtil::
FitCircle(const vector<R2Point>& positionArray, const vector<R2Vector>& tangentArray, R2Point* pCenter)
{
    R2Point centroid = Mean(positionArray);
    double Mxx = 0, Myy = 0, Mxy = 0, Mxz = 0, Myz = 0, Mzz = 0;
    for (int i=0; i<(int)positionArray.size(); i++) {
        double Xi = positionArray[i][0] - centroid[0]; //  centering data
        double Yi = positionArray[i][1] - centroid[1]; //  centering data
        double Zi = Xi*Xi + Yi*Yi;
        Mxy = Mxy + Xi*Yi;
        Mxx = Mxx + Xi*Xi;
        Myy = Myy + Yi*Yi;
        Mxz = Mxz + Xi*Zi;
        Myz = Myz + Yi*Zi;
        Mzz = Mzz + Zi*Zi;
    }
    int n = positionArray.size();
    Mxx = Mxx/n;
    Myy = Myy/n;
    Mxy = Mxy/n;
    Mxz = Mxz/n;
    Myz = Myz/n;
    Mzz = Mzz/n;
    
    // computing the coefficients of the characteristic polynomial
    double Mz = Mxx + Myy;
    double Cov_xy = Mxx*Myy - Mxy*Mxy;
    double A3 = 4*Mz;
    double A2 = -3*Mz*Mz - Mzz;
    double A1 = Mzz*Mz + 4*Cov_xy*Mz - Mxz*Mxz - Myz*Myz - Mz*Mz*Mz;
    double A0 = Mxz*Mxz*Myy + Myz*Myz*Mxx - Mzz*Cov_xy - 2*Mxz*Myz*Mxy + Mz*Mz*Cov_xy;
    double A22 = A2 + A2;
    double A33 = A3 + A3 + A3;
    
    double xnew = 0;
    double ynew = 1e+20;
    double epsilon = 1e-12;
    double IterMax = 20;
    
    // Newton's method starting at x=0
    for (int iter=0; iter<IterMax; iter++) {
        double yold = ynew;
        ynew = A0 + xnew*(A1 + xnew*(A2 + xnew*A3));
        if (fabs(ynew) > fabs(yold)) {
            fprintf(stderr, "Newton-Taubin goes wrong direction: |ynew| > |yold|!\n");
            xnew = 0;
            break;
        }
        double Dy = A1 + xnew*(A22 + xnew*A33);
        double xold = xnew;
        xnew = xold - ynew/Dy;
        // printf("ynew = %g, xold = %g, xnew = %g\n", ynew, xold, xnew);
        if (fabs((xnew-xold)/xnew) < epsilon) {
            printf("Newton-Taubin converges in %d iterations!\n", iter);
            break;
        }
        if (iter == IterMax-1) {
            fprintf(stderr, "Newton-Taubin will not converge (last relative change = %g)!\n",
                    fabs((xnew-xold)/xnew));
            xnew = 0;
        }
        if (xnew < 0.0) {
            fprintf(stderr, "Newton-Taubin negative root:  x=%g\n", xnew);
            xnew = 0;
        }
    }
    
    // computing the circle parameters
    double DET = xnew*xnew - xnew*Mz + Cov_xy;
    if (DET == 0) return FLT_MAX;
    R2Point center((Mxz*(Myy-xnew)-Myz*Mxy)/DET/2.0, (Myz*(Mxx-xnew)-Mxz*Mxy)/DET/2.0);
    double radius = sqrt(center[0]*center[0]+center[1]*center[1]+Mz);
    center = center + centroid.Vector();
    if (pCenter) *pCenter = center;
    return radius;
}

//const R2Point MathUtil::
//FitCircle(const vector<R2Point>& positionArray, const vector<R2Vector>& tangentArray)
//{
//    double mmx = 0;
//    double mnx = 0;
//    double mm = 0;
//    double mn = 0;
//    double nn = 0;
//    double nny = 0;
//    double mny = 0;
//    assert(positionArray.size() == normalArray.size());
//    for (int i=0; i<(int)positionArray.size(); i++) {
//        R2Point pos = positionArray[i];
//        R2Vector vec = tangentArray[i];
//        mmx += vec[0]*vec[0]*pos[0];
//        mnx += vec[0]*vec[1]*pos[0];
//        mm += vec[0]*vec[0];
//        mn += vec[0]*vec[1];
//        nn += vec[1]*vec[1];
//        nny += vec[1]*vec[1]*pos[1];
//        mny += vec[0]*vec[1]*pos[1];
//    }
//    // mm*p + mn*q = mmx + mny
//    // mn*p + nn*q = mnx + nny
//    double u = mmx + mny;
//    double v = mnx + nny;
//    double a = mm;
//    double b = mn;
//    double c = nn;
//    if (b*b == a*c) return R2Point(FLT_MAX, FLT_MAX);
//    return R2Point((v*b-u*c)/(b*b-a*c), (u*b-v*a)/(b*b-a*c));
//}

double MathUtil::
Curvature(const vector<R3Point>& positionArray, const vector<R3Vector>& tangentArray)
{
    if (positionArray.size() == 0) {
        printf("MathUtil::Curvature must have more than 1 points!\n");
        exit(-1);
    }
    // compute principle axes for all the points
    R3Point centroid(0, 0, 0);
    for (int i=0; i<(int)positionArray.size(); i++) {
        centroid[0] += positionArray[i][0];
        centroid[1] += positionArray[i][1];
        centroid[2] += positionArray[i][2];
    }
    centroid[0] /= positionArray.size();
    centroid[1] /= positionArray.size();
    centroid[2] /= positionArray.size();
    
    // Compute covariance matrix
    RNScalar m[9] = { 0 };
    for (int i=0; i<(int)positionArray.size(); i++) {
        R3Point pos = positionArray[i];
        RNScalar x = pos[0] - centroid[0];
        RNScalar y = pos[1] - centroid[1];
        RNScalar z = pos[2] - centroid[2];
        m[0] += x*x;
        m[4] += y*y;
        m[8] += z*z;
        m[1] += x*y;
        m[3] += x*y;
        m[2] += x*z;
        m[6] += x*z;
        m[5] += y*z;
        m[7] += y*z;
    }

    for (int i = 0; i < 9; i++) m[i] /= positionArray.size();
    
    // Compute eigenvalues and eigenvectors
    RNScalar U[9];
    RNScalar W[3];
    RNScalar Vt[9];
    RNSvdDecompose(3, 3, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
    
    // Copy principle axes into more convenient form
    // W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
    R3Vector axes[3];
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            axes[i][j] = Vt[3*i+j];
        }
    }
    axes[0].Normalize();
    axes[1].Normalize();
    // printf("Axes[2] = [%g %g %g]\n", axes[2][0], axes[2][1], axes[2][2]);
    // project all positions and tangent vectors to the first two axes
    vector<R2Point> position2DArray;
    for (int i=0; i<(int)positionArray.size(); i++) {
        R3Vector diff = positionArray[i] - positionArray[0];
        position2DArray.push_back(R2Point(diff.Dot(axes[0]), diff.Dot(axes[1])));
    }
    vector<R2Vector> tangent2DArray;
    for (int i=0; i<(int)tangentArray.size(); i++) {
        tangent2DArray.push_back(R2Vector(tangentArray[i].Dot(axes[0]), tangentArray[i].Dot(axes[1])));
    }
    // fit a 2D circle
    double radius = FitCircle(position2DArray, tangent2DArray);
    return 1.0/radius;
}

double MathUtil::
Tapering(const vector<double>& yArray, const vector<double>& xArray)
{
    assert(yArray.size() == xArray.size());
    double sumXY = 0;
    double sumXX = 0;
    double sumX = 0;
    double sumY = 0;
    for (int i=0; i<(int)yArray.size(); i++) {
        sumXY += xArray[i]*yArray[i];
        sumXX += xArray[i]*xArray[i];
        sumX += xArray[i];
        sumY += yArray[i];
    }
    double k = (yArray.size()*sumXY - sumX*sumY) / (yArray.size()*sumXX - sumX*sumX);
    return fabs(k);
}

double MathUtil::
Correlation(const vector<double>& _data1, const vector<double>& _data2)
{
    vector<double> data1 = _data1;
    vector<double> data2 = _data2;
    double s1 = 0;
    for (int i=0; i<(int)data1.size(); i++) {
        s1 += data1[i]*data1[i];
    }
    s1 = sqrt(s1);
    for (int i=0; i<(int)data1.size(); i++) {
        data1[i] /= s1;
    }
    
    double s2 = 0;
    for (int i=0; i<(int)data2.size(); i++) {
        s2 += data2[i]*data2[i];
    }
    s2 = sqrt(s2);
    for (int i=0; i<(int)data2.size(); i++) {
        data2[i] /= s2;
    }
    assert(data1.size() == data2.size());
    double s = 0;
    for (int i=0; i<(int)data1.size(); i++) {
        s += data1[i]*data2[i];
    }
    return s;
}

//double MathUtil::
//CurvatureDebugging(const vector<R3Point>& positionArray, const vector<R3Vector>& tangentArray)
//{
//    if (positionArray.size() == 0) {
//        printf("MathUtil::Curvature must have more than 1 points!\n");
//        exit(-1);
//    }
//    // compute principle axes for all the points
//    R3Point centroid(0, 0, 0);
//    for (int i=0; i<(int)positionArray.size(); i++) {
//        centroid[0] += positionArray[i][0];
//        centroid[1] += positionArray[i][1];
//        centroid[2] += positionArray[i][2];
//    }
//    centroid[0] /= positionArray.size();
//    centroid[1] /= positionArray.size();
//    centroid[2] /= positionArray.size();
//    
//    // Compute covariance matrix
//    RNScalar m[9] = { 0 };
//    for (int i=0; i<(int)positionArray.size(); i++) {
//        R3Point pos = positionArray[i];
//        RNScalar x = pos[0] - centroid[0];
//        RNScalar y = pos[1] - centroid[1];
//        RNScalar z = pos[2] - centroid[2];
//        m[0] += x*x;
//        m[4] += y*y;
//        m[8] += z*z;
//        m[1] += x*y;
//        m[3] += x*y;
//        m[2] += x*z;
//        m[6] += x*z;
//        m[5] += y*z;
//        m[7] += y*z;
//    }
//    
//    for (int i = 0; i < 9; i++) m[i] /= positionArray.size();
//    
//    // Compute eigenvalues and eigenvectors
//    RNScalar U[9];
//    RNScalar W[3];
//    RNScalar Vt[9];
//    RNSvdDecompose(3, 3, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
//    
//    // Copy principle axes into more convenient form
//    // W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
//    R3Vector axes[3];
//    for (int i = 0; i < 3; i++) {
//        for (int j = 0; j < 3; j++) {
//            axes[i][j] = Vt[3*i+j];
//        }
//    }
//    axes[0].Normalize();
//    axes[1].Normalize();
//    // printf("Axes[2] = [%g %g %g]\n", axes[2][0], axes[2][1], axes[2][2]);
//    // project all positions and tangent vectors to the first two axes
//    vector<R2Point> position2DArray;
//    for (int i=0; i<(int)positionArray.size(); i++) {
//        R3Vector diff = positionArray[i] - positionArray[0];
//        position2DArray.push_back(R2Point(diff.Dot(axes[0]), diff.Dot(axes[1])));
//    }
//    vector<R2Vector> tangent2DArray;
//    for (int i=0; i<(int)tangentArray.size(); i++) {
//        tangent2DArray.push_back(R2Vector(tangentArray[i].Dot(axes[0]), tangentArray[i].Dot(axes[1])));
//    }
//    // fit a 2D circle
//    R2Point circleCenter(0, 0);
//    double radius = FitCircle(position2DArray, tangent2DArray, &circleCenter);
//    // write all points to a file
//    PointCloud localPatch;
//    for (int i=0; i<(int)position2DArray.size(); i++) {
//        localPatch.AddVertex(R3Point(position2DArray[i][0], position2DArray[i][1], 0),
//                             R3Vector(tangent2DArray[i][0], tangent2DArray[i][1], 0));
//    }
//    localPatch.AddVertex(R3Point(circleCenter[0], circleCenter[1], 0), R3Vector(0, 0, 1));
//    localPatch.WriteFile("localpatch-debugging.pts");
//    return 1.0/radius;
//}

double MathUtil::
Combination(int k, int n)
{
    if (k > n) return 0;
    if (k * 2 > n) k = n-k;
    if (k == 0) return 1;
    
    double result = n;
    for( int i = 2; i <= k; ++i ) {
        result *= (n-i+1);
        result /= i;
    }
    //printf("result = %g\n", result);
    return result;
}

double MathUtil::
BinomialSf(int x, int total, double randomProb)
{
    double accProb = 0;
    for (int index=x; index<=total; index++) {
        double combination = Combination(index, total);
        double prob = combination;
        for (int j=0; j<index; j++) {
            prob*=randomProb;
        }
        for (int j=index; j<total; j++) {
            prob*=1.0 - randomProb;
        }
        //printf("prob = %g, randomProb = %g\n", prob, randomProb);
        accProb += prob;
    }
    return accProb;
}

R3Point MathUtil::
GetCameraOriginThatCoversScene(const R3Point& sceneCenter,
                               const R3Vector& cameraTowards,
                               const R3Vector& cameraUp,
                               double xfov,
                               double yfov,
                               const vector<R3Point>& keyPoints)
{
    R3Vector cameraRight = cameraTowards%cameraUp;
    cameraRight.Normalize();
    // center
    double maxx = -FLT_MAX, minx = FLT_MAX;
    double maxy = -FLT_MAX, miny = FLT_MAX;
    for (int i=0; i<(int)keyPoints.size(); i++) {
        R3Vector offset = keyPoints[i] - sceneCenter;
        double xOffset = offset.Dot(cameraRight);
        double yOffset = offset.Dot(cameraUp);
        double zOffset = offset.Dot(cameraTowards);
        double minxCand = xOffset + tan(xfov)*zOffset;
        double maxxCand = xOffset - tan(xfov)*zOffset;
        double minyCand = yOffset + tan(yfov)*zOffset;
        double maxyCand = yOffset - tan(yfov)*zOffset;
        if (minxCand < minx) minx = minxCand;
        if (maxxCand > maxx) maxx = maxxCand;
        if (minyCand < miny) miny = minyCand;
        if (maxyCand > maxy) maxy = maxyCand;
    }
    R3Point sceneCenterNew = sceneCenter + (maxx+minx)*0.5*cameraRight + (maxy+miny)*0.5*cameraUp;
    // r
    double maxR = 0;
    for (int i=0; i<(int)keyPoints.size(); i++) {
        R3Vector offset = keyPoints[i] - sceneCenterNew;
        double xOffset = offset.Dot(cameraRight);
        double yOffset = offset.Dot(cameraUp);
        double zOffset = offset.Dot(cameraTowards);
        double rx = fabs(xOffset)/tan(xfov) - zOffset;
        double ry = fabs(yOffset)/tan(yfov) - zOffset;
        if (rx > maxR) maxR = rx;
        if (ry > maxR) maxR = ry;
    }
    return sceneCenterNew - cameraTowards*maxR;
}

double MathUtil::
EuclideanDistance(const vector<double>& feat1, const vector<double>& feat2)
{
    assert(feat1.size() == feat2.size());
    double sum = 0;
    for (int i=0; i<(int)feat1.size(); i++) {
        double diff = feat1[i] - feat2[i];
        sum += diff*diff;
    }
    return sqrt(sum);
}

